/*
* Return offset from GMT, measured in seconds.  If locale is ahead of
* GMT, offset is positive.
*/
#include <time.h>


struct tm
{
  int tm_sec;                   /* Seconds.     [0-60] (1 leap second) */
  int tm_min;                   /* Minutes.     [0-59] */
  int tm_hour;                  /* Hours.       [0-23] */
  int tm_mday;                  /* Day.         [1-31] */
  int tm_mon;                   /* Month.       [0-11] */
  int tm_year;                  /* Year - 1900.  */
  int tm_wday;                  /* Day of week. [0-6] */
  int tm_yday;                  /* Days in year.[0-365] */
  int tm_isdst;                 /* DST.         [-1/0/1]*/

#ifdef  __USE_BSD
  long int tm_gmtoff;           /* Seconds east of UTC.  */
  __const char *tm_zone;        /* Timezone abbreviation.  */
#else
  long int __tm_gmtoff;         /* Seconds east of UTC.  */
  __const char *__tm_zone;      /* Timezone abbreviation.  */
#endif
};                   

extern struct tm *gmtime(const time_t*);
extern time_t time(time_t*);
extern struct tm *localtime(const time_t*);

int gmtoff(void)
{
  struct {
    struct tm tm;
    int sec;
  } local, gmt;
  int diff;
  time_t c;
  
  time(&c);
  memcpy(&(gmt.tm),gmtime(&c),sizeof(struct tm));
  /* gmt.tm = *(gmtime(&c)); */
  local.tm = *(localtime(&c));

  gmt.sec = gmt.tm.tm_hour * 3600 + gmt.tm.tm_min * 60 + gmt.tm.tm_sec;
  local.sec = local.tm.tm_hour * 3600 + local.tm.tm_min * 60 + local.tm.tm_sec;

  diff = local.sec - gmt.sec;
  if (diff > 86400) diff -= 86400;
  else if (diff < -86400) diff += 86400;
  return diff; 
}
