# 
# Creates GUI, parses command line arguments, sets default parameters
# for the controller
#

# $Id: config.tcl,v 1.37 1998/10/06 08:54:29 dor Exp $
# -------------------------------------------------------------------
#   The site command ist used to retrieve or set entries in the
#   member hash table
#   
#   site ssrc entry [value]
#   site ssrc delete
#
#   The entries are
#      cname          SDES information
#      name           SDES information
#      start          started
#      graphs         active graphs (0/1)
#      last_heard     time of last RTCP packet of PMM message (sender)
#      start          creating time of graph
#      num_rpts       number of RRs received
#      id1            id of first timeout
#      id2            id of second (final) timeout
#      loss           loss rate
#      filtered_loss  filtered loss rate
#      jitter         packet delay jitter
#
# Global variables:
#   
#   myself                  sender cname w/o dots 
#   allowed_bandwidth
#   actual_bandwidth
#   actual_frame_rate
#   allowed_frame_rate
#   address                 IP-address/Port of session
#   control                 control application (0/1)
#   min_bandwidth           minimum bandwidth
#   max_bandwidth           maximum bandwidth
#   alpha                   parameter for the linear filter
#   additive_increase       see description of adjustment algorithm
#   loss_lower_threshold 
#   loss_upper_threshold 
#   congested_recv_fraction 
#   loaded_recv_fraction 
#   algorithm               multicast adjustment algorithm 
#                           0: adjustment according to the worst positioned
#                              receiver
#                           1: adjustment based on the fraction of congested
#                           receiver
#ifloor variables 
set Ifloor(audio) 0 
set Ifloor(video) 0


#build the configuration window for all sessions
proc config_media {{s ""} {type ""} {w_child ""}} {
 global config_modified config_modified_levels  session config_modified_rsvp conf rsvp Levels
        # configuration menu
    set w .config_media
    if [winfo exists $w] {
	destroy $w
	return
    }
    if {![winfo exists $w ]} { 
	toplevel $w
	set x [expr [winfo screenwidth .] /2 ]
	set y [expr [winfo screenheight .] /2 ]
	wm geometry $w +$x+$y
	wm title $w  "$conf(name): Session Configuration"
	tixNoteBook $w.all
	pack $w.all -side top
    }
    set lw [string length "User e-mail (user@host)"]

    set config_procs [info procs config_*]
    if { $s !=""} {
	set type $session($s,type)
	set ses "$session($s,host):$session($s,port)"
	regsub -all {\.} $s {_} w_child
	regsub -all {/} $w_child {_} w_child
	$w.all add $w_child -label $s -underline 0
	set frame [$w.all subwidget $w_child]
	config_$type $s $frame
    }

    for { set i 0} {$i <[llength $conf(sessions)]} {incr i} {
	set Label [lindex $conf(sessions) $i]
	if {![string match *reservation* $Label]} {
	    if { $Label != $s } {
		set type $session($Label,type)
		set ses "$session($Label,host):$session($Label,port)"
		set k 1
		if { $session($Label,type) == "reservation" } {
		    set k 0
		    for { set l 0} {$l <[llength $conf(sessions)]} {incr l} {
			set k [expr [regexp $ses [lindex $conf(sessions) $l]] + $k]
		    }
		}
		if { $k ==1 } {
		    regsub -all {\.} $Label {_} w_child
		    regsub -all {/} $w_child {_} w_child
		    $w.all add $w_child -label $Label -underline 0
		    set frame [$w.all subwidget $w_child]
		    config_$type $Label $frame
		}
	    }
	}
    }
    #config_buttons $w "" config_accept
    config_cancel $w
    #update idletasks
    trace variable Levels w  config_trace_levels
    set config_modified ""
    set config_modified_levels ""
    set config_modified_rsvp ""

}


proc control_ac {{w .ac} {s ""} } {
    global conf Session

    if {[winfo exists $w]} {return}
    toplevel $w
    set x [expr [winfo screenwidth .] /2 ]
    set y [expr [winfo screenheight .] /2 ]
    wm geometry $w +$x+$y
    wm title $w "Application Control"
    set lw [string length "Filter Parameter:    "] 
    #config_read $s

    set session($s,control) 0
    pack \
	[hscale $w.alpha "Filter Parameter: "         $lw 0  1 0.1  "" \
	     session($s,alpha) 100 \
	    "config_parameters $s session($s,alpha)"] \
	[hscale $w.lower "Lower Threshold: "          $lw 0   20 0.1  % \
	     session($s,loss_lower_threshold) 100\
	    "config_parameters $s session($s,loss_lower_threshold)"] \
	[hscale $w.upper "Upper Threshold: "          $lw 0   20 0.1  % \
	     session($s,loss_upper_threshold) 100 \
	    "config_parameters $s session($s,loss_upper_threshold)"] \
	[hscale $w.md "Multiplicative Decrease: "     $lw 0  100 0.1  % \
	     session($s,multiplicative_decrease) 100\
	    "config_parameters $s session($s,multiplicative_decrease)"] \
	[hscale $w.ai  "Additive Increase "           $lw 0   250 1   kb/s \
	     session($s,additive_increase) 100\
	    "config_parameters $s session($s,additive_increase)"] \
	[hscale $w.congested "Congested Receiver: "   $lw 0   20  0.1 % \
	     session($s,congested_recv_fraction) 100\
	     "config_parameters $s session($s,congested_recv_fraction)"]\
	[hscale $w.loaded "Loaded Receiver: "         $lw 0   20  0.1 % \
	     session($s,loaded_recv_fraction) 100 \
	    "config_parameters $s session($s,loaded_recv_fraction)"  ] \
	[radiobutton $w.0 -text "Worst receiver decision" -anchor w \
	     -relief flat -variable session($s,algorithm) -value 0 \
	    -command "set session($s,control) 1; \
                 config_parameters $s session($s,control);\
                 config_parameters $s session($s,algorithm);\
                 set session($s,alpha) 0.3\;
                 config_parameters $s session($s,alpha);\
                 set session($s,loss_lower_threshold) 2\;
                 config_parameters $s session($s,loss_lower_threshold);\
                 set session($s,loss_upper_threshold) 4;\
                 config_parameters $s session($s,loss_upper_threshold);\
                 hscale_label $w.md {Multiplicative decrease} \;
                 hscale_unit $w.md {%};\
                 hscale_active $w.ai 1;\
                 hscale_active $w.congested 0;\
                 hscale_active $w.loaded 0"]\
        [radiobutton $w.1 -text "Congestion ratio decision" -anchor w \
	     -relief flat -variable session($s,algorithm) -value 1\
	    -command "set session($s,control) 1; \
                 config_parameters $s session($s,control);\
                 config_parameters $s session($s,algorithm);\
                 set session($s,loss_lower_threshold) 2\;
                 config_parameters $s session($s,loss_lower_threshold);\
                 set session($s,loss_upper_threshold) 4;\
                 config_parameters $s session($s,loss_upper_threshold);\
                 set session($s,alpha) 0.3\;
                 config_parameters $s session($s,alpha);\
                 hscale_label $w.md {Multiplicative decrease} \;
                 hscale_unit $w.md {%};\
                 hscale_active $w.ai 1;\
                 hscale_active $w.congested 1;\
                 hscale_active $w.loaded 1"]\
        [radiobutton $w.2 -text "Direct Adjustment decision" -anchor w \
	     -relief flat -variable session($s,algorithm) -value 2\
	     -command "set session($s,control) 1; \
                 config_parameters $s session($s,control);\
                 config_parameters $s session($s,algorithm);\
                 set session($s,alpha) 0.3\;
                 config_parameters $s session($s,alpha);\
                 set session($s,loss_lower_threshold) 2\;
                 config_parameters $s session($s,loss_lower_threshold);\
                 set session($s,loss_upper_threshold) 4;\
                 config_parameters $s session($s,loss_upper_threshold);\
                 hscale_label $w.md {Multiplicative decrease} \;
                 hscale_unit $w.md {%};\
                 hscale_active $w.ai 1;\
                 hscale_active $w.congested 0;\
                 hscale_active $w.loaded 0"]\
        [radiobutton $w.3 -text "Loss Delay Based Adjustment" -anchor w \
	     -relief flat -variable session($s,algorithm) -value 3\
	     -command "set session($s,control) 1; \
                 config_parameters $s session($s,control);\
                 config_parameters $s session($s,algorithm);\
                 set session($s,additive_increase)  10;\
                 config_parameters $s session($s,additive_increase);\
                 set session($s,alpha) 1\;
                 config_parameters $s session($s,alpha);\
                 set session($s,loss_lower_threshold) 1\;
                 config_parameters $s session($s,loss_lower_threshold);\
                 set session($s,loss_upper_threshold) 10;\
                 config_parameters $s session($s,loss_upper_threshold);\
                 hscale_label $w.md {Reduction factor} \;
                 hscale_unit $w.md ;\
                 set session($s,multiplicative_decrease) 3;\
                 config_parameters $s session($s,multiplicative_decrease);\
                 hscale_active $w.congested 0;\
                 hscale_active $w.loaded 0"]\
        [radiobutton $w.4 -text "No Adapatation" -anchor w \
	     -relief flat -variable session($s,algorithm) -value -1\
	     -command "set session($s,control) 0; \
                 config_parameters $s session($s,control)"]\
	    -anchor w
    tab_init [pack slaves $w]
    config_cancel $w
    
}





##### end application controller


#
# Copy current settings from session(s,) to Session().
#
proc config_read {s} {
  global Session session
  if {$s != ""} {
    foreach e [array names Session] {
	if {![string match "*reservation/*" $e]} {
	    set Session($e) $session($s,$e)
	}
    }
  }
}

#
# Create buttons for accepting/dismissing parameter changes.
#
proc config_buttons {w s command} {
    global config_modified config_modified_levels Session Levels session config_modified_rsvp

    if [winfo exists $w.action] return
    pack \
	[button $w.cancel -text "Cancel" -relief raised -borderwidth 1\
	     -command "set config_modified {}; destroy $w;"]\
	[button $w.accept -text "Accept"  -relief raised -borderwidth 1\
	     -command "$command $s;destroy $w;" ]\
	[button $w.apply -text "Apply"  -relief raised -borderwidth 1\
	     -command "$command $s;set config_modified {}" ]\
	-fill both -expand yes -anchor nw -side left

# For "Apply" added in the command line: set config_modified {}; !!!!!!!! 

    update idletasks
}

#
# the cancel button to be used for all windows
#
proc config_cancel {w } {
    pack \
	[button $w.cancel -text "Close" -relief raised -command "destroy $w"]\
	-fill both -expand yes -anchor nw -side left
}	



#
# General configuration (media-independent) for session 's'.
#
proc config_general {{s ""} {w .general}} {
  global conf Session session

  if [winfo exists $w] return
  toplevel $w
    set x [expr [winfo screenwidth .] /2 ]
    set y [expr [winfo screenheight .] /2 ]
    wm geometry $w +$x+$y
  wm title $w "$conf(name): Configure Session $s"
  set lw [string length "User e-mail (user@host)"] 
  #config_read $s

  # options buttons
  frame $w.mode
  pack \
    [checkbutton $w.mode.mute_new -text "mute new" -relief flat \
    -variable Session(mute_new)] \
    -side top -anchor w

  # net mutes mike vs. mike mutes net
  frame $w.priority
  pack \
    [label $w.priority.label -text "Muting:" -width $lw -anchor nw] \
    [radiobutton $w.priority.net    -text "net mutes local" -value net \
     -variable Session(priority) -relief flat] \
    [radiobutton $w.priority.local  -text "local mutes net" -value local \
     -variable Session(priority) -relief flat] \
    [radiobutton $w.priority.duplex -text "full duplex"    -value duplex \
     -variable Session(priority) -relief flat] \
    -side left -anchor w

  frame $w.encrypt
  pack \
    [field_str $w.encrypt.key     "Encryption key" $lw 10 "" Session(key)] \
    [checkbutton $w.encrypt.enable -relief flat -variable Session(encrypt)] \
    -side left -anchor w

  pack \
    $w.encrypt \
    $w.priority \
    $w.mode \
    [field_num $w.interval   "Statistics interval"     $lw 2  "sec" \
      conf(stats_period)] \
    -side top -anchor nw

  tab_init [pack slaves $w]
  config_buttons $w $s config_accept
}


#
# Transport-specific settings.
#
proc config_transport {w_protocol w_media s} {
  global Session session

  if {$session($s,transport) == "VAT"} {
      enable_win $w_protocol
      $w_protocol config  -step 1 
      $w_protocol config -command "config_parameters $s session($s,confid)"
      $w_protocol subwidget entry configure -state normal
      $w_protocol subwidget incr configure -state normal
      $w_protocol subwidget decr configure -state normal

    } else {
	$w_protocol config -step 0 
	$w_protocol config -command ""
	$w_protocol subwidget entry configure -state disabled
	$w_protocol subwidget incr configure -state disabled
	$w_protocol subwidget decr configure -state disabled

   }
    # and do not forget the config parameters 
   config_parameters $s session($s,transport)
}

proc config_limit {w s} {
  global Session limit session

  if {$session($s,media) == "YUV"} {
	set limit 100000
  } else {
        set limit 3000
  }
  config_parameters $s session($s,media)
  $w.s configure -to $limit
}




# convert the coding into a longer name
# this is required as nevot needs these values in the longer name form
proc code_2_name { code } {
    set x $code
    if { $code == "DVI4 1 8000" } {set x }
    if { $code == "GSM 1 8000"  } {set x 13 }
    if { $code ==  "LPC 1 8000" } {set x 6}
    if { $code ==  "PCMU 1 8000"}  {set x 64}
    if { $code == "DVI4 1 16000" } {set x 64}
    if { $code == "L16 1 44100" } {set x 705 }
    if { $code == "L16 2 44100" } {set x 1400}
    
    if { $code == "dvi4" } {set x "DVI4 1 8000"}
    if { $code == "gsm"  } {set x "GSM 1 8000" }
    if { $code ==  "lpc" } {set x "LPC 1 8000"}
    if { $code ==  "pcmu"}  {set x "PCMU 1 8000"}
    if { $code == "l16" } {set x  "L16 1 44100"}
    if { $code == "l16_1" } {set x  "L16 1 44100"}
    if { $code == "l16_2" } {set x  "L16 2 44100"}
    if { $code == "h261" } {set x  "H.261"}
    if { $code == "jpeg" } {set x  "JPEG"}
    if { $code == "cellb" } {set x  "CellB"}
    if { $code == "mpeg" } {set x  "MPEG"}
    return $x
}
# convert the coding name used in isc to the longer form
proc name_2_code { name } {
    set x $name
    if { $name == "DVI4 1 8000" } {set x "dvi4"}
    if { $name == "GSM 1 8000"  } {set x "gsm" }
    if { $name ==  "LPC 1 8000" } {set x "lpc"}
    if { $name ==  "PCMU 1 8000"}  {set x "pcmu"}
    if { $name == "L16 1 44100" } {set x  "l16"}
    if { $name == "L16 1 44100" } {set x  "l16_1"}
    if { $name == "L16 2 44100" } {set x  "l16_2"}
    if { $name == "H.261" } {set x  "h261"}
    if { $name == "JPEG" } {set x  "jpeg"}
    if { $name == "CellB" } {set x  "cellb"}
    if { $name == "MPEG" } {set x  "mpeg"}
    return $x
}


#
# Audio-specific settings.
#
proc config_audio {{s ""} {w .audio}} {
    global conf session Session Ifloor
    
    set lw [string length "Time constant - var."] 
    #config_read $s
    
    # convert the coding into a bandwdith value
    proc code_2_bw { code } {
	set x 0
	if { $code == "DVI4 1 8000" } {set x 32}
	if { $code == "GSM 1 8000"  } {set x 13 }
	if { $code ==  "LPC 1 8000" } {set x 6}
	if { $code ==  "PCMU 1 8000"}  {set x 64}
	if { $code == "DVI4 1 16000" } {set x 64}
	if { $code == "L16 1 44100" } {set x 705 }
	if { $code == "L16 2 44100" } {set x 1400}

	if { $code == "dvi4" } {set x 32}
	if { $code == "gsm"  } {set x 13 }
	if { $code ==  "lpc" } {set x 6}
	if { $code ==  "pcmu"}  {set x 64}
	if { $code == "l16" } {set x 705 }
	return $x
    }

    # set the appropriate reservation values
    update_rsvp $s [code_2_bw $session($s,media)]
    # frame for the audio format
    set w1 [frame $w.stream -relief groove  -borderwidth 1]
    tixLabelFrame $w1.media -label "Audio format" 
    set f_media [$w1.media subwidget frame]
    
    global mailcap_rules mailcap_subtypes

    for {set i 0} {$i < [llength $mailcap_subtypes(audio)]} {incr i} {
	set type [lindex $mailcap_subtypes(audio) $i]
	if { [winfo exists $f_media.$type] } {
	    isc_message "WARNING ~/.mailcap contains repeated entries"
	} else {
	    grid \
		[radiobutton $f_media.$type  -variable session($s,media) \
		     -text "[code_2_name $type] ([code_2_bw $type] kb/s)" \
		     -command "set  session($s,media) {[code_2_name $type]};\
		 config_parameters $s session($s,media); \
                update_rsvp $s [code_2_bw $type]" \
		     -value [code_2_name $type]] -sticky w -row $i -column 0
	}
    }
    # frame for the audio protocol
    tixLabelFrame $w1.protocol -label "Audio Protocol" 
    set f_protocol [$w1.protocol subwidget frame]
    pack \
	[radiobutton $f_protocol.rtp -variable session($s,transport) \
	     -value RTP -text RTP \
	    -command "config_parameters $s session($s,transport)"]\
	[radiobutton $f_protocol.vat -variable session($s,transport) \
	     -value VAT -text VAT \
	    -command "config_parameters $s session($s,transport)"]\
	[tixControl $f_protocol.confid -label "VAT Conf. ID"   \
	     -max 10000 -min 0 -step 1 -variable session($s,confid) \
	     -command "config_parameters $s session($s,confid)" ] \
	-side top -anchor nw  -expand yes -fill both
 		

    pack $w1.media $w1.protocol -side left -anchor nw  -expand yes -fill both

    $f_protocol.vat configure \
	-command "config_transport $f_protocol.confid $f_media $s"
    $f_protocol.rtp configure \
	-command "config_transport $f_protocol.confid $f_media $s"
    config_transport $f_protocol.confid $f_media $s   
    # frame for the QoS stuff and utilities
    set w2 [frame $w.stuff -relief groove  -borderwidth 1]
    tixLabelFrame $w2.qos -label "QoS Control" 
    set f_qos [$w2.qos subwidget frame]
    set f_delay [frame $f_qos.delay]
    pack \
	[checkbutton $f_delay.late_ts -text "Late = Talkspurt"\
	     -relief flat -variable session($s,late_ts)\
	    -command "config_parameters $s session($s,late_ts)"] \
	[checkbutton $f_delay.immediate -text "No playout delay" \
	     -relief flat -variable session($s,immediate)\
	     -command "config_parameters $s session($s,immediate)"] \
    -anchor nw -side left
    
    
    pack \
	$f_delay \
	[tixControl $f_qos.bpm -label "Audio per packet"  \
	     -max 0.20 -min 0.020 -step 0.010 -variable session($s,frame) \
	     -command "config_parameters $s session($s,frame)" ] \
	[hscale $f_qos.delay_smooth  "Time constant - var."    $lw 1 30 1 pkt \
	     session($s,delay_smooth) 150\
	    "config_parameters $s session($s,delay_smooth)"] \
	[hscale $f_qos.initial_delay "Initial delay"      $lw 0.01 1 0.01 sec \
	     session($s,initial_delay) 150\
	    "config_parameters $s session($s,initial_delay)"] \
	[hscale $f_qos.min_delay     "Minimum delay"      $lw 0.0 5 0.01 sec \
	     session($s,min_delay) 150\
	    "config_parameters $s session($s,min_delay)"] \
	[hscale $f_qos.max_delay     "Maximum delay"      $lw 0.0 5 0.01 sec \
	     session($s,max_delay) 150\
	    "config_parameters $s session($s,max_delay)"] \
	[hscale $f_qos.delay_mult    "Variance multiplier"     $lw 1 10 1 "" \
	     session($s,delay_multiplier) 150\
	    "config_parameters $s session($s,delay_multiplier)"] \
    -anchor nw

    #floor control
    set sess "$conf(id)/$session($s,type)/$session($s,host):$session($s,port)"
 
    #frame for the utilities
    tixLabelFrame $w2.util -label "Utilities" 
    set f_util [$w2.util subwidget frame]

    # frame for the sound directory
    tixLabelFrame $w2.sound -label "Sound Directory" 
    set f_sound [$w2.sound subwidget frame]
    pack \
	[entry  $f_sound.play_dir -textvariable sound(dir)]\
	-expand 1   -anchor nw
	
    pack \
	 $w2.sound \
	[button $f_util.play -text "Play" -relief raised -command play_panel] \
	[button $f_util.record -text "Record"  -relief raised \
	     -command record_panel] \
	[checkbutton $f_util.fc  -text "Floor Control" -relief raised  \
             -variable Ifloor(audio) \
             -command "i_session_ctrl Ifloor(audio) $sess"] \
   	[button $f_util.rsvp -text "Reservation"  -relief raised \
	     -command "do_reservation $s {[code_2_bw $session($s,media)]} "] \
	-expand 1 -fill both  -anchor nw

    
    pack $w2.qos $w2.util -side left -anchor nw  -expand yes -fill both
    pack $w1 $w2 -anchor nw  -expand yes -fill both
    
}


set limit 3000

#
# this part handles layered transmission. That is a video stream is divided
# into different layers and sent to seperate multicast addresses. 
# This is particularly beneficial for teh case of heterogenous environments
#
#

#
# check if the sum of the bandwidth of all the layers exceeds the maximum
# bandwidth the user wants to use
#
proc layer_bw {name index op} {
    global Levels Session
    set layer_bandwidth 0
    foreach e [array names Levels ?,bandwidth] { 
	set layer_bandwidth [expr $layer_bandwidth + $Levels($e)]
	if {$layer_bandwidth >$Session(max_bandwidth)} {
	  tkwarning "Sum of layers bandwidth exceeds maximum bandwidth allowed"
	}
    }
    
}   
#
# remove this level and all upper levels as well
#

proc layer_remove {win  win_level level {s ""}} {
    global conf
    if { $level >0 } {
	set b [tkwarning " Removing level [expr $level] and all upper \
                      levels" 2]
	if { $b==1 } {
	    set i [expr $level]
	    set b [$win_level index last]
	    while {$i< $b}  {
		$win_level delete $b
		set cs $conf(id)/$s
		control_send "$cs Levels [expr $level] remove"
		set b [expr $b -1]
	    }
	    destroy $win
	}
	    return
    }
    tkwarning "Can not remove the base layer, to leave the session \
                 please use the Leave button"
}

#
# make sure a new layer is added
#
proc layer_add {level {s ""}} {
    global Levels conf
    if { $level >0 } {
	set cs $conf(id)/$s
	foreach e [array names Levels [expr $level],*] {
	    lappend param_list [lindex [split $e ","] 1] $Levels($e)
	}
	control_send "$cs Levels [expr $level] $param_list "
    }
}
    
#
# build the window for configuring the layer
#
proc level_com {{level 0} win_level {s ""}} {
  global Levels Session
  set w .win_$level
  if {[winfo exists $w]} {return}
  set e [array names Levels $level,ttl]
  if {$e =="" } { 
      set Levels($level,port)  $session($s,port)
      set Levels($level,ttl)   $session($s,ttl)
      set Levels($level,local) $session($s,local)
      set Levels($level,host) [join [lreplace [split $session($s,host) .] 3 3 \
		      [expr [lindex [split $session($s,host) .] 3]+$level ]] .]
      set Levels($level,bandwidth) 0				 
      trace variable Levels($level,bandwidth) w layer_bw
  }
 
  set layer_limit $Session(max_bandwidth)
  set lw 15
  toplevel $w
    set x [expr [winfo screenwidth .] /2 ]
    set y [expr [winfo screenheight .] /2 ]
    wm geometry $w +$x+$y
  wm title $w "Layering Window: Level $level"

  frame $w.f -relief groove -borderwidth 1
  frame $w.f.network -relief groove -borderwidth 1
  pack \
	[field_str $w.f.network.host "Host(s)" $lw 20 "" Levels($level,host)] \
	[field_str $w.f.network.port "Port"    $lw 20 "" Levels($level,port)] \
	[field_num $w.f.network.ttl  "TTL"     $lw 6  "" Levels($level,ttl)] \
	[field_str $w.f.network.local "Local"  $lw 20 "" Levels($level,local)]\
  -side top -anchor w  
  frame $w.f.work 
  button $w.f.work.remove -text "Remove"  -relief raised\
      -command "layer_remove $w $win_level [expr $level] $s"
  button $w.f.work.add -text "Add"  -relief raised\
      -command "layer_add [expr $level] $s"
  pack  $w.f.work.remove $w.f.work.add -side top -anchor nw
  pack $w.f.network $w.f.work -side left 
  pack \
    $w.f \
    [hscale $w.bandwidth "Bandwidth" $lw 1 $layer_limit 1\
	 kbit/s Levels($level,bandwidth) 100] \
  -side top -anchor w  
  
    return [config_buttons $w $s config_accept_levels]
}
#
# check if this is a new level
#

#
# add a new level for video transmission
#
proc level_new {win_level {s ""}} {

# the first index of win_level (0) is reserved for the new entry
# but we would still like to name the layers starting from 0
    set name [$win_level  index last]
    level_com $name $win_level $s
    $win_level add command -label "level $name" -command "level_com $name \
         $win_level"
}


# determine which video input port to use
proc get_input {s  input } {
    global session flag_new

    if { $flag_new == 1 } {
	return
    }
    switch $input {
	"Input 1" {
	    set session($s,video_input) 1
	}
	"Input 2" {
	    set session($s,video_input) 2
	}
	"S Video" {
	    set session($s,video_input) 0
	}
	"Television" {
	    set session($s,video_input) 3
	}
	"Composite 1" {
	    set session($s,video_input) 4
	}
	"SVHS" {
	    set session($s,video_input) 5
	}
    }
    config_parameters $s session($s,video_input)
    return $session($s,video_input)
}

# determine the name of the video input port given its numerical representation
proc get_input_name { input } {
    global session 
    switch $input {
	"1" {
	   return "Input 1"
	}
	"2" {
	    return "Input 2"
	}
	0 {
	   return "S Video"
	}
	"3" {
	   return "Television"
	}
	"4" {
	   return "Composite 1"
	}
	"5" {
	   return "SVHS"
	}
    }
}


# determine which video device to use
proc get_device {s  device } {
    global session flag_new
    if { $flag_new ==1 } {
	return 
    }
    switch $device {
	"SunVideo" {
	    set session($s,video_device) sunvideo
	}
	"Parallax" {
	    set session($s,video_device) parallax
	}
	"Video4Linux" {
	    set session($s,video_device) video4linux
	}
    }
    config_parameters $s session($s,video_device)
    return $session($s,video_device)
}

# return the actual value of a variable
proc get_var { v } {
    global ${v}
    return ${v}
}


# whenever a the input and device lists are created the attached command is
# called leading to reinitialization of the parameters
# flag_new 0  -config_video has not been called yet
# flag_new 1  - config_video has been called so the session(,,.parameter has
#               a value
# flag_new 2  - the commands are called due to clicking the menues and not 
#               due to buildup of the neues

set flag_new 0
#
# Video-specific settings.
#
proc config_video {{s ""} {w .video} } {
    global conf Session  limit Levels session
    global flag_new mailcap_subtypes

    if { $flag_new == 2 } {
	set flag_new 1
    }
    if { $flag_new == 0 } {
	set flag_new 2
    }
    

    set lw [string length "Max. frame rate: "] 
    #config_read $s
    set session($s,transport) RTP
    
    # frame for the codecs and qos parameters
    set w1 [frame $w.stream -relief groove  -borderwidth 1]

    tixLabelFrame $w1.media -label "Video format" 
    set f_media [$w1.media subwidget frame]
    global mailcap_subtypes  
    mailcap_load
    for {set i 0} {$i < [llength $mailcap_subtypes(video)]} {incr i} {
	set mail_type [lindex $mailcap_subtypes(video) $i]
	regsub -all {\.} $mail_type {} type
	if { [winfo exists $f_media.$type] } {
	    isc_message "WARNING
~/.mailcap contains repeated entries"
	} else {
	    grid \
		[radiobutton $f_media.$type  -variable session($s,media) \
		     -text "[string toupper $mail_type] " \
		     -command "set session($s,media) [string toupper $type];\
		 config_parameters $s session($s,media)" \
		     -value [string toupper $type]] -sticky w -row $i -column 0
	}
    }
    set session($s,media) $session($s,media)
    regsub -all {\.} [lindex $mailcap_subtypes(video) 0] {} type
    set firstvideocoding $type

    # set the appropriate reservation values
    update_rsvp $s $session($s,max_bandwidth)
    tixLabelFrame $w1.qos -label "Quality control" 
    set f_qos [$w1.qos subwidget frame]
    pack \
    	[hscale $f_qos.max_bandwidth   "Max. bandwidth" $lw 1 $limit \
	     1 kbit/s    session($s,max_bandwidth) 100 \
	     "update_rsvp $s $[get_var session($s,max_bandwidth)];\
                  config_parameters $s session($s,max_bandwidth) "]\
	[hscale $f_qos.max_frame_rate "Max. frame rate"  $lw 1 30 1 f/s \
	     session($s,max_frame_rate) 100 \
	     "config_parameters $s session($s,max_frame_rate)"] \
	[hscale $f_qos.quality "Quality"  $lw 1 100 1 "" \
	     session($s,quality) 100 \
	     "config_parameters $s session($s,quality)"] \
    -side top -anchor nw -expand yes -fill both

    pack  $w1.media  $w1.qos -side left -anchor nw -expand yes -fill both

    # adjust the bandwidth scale for jpeg and YUV
    $f_media.$firstvideocoding configure \
	-command "config_limit $f_qos.max_bandwidth $s"
    #$f_media.yuv configure -command "config_limit $f_qos.max_bandwidth $s"

    # frame for the hardware setting, utilities and control buttons
    set w2 [frame $w.hard -relief groove  -borderwidth 1]

    tixLabelFrame $w2.hardware -label "Hardware" 
    set f_hard [$w2.hardware subwidget frame]
 
    pack \
	[tixOptionMenu $f_hard.input -variable input \
	     -width 15 -command "get_input $s " -options {
		 menubutton.width 10
	         menubutton.relief raised }] \
	[tixOptionMenu $f_hard.dev  -variable device \
	     -width 15 -command "get_device $s" -options {
		 menubutton.width 10
	         menubutton.relief raised }] \
	[tixOptionMenu $f_hard.opt \
	    -width 15 -options {
		 menubutton.width 10
	         menubutton.relief raised }] \
	-anchor nw -expand yes -fill both	

    # hardware option menu
    set menu [$f_hard.opt subwidget menu]
    $menu add checkbutton -label "Sending Slides" -variable session($s,sending_slides) \
	-command "config_parameters $s session($s,sending_slides)  "
    $menu add checkbutton -label "Use JPEG for H261"  -variable session($s,use_jpeg_for_h261) \
        -command "config_parameters $s session($s,use_jpeg_for_h261)  "
    set menubutton [$f_hard.opt subwidget menubutton]
    $menubutton configure -text "Options ..."
    
    # video input -create the names dynamically
    $f_hard.input add command [get_input_name [expr $session($s,video_input)%6]]
    $f_hard.input add command [get_input_name [expr [expr $session($s,video_input)+1] %6]]
    $f_hard.input add command [get_input_name [expr [expr $session($s,video_input)+2] %6]]
    $f_hard.input add command [get_input_name [expr [expr $session($s,video_input)+3] %6]]
    $f_hard.input add command [get_input_name [expr [expr $session($s,video_input)+4] %6]]
    $f_hard.input add command [get_input_name [expr [expr $session($s,video_input)+5] %6]]

    # video devices
    $f_hard.dev add command "Video4Linux"
    $f_hard.dev add command "SunVideo"
    $f_hard.dev add command "Parallax"


    #ifloor settings
    set sess "$conf(id)/$session($s,type)/$session($s,host):$session($s,port)" 

    # utilities frame -reservation, layerd coding, adaptation ...
    tixLabelFrame $w2.util -label "Utilities"
    set f_util [$w2.util subwidget frame]
    
    pack \
	[ button $f_util.ac -text "Application Control" -relief raised\
	      -command "control_ac $w.ac $s" ]\
	[checkbutton $f_util.fc  -text "Floor Control" -relief raised  \
             -variable Ifloor(video) \
             -command "i_session_ctrl Ifloor(video) $sess"] \
	[ button $f_util.res -text "Reservation" -relief raised\
	      -command "do_reservation $s $session($s,max_bandwidth)" ] \
 	-anchor nw -expand yes -fill both	
   
    

    # changed by stb
    #extras display options for vic-pmm
    
    set w3 [frame $w.display -relief groove  -borderwidth 1]
    tixLabelFrame $w3.dither -label "Dithering"
    set f_dither [$w3.dither subwidget frame]
    pack [radiobutton $f_dither.ordered -text "Ordered" \
	     -variable session($s,dither) -value ordered \
	     -command "config_parameters $s session($s,dither)" ]\
	-side left -anchor n -expand yes -fill y
    pack [radiobutton $f_dither.error_diff -text "Error Diff" \
	     -variable session($s,dither) -value error_diff \
	     -command "config_parameters $s session($s,dither)" ]\
	-side left -anchor n -expand yes -fill y
    pack [radiobutton $f_dither.quantize -text "Quantize" \
	     -variable session($s,dither) -value quantize \
	     -command "config_parameters $s session($s,dither)" ]\
	-side left -anchor n -expand yes -fill y
    pack [radiobutton $f_dither.gray -text "Gray" \
	     -variable session($s,dither) -value gray \
	     -command "config_parameters $s session($s,dither)" ]\
	-side left -anchor n -expand yes -fill y
    tixLabelFrame $w3.gamma -label "Gamma Correction"
    set f_gamma [$w3.gamma subwidget frame]
    pack \
    	[hscale $f_gamma.gamma "8-bit"  $lw 0 1 0.01 "" \
	     session($s,gamma) 100 \
	     "config_parameters $s session($s,gamma)"] \
    -side top -anchor nw -expand yes -fill both

    pack  $w3.dither $w3.gamma -side top -anchor w -expand yes -fill both

    #changed by mkr

    set w4 [frame $w.frame4 -relief groove  -borderwidth 1]
    set frame [frame $w4.save]

    #changed by mkr
    menubutton $frame.mb -text "Preferences" -menu $frame.mb.menu \
	-relief raised
    set m [menu $frame.mb.menu -tearoff 1]

    $m add radio -label Default -variable connect_env(type) \
	-command "load_config; set_config $s"
    $m add radio -label User -variable connect_env(type) \
	-command "load_config; set_config $s"
    $m add separator
    $m add radio -label LAN -variable connect_env(net) \
	-command "load_config; set_config $s"
    $m add radio -label WAN -variable connect_env(net) \
	-command "load_config; set_config $s"
  
    pack  $frame.mb \
	[button $frame.save -text "Save" \
             -command "save_config $s"] \
	-side left -expand yes
    #

    pack  $w2.util $w2.hardware -side left -anchor w -expand yes -fill both
    pack $w4.save -side bottom -fill x

    pack $w1 $w2 $w3 $w4 -side top -anchor w -expand yes -fill x


    set flag_new 2
}

# changed by mkr
proc config_profile {{s ""} {w ".configure.all.nbframe.connect_env"}} {
    global connect_env session conf Session

    if {$s == ""} {
	set Label [lindex $conf(sessions) 0]
	set s $Session(type)/$Session(host):$Session(port)
    }
	
    set w1 [frame $w.frame1 -relief groove  -borderwidth 1]

    tixLabelFrame $w1.type -label "Type"

    set frame [$w1.type subwidget frame]
    grid \
	[radiobutton $frame.env_Def -variable connect_env(type) \
	     -text "Default" -value Default] \
	-sticky w -row 0 -column 0
    grid \
	[radiobutton $frame.env_User -variable connect_env(type) \
	     -text "User" -value User] \
	-sticky w -row 1 -column 0

    tixLabelFrame $w1.net -label "Net"
    set frame [$w1.net subwidget frame]
    grid \
	[radiobutton $frame.env_lan -variable connect_env(net) \
	     -text "LAN" -value LAN] \
	-sticky w -row 0 -column 0
    grid \
	[radiobutton $frame.env_wan -variable connect_env(net) \
	     -text "WAN" -value WAN] \
	-sticky w -row 1 -column 0

    set w2 [frame $w.frame2 -relief groove  -borderwidth 1]
    set frame [frame $w2.save] 
    pack \
	[button $frame.but1 -text "Apply" \
	     -command "set_config $s"] \
	[button $frame.but2 -text "Save" \
	     -command "save_config $s"] \
	-side left -expand yes

    pack $w1.type $w1.net -side left -anchor w -expand yes -fill both
    pack $w2.save -side bottom -fill x

    pack $w1 $w2 -side top -anchor w -expand yes -fill both

}

#
# choose a file for transmission
#
proc config_file { w s} {
    proc send_file {s f } {
	global conf
	control_send "$conf(id)/$s file $f"
    }	
    # panel
    pack [fileselect $w.fs "send_file $s" " " \
	      "File" . *.gif Send] -side top
}
#
# Data-specific settings.
#
proc config_picture {{s ""} {w .picture}} {
    global conf Session  limit Levels session mailcap_subtypes

    set lw [string length "Max. frame rate: "] 
    #config_read $s
    set session($s,transport) RTP
    
    # frame for the codecs and qos parameters
    set w1 [frame $w.stream -relief groove  -borderwidth 1]

    tixLabelFrame $w1.media -label "Video file" 
    set f_media [$w1.media subwidget frame]
    config_file $f_media $s
    set session($s,media) $session($s,media)
    
    # set the appropriate reservation values
    update_rsvp $s $session($s,max_bandwidth)
    tixLabelFrame $w1.qos -label "Quality control" 
    set f_qos [$w1.qos subwidget frame]
    pack \
    	[hscale $f_qos.max_bandwidth  "Max. bandwidth" $lw 1 $limit \
	 1 kbit/s    session($s,max_bandwidth) 100 \
	     "update_rsvp $s $session($s,max_bandwidth);\
                 config_parameters $s session($s,max_bandwidth) "]\
	[hscale $f_qos.max_frame_rate "Max. frame rate"  $lw 1 30 1 f/s \
	     session($s,max_frame_rate) 100 \
	    "config_parameters $s session($s,max_frame_rate)"] \
	[hscale $f_qos.quality "Quality"  $lw 1 100 1 "" \
	     session($s,quality) 100\
	    "config_parameters $s session($s,quality)"	    ] \
    -side top -anchor nw -expand yes -fill both

    pack  $w1.media  $w1.qos -side left -anchor nw -expand yes -fill both

    # frame for the utilities and control buttons
    set w2 [frame $w.hard -relief groove  -borderwidth 1]

    # utilities frame -reservation, layerd coding, adaptation ...
    tixLabelFrame $w2.util -label "Utilities"
    set f_util [$w2.util subwidget frame]
    
    pack \
	[ button $f_util.ac -text "Application Control" -relief raised\
	      -command "control_ac $w.ac $s" ]\
	[ button $f_util.res -text "Reservation" -relief raised\
	      -command "do_reservation $s $session($s,max_bandwidth)" ] \
 	-anchor nw -expand yes -fill both

    # frame for the control buttons
    tixLabelFrame $w2.control -label "Control"
    set f_control [$w2.control subwidget frame]
    pack [checkbutton $f_control.transmit  -text "Transmit" \
	      -variable session($s,send) \
	     -command "config_parameters $s session($s,send)"]\
	[checkbutton $f_control.preview  -text "Preview" \
	     -variable session($s,preview) \
	     -command "config_parameters $s session($s,preview)"]\
	-side top -anchor w -expand yes -fill both
    
    pack  $w2.util $w2.control -side left -anchor w -expand yes -fill both

    
    pack $w1 $w2 -side top -anchor w -expand yes -fill both


}


# netghost-specific settings.
#
proc config_netghost {{s ""} {ww .netghost}} {
    global conf Session session netghost mailcap_subtypes
 
 
    #config_read $s
    set session($s,transport) RTP 
    set cs $conf(id)/$s
 
# mode sender / receiver
 
 
    tixLabelFrame $ww.location -label "File Location:"
    set f_d [$ww.location subwidget frame]

    pack [field_str $f_d.hostname  "http://" 5 20 "" session($s,netghost_host) ] \
	[field_str $f_d.path  ":" 1 20 "" session($s,netghost_path) ] \
	-side left -anchor n -expand no -fill both
    
    pack [button $f_d.getit -text "Load File" -relief raised -borderwidth 1 -command "getfile $s"]\
	-side right -anchor n -expand no -fill both

    
    tixLabelFrame $ww.source -label "Configuration:"
    set f_s [$ww.source subwidget frame]
    pack [radiobutton $f_s.local   \
              -variable session($s,netghost_protocol)  -value LOCAL\
	      -text "local file" -command "$f_d.hostname.in configure -state disabled -bg grey; \
               $f_d.hostname.in select clear; \
               focus $f_d.path.in; \
               set session($s,netghost_protocol) LOCAL; \
	       config_parameters $s session($s,netghost_protocol)" ]\
        [radiobutton $f_s.www   \
             -variable session($s,netghost_protocol)  -value HTTP -text "WWW"\
	     -command "$f_d.hostname.in configure -state normal -bg white;\
              focus $f_d.hostname.in ; \
              set session($s,netghost_protocol) HTTP;\
	      config_parameters $s session($s,netghost_protocol)" ]\
	[field_num $f_s.entry1 "  TTL"     6 4  "" session($s,ttl)]  \
	[field_num $f_s.entry2 "  ID"     5 6  "" session($s,netghost_id)]  \
	[radiobutton $f_s.sender   \
	     -variable session($s,mode)  -value sender -text "Sender" \
	     -command "set session($s,mode) sender;\
	      config_parameters $s session($s,mode)" ]\
	[radiobutton $f_s.receiver   \
	     -variable session($s,mode)  -value receiver -text "Receiver"\
	     -command "set session($s,mode) receiver;\
	      config_parameters $s session($s,mode)" ]\
        -side left -anchor n -expand yes -fill both
    
    iscBalloon $f_d.path.in "netghostpath"
    iscBalloon $f_d.hostname.in "netghosthost"
    iscBalloon $f_s.sender "netghostmode"
    iscBalloon $f_s.receiver "netghostmode"
    iscBalloon $f_s.local "netghostlocation"
    iscBalloon $f_s.www "netghostlocation"

    bind  $f_d.path.in <Return> "getfile $s"
    bind  $f_d.hostname.in <Return> "getfile $s"
    bind  $f_d.getit <Return> "getfile $s"
    bind $f_d.hostname.in <FocusIn> "$f_d.hostname.in select range 0 end"
    bind $f_d.path.in <FocusIn> "$f_d.path.in select range 0 end"

    focus $f_d.hostname.in ; 
    $f_d.hostname.in select range 0 end;



    pack  $ww.source $ww.location -side top -anchor w -expand no -fill both


    proc getfile {s} {
	global conf session netghost config_modified env
	



	# we know the excat location of the PS file -> local or in the WWW
	set _host $session($s,netghost_host);
	set _path $session($s,netghost_path);
	set _file [string range $_path [expr [string last "/" $_path] +1] [string length $_path]]
	set _prot $session($s,netghost_protocol);

	# if the file must be downloaded from the WWW, check the protocol variable
	if {$_prot=="HTTP"} {
	    if {$_host==""} { return 0 }
	    
	    set url "$_host/$_path"

	    # where to store, $HOME or /tmp
	    if {[info exists env(HOME)]} {
		set localpath "$env(HOME)"
	    } else {
		set localpath "/tmp"
	    }
	    
	    # create the complete path for the file, e.g. /tmp/xxx.ps
	    # and open the file
	    set file "$localpath/$_file"
	    set out [open $file w]
	    
	    proc progress {token total current} {
		puts -nonewline "."
	    }
	    
	    http_config
	    set token [http_get $url -channel $out]
	    close $out
	
	} else {
	    set file "$_path"
	}


	set name_end $file
	set found 0
	set found [eval regexp {.gz}  $name_end]
	if {$found==1} { 
	    set length [eval string length $name_end]
	    set length [expr ($length-4)]
	    set  name_end [string range  $name_end 0 $length]
	}
	set found 0
	set found [eval regexp {.Z}  $name_end]
	if {$found==1} { 
	    set length [eval string length  $name_end]
	    set length [expr ($length-3)]
	    set name_end [string range  $name_end 0 $length]
	}
	set session($s,netghost_file) $name_end
	
	set $session($s,netghost_protocol) LOCAL
	set $session($s,netghost_path) $file

	config_parameters $s session($s,netghost_host)
	config_parameters $s session($s,netghost_path)
	config_parameters $s session($s,netghost_file)
    }
    
    
}


 


#proc ftp_getfile { _protocol _host _path _name } {
#    global conf Session env
# 
#    if {$_name==""} {
#        return 0
#    }
#    if {$_protocol=="LOCAL"} { return 0 }
#    if {$_protocol == "HTTP"} { 
#
#    }
#}




# 
# Iloof specific settings
#

#get actual ifloor or media session(s)
proc get_session {name} {
global session conf i_sessions debug
    if [info exists i_sessions] { 
        unset i_sessions
    } 
    set i 0
    foreach e [array names session *,$name] {
	set s [lindex [split $e ","] 0]
	set t [lindex [split $s "/"] 0]
	if {$t == $name} {
	    set i_sessions($name,$i) $conf(id)/$s
	    if { $debug == "pmm" } {
		puts stdout "isc-->get_session:$i session == $i_sessions($name,$i)"
	    }
            incr i
	} 
    }
}

#get actual ifloor session
proc i_mediactl {i_agent} {
    global $i_agent i_sessions debug
    upvar #0 $i_agent v 
    #upvar fails here!! (sometimes ...)
    #send pmm msg to ifloor: addmedia/rmmedia a media agent
    set agentnam [lindex [split $i_agent "("] 1]
    set agentnam [string trimright $agentnam ")"]
    #   set v $agent
    if { $debug == "pmm" } {
	puts stdout "isc-->mediactl:agent $agentnam == $v"
    }
    get_session "ifloor"
    if {![info exists i_sessions]} {
	if { $debug == "pmm" } {
	    puts stdout "isc-->mediactl: no session(ifloor)"
	}
	return
    }
    foreach e [array names i_sessions ifloor,*] {
        set to $i_sessions($e)
        if {$v == 1} {
            control_send [list $to addmedia $agentnam]
            get_session $agentnam
            if [info exists i_sessions] {
		foreach e [array names i_sessions $agentnam,*] {
		    control_send [list $to s_control $i_sessions($e)]
		}
            }
        } else {
            control_send [list $to rmmedia $agentnam]
        }
    }  	
}


proc do_ifloor {s} {
    global Session conf session
    set Session(type) ifloor
    set Session(host) $session($s,host)
    set Session(port) 3452
    session_create
    
    set s "ifloor/$Session($host):$Session(port)"
    if { [lsearch -exact $conf(sessions) $s] < 0 } {
	footer "Could not Establish: $s"
	after $delay(display) "footer {}"
    } else {		    
	configure_ifloor $s
    }
}
proc i_session_ctrl_nocreate {i_media s} {
    global session conf $i_media i_sessions debug
    upvar #0 $i_media v
    # send pmm msg to ifloor: control/uncontrol a media session  
    if { $debug == "pmm" } {
	puts stdout "isc-->i_session_ctrl:agent = $i_media, sess = $s"
    }
    get_session "ifloor"
    if {![info exists i_sessions]} { 
	if { $debug == "pmm" } {
	    puts stdout "isc--> sessionctrl: no session(ifloor)"
	}
	return
    }
    foreach e [array names i_sessions ifloor,*] {
        set to $i_sessions($e)
        if {$v == 1 } {
	    if { $debug == "pmm" } {
		puts stdout "isc-->sessionctrl:$e : $to session_control $s" 
	    }
	    control_send [list $to s_control $s]
        } else {
	    if { $debug == "pmm" } {
		puts stdout "isc-->sessionctrl:$e : $to session_uncontrol $s"
	    } 
            control_send [list $to s_uncontrol $s]
        } 
    } 		
}


proc ifloor_close {s } {
    global Ifloor session conf
    if { [info exists Ifloor($session($s,type))] } {
	set Ifloor($session($s,type)) 0
    }
    if { [ info exists session($s,type) ] } {
	if { $session($s,type) =="ifloor" } {
	    foreach e $conf(sessions) {
		if { [info exists session($e,type)] } {
		    if { [info exists Ifloor($session($e,type))] } {
			set Ifloor($session($e,type)) 0
		    }
		}
	    }
	}
    }
}


proc i_session_ctrl {i_media media_s} {
    global session conf $i_media i_sessions delay Ifloor debug
    upvar #0 $i_media v
    # send pmm msg to ifloor: control/uncontrol a media session  
    if { $debug == "pmm" } {
	puts stdout "isc-->session_ctrl:agent = $i_media, sess = $media_s"
    }
    get_session ifloor
    # if floor control is switched ON:
    if {$v == 1} {
	# create ifloor session, if it does not exist
        if {![info exists i_sessions]} {
            set Ifloor(session) " "
            # invoke "initialize media" menue
#             if [catch {set audio_fd [open /dev/audio {NONBLOCK RDWR}]}] {
#                exec rm /var/tmp/pmm_audioctl.pid
#                exit
#             }
# 	    close $audio_fd
            choose_media 2
            choose_network_enabled $Ifloor(rowwin) $Ifloor(row) \
                $Ifloor(column) ifloor
	    
            set id [after 5000 "set Ifloor(session) no"]
            tkwait variable Ifloor(session)
            after cancel id
            if {[string match $Ifloor(session) "no"] } {
                #ifloor could NOT be created! (reported elsewhere)
                return
            }
            #ifloor has been created!
	    
            get_session ifloor
        }
    } else {
        #if floor control is switched OFF:
        #remove session from ifloor control-session list and return
	if { ![ info exists Ifloor(c_sessions) ] } {
	    set  $i_media 0
	    return
	}
        set i [lsearch -exact $Ifloor(c_sessions) $media_s]
        if {$i > 0} {
            set Ifloor(c_sessions) [lreplace  $Ifloor(c_sessions) $i $i]
        } else {
            if {$i == 0} {
                unset Ifloor(c_sessions)
            }
        }
        #if no ifloor session exists, nothing more is to do now for 
        #session UNcontrol
        if {![info exists i_sessions]} {
            return
        }
    }
    
    #all sessions to be floor controlled are in "c_session" list:
    set Ifloor(c_sessions) [lappend Ifloor(c_sessions) $media_s]
    if { $debug == "pmm" } {
	puts stdout "isc-->i_session_ctrl:$media_s to be controlled"
    }
    #send "control" messages to ifloor:              
    if {[info exists i_sessions]} {
        set to $i_sessions(ifloor,0) 
	
        if {$v == 1 } {
	    if { $debug == "pmm" } {
		puts stdout "isc-->sessioncntrl: session_control $media_s"
	    } 
	    control_send [list $to s_control $media_s]
        } else {
	    if { $debug == "pmm" } {
		puts stdout "isc-->sessioncntrl: session_uncontrol $media_s" 
	    }
	    control_send [list $to s_uncontrol $media_s]
        }
    }  		
}

#if ifloor crashes and starts again, the relavant config. variables 
#are still valid in isc. they must be explicitly asked for.

proc config_read_ifloor {{s ""} {w .ifloor}} {
    global Ifloor i_sessions conf debug
    if { $debug == "pmm" } {
	puts stdout "isc-->config_read_ifloor for session $s"
    }
    get_session "ifloor"
    
    if {[info exists Ifloor(c_sessions)]} {
        foreach e $Ifloor(c_sessions) {
	    if { $debug == "pmm" } {
		puts stdout "isc-->config_read_ifloor: under control $e"
	    }
            set s [lindex [split $e "/"] 1]
            if {[lsearch $conf(sessions) $e] >= 0} {
		if { $debug == "pmm" } {
		    puts stdout "isc-->config_read_ifloor: control $e"
		}
                control_send [list $to s_control $e]
            }
        }
    }
    
    if {[info exists Ifloor(audio)]} { 
	if {$Ifloor(audio) == 1} {
	    i_mediactl Ifloor(audio)
	}
    }
    
    if {[info exists Ifloor(video)]} {
	if {$Ifloor(video) == 1} {
	    i_mediactl Ifloor(video)
	} 
    } 
}





proc config_ifloor {{s ""} {w .ifloor}} {
    global conf session Ifloor debug
    if { $debug == "pmm" } {
	puts stdout "isc-->config_ifloor1: windowname:$w, s:$s"
    }
#default setting!!!

#control media
frame $w.head1 -relief flat -borderwidth 4
    pack [label $w.head1.msg -anchor center -text "Control Media" \
	      -font *-*-*-bold-i-*-12-*-*-*-*-*-*-*]
    pack $w.head1 -side top -fill x

#put buttons:
#buttons must be active, if buttons change to 0, <rmmedia> must called
frame $w.buttons -borderwidth 10
  set Vbut  [checkbutton $w.buttons.video -text Video -relief flat \
	   -variable Ifloor(video)  -command "i_mediactl Ifloor(video)" \
	       -bg LightGrey -activebackground SlateGrey]
  set Abut [checkbutton $w.buttons.audio -text Audio -relief flat \
	       -variable Ifloor(audio) -command "i_mediactl Ifloor(audio)" \
	      -bg LightGrey -activebackground SlateGrey]

  tab_init [pack slaves $w]
#  config_buttons $w $s
  pack $w.buttons.audio -side top -padx 20
  pack $w.buttons.video -side top -padx 20 
  pack $w.buttons -side top

#read settings, already valid in isc:
# config_read_ifloor $s
}




#
# SDES information.
#
proc config_sdes {{s ""} {frame .sdes}} {
  global conf

  set lw [string length "User e-mail (user@host)"]

  pack \
    [field_str $frame.name       "User real-life name"     $lw 40 "" \
      Session(name)] \
    [field_str $frame.email      "User e-mail (user@host)" $lw 40 "" \
      Session(email)] \
    [field_str $frame.img        "User image file name"    $lw 40 "" \
      Session(img)] \
    [field_str $frame.phone      "User phone number"       $lw 40 "" \
      Session(phone)] \
    [field_str $frame.loc        "User location"           $lw 40 "" \
      Session(loc)] \
    [field_str $frame.note       "Note to conference"      $lw 40 "" \
      Session(note)] \
    -side top
}



#
# Trace for dynamically changeable variables. (session)
# Strangely enough, upvar fails (but without error indication) 
#   if called through the backspace binding.
#
proc config_trace {name element op} {
  global Session config_modified session

  #
  # append to list of modified variables
  #
  if {$element != ""} {
    set name ${name}($element)
  }
  ladd config_modified $name
  #config_accept [lindex [split $element ,] 0]
}

#
# Trace for dynamically changeable variables.
# Strangely enough, upvar fails (but without error indication) 
#   if called through the backspace binding.
#
proc config_trace_levels {name element op} {
  global Levels config_modified_levels 

  #
  # append to list of modified variables
  #
  if {$element != ""} {
    set name ${name}($element)
  }
  ladd config_modified_levels $name
}

proc config_parameters {s name {value ""}} {
    global config_modified connect_env
    set connect_env(type) "User"
    lappend config_modified $name
    config_accept $s
    set config_modified ""
}


#
# Changes parameters for session 'cs'.  Invoked by accept button;
# checks for modified values.
#
proc config_accept {{s ""}} {
  global conf  session  config_modified  member rsvp

    if {$s != ""} {
	set cs $conf(id)/$s
	foreach c $config_modified {
	    set e  [lindex [split [lindex [split $c "()"] 1] ","] 1]
	    set e2  [lindex [split [lindex [split $c "()"] 1] ","] 0]

	    if { ![string compare $s $e2] } {
		if {[regexp -nocase source $e]} {
		    if { $rsvp($s,filtertype) == 1 } {
			set value $rsvp($s,$e)
			set source [string range $e 6 end]
			set e "reserve $value"
			regsub -all {_} $source {.} source 
			regsub -all {\/} $source {.} source
			set value $source
			control_send "$cs $e [list $value]"
		    }
		    if { $rsvp($s,filtertype) == 0 } {
			set value $rsvp($s,$e)
			set source [string range $e 6 end]
			set e "reserve $value"
			regsub -all {_} $source {.} source 
			regsub -all {\/} $source {.} source
			set value $source
			control_send "$cs $e [list $value]"
		    }

		} else {
		    if {[regexp -nocase flowspec $e]} {
			set source [string range $e 8 end]
			regsub -all _ $source . source 
			regsub -all \/ $source . source
			set e "flowspec $source"
		    } else {
			set value [set $c]
			control_send "$cs $e [list $value]"
			#make sure that the control buttons and the control icons
			#show the same state for the previewing and transmission
			if { [regexp -nocase send $e] } {
			    set m [lindex $conf(us) 0]	
			    set member($m,$s,rx) $session($s,send)
			}
			if { [regexp -nocase preview $e] } {
			    set m [lindex $conf(us) 0]
			    if { $session($s,preview) > 0} {
				set member($m,$s,state) A
			    } else {
				set member($m,$s,state) r
			    }		
			}    
		    }
		}
	    }
	}
    } else {
	foreach s $conf(sessions) {
	    config_accept $s
	}
    }
}
proc config_accept_levels {{s ""}} {

    global config_modified_levels conf Levels
    set cs $conf(id)/$s
    if { $config_modified_levels != ""} {
	set c [lindex $config_modified_levels 0]
	set param [lindex [split $c "()"] 0]
	set param_num\
	    [list [lindex [split [lindex [split $c "()"] 1] ","] 0]]
	foreach e [array names Levels $param_num,*] {
	    lappend param_list [lindex [split $e ","] 1] \
		[subst $${param}($e)]
	}
	control_send "$cs $param $param_num $param_list "
	set config_modified_levels "" 
    }
}


proc config_vote { {s ""} {w .vote} } {
    global conf session vote
    
    #config_read $s

    label $w.label -text "No parameters to configure!"

    pack $w.label -side left 
}
