#
# Invoked when graphing buttons are pressed.
# Draws for session 's', member 'm', parameter 'p', row 'y'.
#
# $Id: graph.tcl,v 1.3 1998/07/20 16:41:34 stb Exp $
#


#set Font  -*-helvetica-small-r-normal--*-100-* 

set Width 350 
set Height 300
set W_Edge 40
set H_Edge 40
set Curve_W [expr $Width  - (2*$W_Edge)]
set Curve_H [expr $Height - (2*$H_Edge)]

set X_Zero $W_Edge
set Y_Zero [expr $Curve_H+$H_Edge]


proc graph {w title x_axis} {
   global Curve_H Curve_W X_Zero Y_Zero members W_Edge H_Edge Width Height 

    if { [winfo exists $w]} {
	destroy $w
    }

    canvas $w -width $Width -height $Height
    $w create rectangle [expr $Curve_W+$W_Edge] [expr $Curve_H+$H_Edge]\
	$W_Edge $H_Edge  -outline black -fill white -tag rec_$w

    $w create text \
	[expr $Width/2] [expr $H_Edge-20]  -text $title -justify center 
    $w create text \
	[expr $Width - $W_Edge] [expr $Height-$H_Edge+30] -text $x_axis \
	-justify center 

    pack $w

    set members($w,div) 0
    set members($w,Max_x) 1
    set members($w,Max_y) 1
    set members($w,Div_x) 10
    set members($w,Div_y) 10
    set members($w,tag_div) tag_div_$w
}

proc element {w list_x list_y} {
    global W_Edge H_Edge Width Height members Curve_W Curve_H 

    set members($w,Div_x) 10
    set members($w,Div_y) 10

    set members($w,Max_x)   [lindex $list_x [expr [llength $list_x]-1 ]]

    set value_y [lindex $list_y [expr [llength $list_y]-1 ]]

    # find the maximum y value and make sure it is not 0
    # determine the number of y-axis devisions
    if { $value_y > $members($w,Max_y) } {
	set members($w,Max_y) $value_y
    }
    if { $members($w,Max_y) < $members($w,Div_y) } {
	set members($w,Div_y) $members($w,Max_y)
	if { $members($w,Div_y) <=1 } {
	    set members($w,Div_y) 2
	}
    }

    # find the maximum x value and make sure it is not 0 and 
    # determine the number of x-axis devisions
    if { $members($w,Max_x) < $members($w,Div_x) } {
	set members($w,Div_x) $members($w,Max_x)
	if { $members($w,Div_x) <=1 } { 
	    set members($w,Div_x) 2
	}
     }
    
    # delete the old tags
    if { $members($w,div) >0 } {
	$w delete tag_$w
    }

    # draw the x axis
    for { set i 0} {$i <= $members($w,Div_x) } {incr i} {
	$w create text \
	    [expr ($W_Edge )+(($i*$Curve_W)/($members($w,Div_x)))] \
	    [expr $Height -$H_Edge +10] \
	    -text [expr ($i*$members($w,Max_x))/$members($w,Div_x)] \
    -justify center -tag tag_$w 
	set members($w,div) $members($w,Div_x)
    }

    #draw the y axis
    for { set i 0} {$i <= $members($w,Div_y) } {incr i} {
	set value \
	    [format "%.1f" [expr ($i*$members($w,Max_y))/$members($w,Div_y)]]
	$w create text   [expr $W_Edge -20] \
	    [expr ($Height -$H_Edge)-(($i*$Curve_H)/$members($w,Div_y))] \
	    -text $value -justify right -tag tag_$w 

    }	
    draw $w $list_x $list_y
}


# draw the actual curve    
proc draw {w list_x list_y} {
    global List Curve_W Curve_H X_Zero Y_Zero  members 

    set length [expr [llength $list_x]-1 ]
    for { set i 0} {$i <= $length} {incr i} {
	# scale the single points with the actual size of the curve
	set List [lappend List \
		 [expr ([lindex $list_x $i]*$Curve_W)/($members($w,Max_x)+1)]]
	set List [lappend List \
		[expr ([lindex $list_y $i]*$Curve_H)/($members($w,Max_y)+1)]]
    }
    #clean up the last curve 
    $w delete -tag {$w_rec_tag}

    set x1 $X_Zero
    set y1 $Y_Zero
    #draw a new curve
    set length [expr [llength $List]/2]
    for { set i 0} {$i < $length} {incr i} {
	set x2 [expr [lindex $List [expr ($i*2)]]+$X_Zero]
	set y2 [expr $Y_Zero - [lindex $List [expr ($i*2)+1]]]
	$w create line $x1 $y1  $x2 $y2 \
	    -fill green -tag {$w_rec_tag}
	set x1 $x2
	set y1 $y2
    }
    if [info exists List] {unset List}
}

#
# destroy the statistics graph
#
proc stats_destroy {w s m p} {
    global member 
    global xdata ydata
    
    incr member($m,$s,graphs) -1
    if [info exists xdata($m,$s,$p)] {unset xdata($m,$s,$p)}
    if [info exists ydata($m,$s,$p)] {unset ydata($m,$s,$p)}
    set member($m,$s,G,$p) 0
    if {[winfo exists $w]} {destroy $w} 
}

proc stats_graph {s m p map y} {
  global member 
  global xdata ydata

    set w .[noperiod $s]/[noperiod $m]/[noperiod $p]
    if {$member($m,$s,G,$p)} {
	incr member($m,$s,graphs)
	# create graphing widget and print button if they don't already exist
	if {[winfo exists $w.g] == 0} {
	    toplevel $w
	    wm title $w $s/$m
	    set member($m,$s,start_$p) [getclock]

	    graph $w.g $p Seconds
	    grid configure $w.g -row $y -column 0 -columnspan 4
	    if {[winfo exists $w.print] == 0} {
		grid configure \
		    [button $w.print -text "Save to File" -width 23 \
			 -command "print_draw $s $m $p"] \
		    [button $w.destroy -text "Destroy" -width 23\
			 -command "stats_destroy $w $s $m $p "]\
		    -sticky ns
	    }
	    
	} 
    } else {
	stats_destroy $w $s $m $p
    }
}

# print values of graph for member 'm' in session 's'
#
proc print_draw {s m p} {
    global member 
    global xdata ydata

    if { [info exists xdata($m,$s,$p)] == 0 } {
	return;
    }
    set pid_id [pid]
    set type [lindex [split $s /] 0]
    if { [file exists $member($m,cname).$s.$pid_id.$p]} {
	exec rm $member($m,cname).$type.$pid_id.$p
    }
    if [ catch {open $member($m,cname).$type.$pid_id.$p w} id] {
	puts stderr " can not open file $member($m,cname).$type.$pid_id.$p"
    } else {
	for {set i 0} {$i < [llength $xdata($m,$s,$p)]} {incr i} {
	    puts $id "[lindex [split $xdata($m,$s,$p)] $i]\
                      [lindex [split $ydata($m,$s,$p)] $i]"
	 
	}
	close $id
    }
}

#
# add points to existing graph for member 'm' in session 's'
#
proc stats_draw {s m p v} {
    global member 
    global xdata ydata
    set w .[noperiod $s]/[noperiod $m]/[noperiod $p]
    lappend xdata($m,$s,$p) [expr [getclock] - $member($m,$s,start_$p)] 
    lappend ydata($m,$s,$p) $v
    element $w.g $xdata($m,$s,$p) $ydata($m,$s,$p)
}

#
# destroy graph if dismiss button in stats window is pressed
# 
proc destroy_graph {s m} {
    global xdata ydata

    foreach p {jitter rr_jitter actual_framerate framerate actual_bandwidth \
		   bandwidth delay lost rr_lost late dup order} {
	set w .[noperiod $s]/[noperiod $m]/[noperiod $p]
	if {[winfo exists $w]} {destroy $w} 
	if [info exists xdata($m,$s,$p)] {unset xdata($m,$s,$p)}
	if [info exists ydata($m,$s,$p)] {unset ydata($m,$s,$p)}
    }
}


