#
# invite.tcl
# procedures to invite other users to sessions
# - read session parameters of existing sessions
# - create and send sip requests
# - timeout handler
#


#
# check, if there are already active sessions
# get parameters of existing sessions
#
proc get_active_session_parameters {} {
  global invite conf session

  catch {unset invite(media_list)}
  if [llength $conf(sessions)] {
    # create media list of existing sesions
    set invite(media_list) ""
    foreach s $conf(sessions) {
      if {$session($s,type) == "audio" || $session($s,type) == "video"} {
        set ml $session($s,type)
	lappend ml $session($s,port)
        # handle unicast conference addr
        set a [lindex [split $session($s,host) .] 0]
        if {$a < 224 || $a > 255} {
	    set invite(unicast) 1
   	  lappend ml [idstring %d]
	} else {
	  lappend ml $session($s,host)
	}
	lappend ml $session($s,ttl)
	lappend ml [string tolower [lindex $session($s,media) 0]]
	lappend ml ""
	lappend invite(media_list) $ml
      }
    }
  }
  invite_proc
} ;# get_active_session_parameters



proc invite_proc {} {
    global invite image conf m_rules

    set w .invite
    if [winfo exists $w] return 
    toplevel $w
    set title "Invitation Parameters"
    wm title $w $title
    #put the window in a fixed place
    set x [expr [winfo screenwidth .] /2 ]
    set y [expr [winfo screenheight .] /2 ]
    wm geometry $w +$x+$y
    
    frame $w.top -relief groove  -borderwidth 1
    frame $w.top.left
    pack \
    [field_str $w.top.left.guest        "Callee(s)"   20 30 "" invite(guest)] \
    [field_str $w.top.left.session_name "Conference Title"   20 30 "" conf(name)] \
	[field_str $w.top.left.information  "Conference Subject" 20 30 "" invite(information)] \
	[field_num $w.top.left.ttl "TTL"     20 6  "" invite(ttl)]  \
	[checkbutton $w.top.left.unicast -text "Unicast Conference" \
	    -variable invite(unicast) -command inv_cast]\
    -side  top -anchor nw
    
    if [llength $conf(sessions)] {
    disable_win $w.top.left.session_name
    $w.top.left.session_name.in configure -foreground black 
  }
  pack $w.top -expand yes -fill both -ipady 1m -pady 1m

  
  frame $w.top.right 
  pack \
    [button $w.top.right.phonebook -text "Phonebook" -image $image(phonebook) -command sip_phonebook] \
     -side top -fill both

  pack \
    $w.top.left \
    $w.top.right\
    -side left -fill both -padx 2m -pady 1m

    set wbody [frame $w.body -relief groove]
    pack $wbody -expand yes -side top -fill both -ipady 1 -pady 1
 
    set i 0
    set row 0
    set column 0
    set invite(disabled_rows_audio) 0
    set invite(disabled_rows_video) 0

    foreach e {Media Destination Port Encoding Subject} {
	grid  configure \
	    [label $wbody.$row$column -text $e] \
	    -row $row -column $column -sticky w 
	set column [incr column]	
    }

    set wend [frame $w.end  -relief groove -borderwidth 1 ]
    pack \
	[button $wend.cancel -text "Send Invitation" -command "process_invitation" -relief raised]\
	[button $wend.add -text "Cancel" -command "destroy $w" -relief raised]\
	-side left -expand yes -fill both
    pack $wend -expand yes -side top -fill both -ipady 5 -pady 1

    # check for media_list (already active sessions or session parameters specified by a SIP response)
    if {[info exists invite(media_list)]} {
	foreach media $invite(media_list) {
	    # create a row in the invitation window
	    set type [lindex $media 0]
	    if {[lsearch {audio video} $type] > -1} {
		
		# create row
		set row [inv_add_row_$type $wbody]
		set invite($row) 1

		enable_row $wbody $row
		set invite(disabled_rows_$type) [incr invite(disabled_rows_$type) -1]
		# set parameters
		set invite($row,code_list) ""
		set invite($row,profile_list) ""
		foreach code {lpc gsm dvi4 pcmu dvi16 l16_1 l16_2 h261 cellb jpeg mpeg} {
		    set invite($row,$code) 0
		}
		set invite($row,session_addr) [lindex $media 2]
		set invite($row,port) [lindex $media 1]
		set invite($row,ttl) [lindex $media 3]
		foreach code [lindex $media 4] {
		    if {[lsearch $m_rules $code] > -1} {
			set invite($row,$code) 1
			update_code_list $row $code
		    }
		}
		set invite($row,information) [lindex $media 5]
	    }
	}
    }

    inv_add_row_audio $wbody 
    inv_add_row_video $wbody 
}


proc inv_add_row_audio {w} {
    global invite

    if ![info exists invite(disabled_rows_audio)] {
	set invite(disabled_rows_audio) 0
    }	
    set column 0
    set row [lindex [grid size $w] 1]

    if {$invite(unicast)} {
	set invite($row,session_addr) [idstring %i]
    } else {
	set invite($row,session_addr) $invite(session_addr)
    }
#[join [lreplace [split  $invite(session_addr) .] 3 3 [expr [lindex [split $invite(session_addr) .] 3]+$row + $column ]] .]
    set invite($row,port) [expr $invite(port) + 2 * $row]
    set invite($row,ttl) $invite(ttl)
    grid configure \
	[button $w.$row$column  -relief raised  \
	     -padx 4 -width 100\
	     -command "row_select $w $row audio" ] \
	-row $row -column $column -sticky w

    compound_win $w.$row$column audio "Audio"
    set column [expr $column+1]

    grid configure \
 	[entry $w.$row$column  -width 15 -textvariable invite($row,session_addr) \
	     -relief groove -state disabled -foreground grey \
	     -background white] \
	-row $row -column $column -sticky w
    set column [expr $column+1]

    grid configure \
	[entry $w.$row$column  -width 6 -textvariable invite($row,port) \
	     -relief groove -state disabled -foreground grey \
	     -background white ] \
	-row $row -column $column -sticky w
    set column [expr $column+1]

    grid configure \
	[menubutton  $w.$row$column  \
	     -width 20  -relief groove -state disabled \
	     -text "Encoding" \
	     -menu $w.$row$column.menu] \
	-row $row -column $column -padx 5 -sticky w 
    set m [menu $w.$row$column.menu]
    
    set invite($row,code_list) ""
    set invite($row,profile_list) ""
    foreach code {lpc gsm dvi4 pcmu dvi16 l16_1 l16_2 h261 cellb jpeg mpeg} {
	set invite($row,$code) 0
    }
    add_code_menu $m $row "LPC, 8000 Hz, mono (5.6 kb/s)" audio lpc
    add_code_menu $m $row "GSM, 8000 Hz, mono (13 kb/s)" audio gsm
    add_code_menu $m $row "DVI, 8000 Hz, mono (32 kb/s)" audio dvi4
    add_code_menu $m $row "PCM, 8000 Hz, mono (64 kb/s)" audio pcmu
    add_code_menu $m $row "DVI, 16000 Hz, mono (64 kb/s)" audio dvi16
    add_code_menu $m $row "16-bit, 44100 Hz, mono (705 kb/s)" audio l16_1
    add_code_menu $m $row "16-bit, 44100 Hz, stereo (1.4 Mb/s)" audio l16_2
    
    disable_win $w.$row$column 

    set column [expr $column+1]
    set invite($row,information) ""
    grid configure \
	[entry $w.$row$column  -width 20 -textvariable invite($row,information) \
	     -relief groove -state disabled -foreground grey \
	     -background white] \
	-row $row -column $column -sticky w

    set invite(disabled_rows_audio) [incr invite(disabled_rows_audio)]
    set invite($row) 0
    set invite($row,type) audio
    return $row

}


proc inv_add_row_video {w} {
    global invite

    if ![info exists invite(disabled_rows_video)] {
	set invite(disabled_rows_video) 0
    }	
    set column 0
    set row [lindex [grid size $w] 1]

    if {$invite(unicast)} {
	set invite($row,session_addr) [idstring %i]
    } else {
	set invite($row,session_addr) $invite(session_addr)
    }
#[join [lreplace [split  $invite(session_addr) .] 3 3 [expr [lindex [split $invite(session_addr) .] 3]+$row + $column ]] .]
    set invite($row,port) [expr $invite(port) + 2 * $row]
    set invite($row,ttl) $invite(ttl)

    grid configure \
	[button $w.$row$column  -relief raised  \
	     -padx 4 -width 100\
	     -command "row_select $w $row video" ] \
	-row $row -column $column -sticky w

    compound_win $w.$row$column video "Video"
    set column [expr $column+1]

    grid configure \
 	[entry $w.$row$column  -width 15 -textvariable invite($row,session_addr) \
	     -relief groove -state disabled -foreground grey \
	     -background white] \
	-row $row -column $column -sticky w
    set column [expr $column+1]

    grid configure \
	[entry $w.$row$column  -width 6 -textvariable invite($row,port) \
	     -relief groove -state disabled -foreground grey \
	     -background white ] \
	-row $row -column $column -sticky w
    set column [expr $column+1]

    grid configure \
	[menubutton  $w.$row$column  \
	     -width 20  -relief groove -state disabled \
	     -text "Encoding" \
	     -menu $w.$row$column.menu] \
	-row $row -column $column -padx 5 -sticky w 
    set m [menu $w.$row$column.menu]

    set invite($row,code_list) ""
    set invite($row,profile_list) ""
    foreach code {lpc gsm dvi4 pcmu dvi16 l16_1 l16_2 h261 cellb jpeg mpeg} {
	set invite($row,$code) 0
    }
    add_code_menu $m $row "H.261 (high quality)" video h261
    add_code_menu $m $row "CellB (low quality, fast)" video cellb
    add_code_menu $m $row "JPEG (high quality, high rate)" video jpeg
    add_code_menu $m $row "MPEG (high quality, lower rate)" video mpeg

    disable_win $w.$row$column 

    set column [expr $column+1]
    set invite($row,information) ""
    grid configure \
	[entry $w.$row$column  -width 20 -textvariable invite($row,information) \
	     -relief groove -state disabled -foreground grey \
	     -background white] \
	-row $row -column $column -sticky w

    set invite(disabled_rows_video) [incr invite(disabled_rows_video)]
    set invite($row) 0
    set invite($row,type) video

    return $row
}

proc inv_cast {} {
    global invite

    if {$invite(unicast)} {
	set addr [idstring %i]
    } else {
	set addr $invite(session_addr)
    }
    for {set row 1} {$row < [lindex [grid size .invite.body] 1]} {incr row} {
	set invite($row,session_addr) $addr
    }
}


proc row_select {w row type} {
    global invite Session

    if {$invite($row) == 1} {
	disable_row $w $row
	$w.${row}0 configure -state normal
	set invite(disabled_rows_$type) [incr invite(disabled_rows_$type)]
	set invite($row) 0
	# reset the code values to none (dor)
	set invite($row,[ name_2_code $Session($type)])  0
	.invite.body.${row}3 configure -text Encoding
	set invite($row,code_list) ""
    } else {
	enable_row $w $row
	set invite(disabled_rows_$type) [incr invite(disabled_rows_$type) -1]
	set invite($row) 1
	# set the value of the code to the default code value for media type 
	# (dor)
	set invite($row,[ name_2_code $Session($type)]) 1
	.invite.body.${row}3 configure -text $Session($type)
	update_code_list $row  [name_2_code $Session($type)]
	if { $invite(disabled_rows_$type) <= 0} {
	    inv_add_row_$type $w
	}
    }	
}


#
# update list of selected codings
#
proc update_code_list {row code} {
  global invite

  if {$invite($row,$code)}  {
    lappend invite($row,code_list) $code
    lappend invite($row,profile_list) [get_profile $code]
  } else {
    set index [lsearch  $invite($row,code_list) $code] 
    catch {set invite($row,code_list) [lreplace $invite($row,code_list) $index $index]}
    catch {set invite($row,profile_list) [lreplace $invite($row,profile_list) $index $index]}
  }

    if {$invite($row,code_list) != ""} {
	.invite.body.${row}3 configure -text [string toupper $invite($row,code_list)]
    } else {
	.invite.body.${row}3 configure -text Encoding
    }

} ;# update_code_list


proc add_code_menu {m row label type code} {
    global mailcap_rules m_rules

    set m_rules ""
    foreach rule $mailcap_rules($type) {
	lappend m_rules [lindex $rule 0]
    }
    set invite($row,$code) 0
    if {[lsearch $m_rules $code] > -1} {
	$m add check -label $label -variable invite($row,$code) -command "update_code_list $row $code"
    }
}




#
# manage creation and sending of a SIP request
# invoked when pressing the "Send invitation" button
#
proc process_invitation {} {
    global SIP invite instance
    
    if {$invite(guest) == ""} {
	warning "No guest selected!\nPlease modify your entries."
	invite_proc
	return
    }
    if {[create_media_list]} {
	invite_proc
	return
    }
    catch {destroy .invite}
    
  # open and read phonebook file
  set phonebook_entries [read_file .sipphonebook]
 
  set unaliased_guests ""
  foreach guest [string trim [split $invite(guest)]] {
    set index [lsearch $phonebook_entries "$guest:*"]
    if {$index >= 0} {
	lappend unaliased_guests [string trim [lindex [split [lindex $phonebook_entries $index] ":"] 1]]
    } else {
      lappend unaliased_guests $guest
    }
  }

  set invite(guest) [join $unaliased_guests]

  foreach guest [string trim [split $invite(guest)]] {
    # create globally unique request-id
    incr instance
    set call_id "[idstring %i]/[pid]/$instance"
    lappend invite(active_invitations) $call_id
      
    if {[create_sip_request $guest $call_id]} {
      invite_proc
      return
    }
    send_sip_request $call_id
  }
} ;# process_invitation



#
# create a SIP request to user "guest"
#
proc create_sip_request {guest call_id} {
  global SIP invite request conf

  # create SDP payload
  
  # session description
  set payload "v=0\r\n"
   
  set request($call_id,origin) "[idstring %u] $conf(id) $call_id IN IP4 [idstring %i]"
  append payload \
    "o=$request($call_id,origin)\r\n"

  set request($call_id,session_name) $conf(name)
  append payload \
    "s=$request($call_id,session_name)\r\n"

  if {$invite(information) != ""} {
    set request($call_id,information) $invite(information)
    append payload \
      "i=$request($call_id,information)\r\n"
  }

  if [info exists invite(uri)] {
    set request($call_id,uri) $invite(uri)
    append payload \
      "u=$request($call_id,uri)\r\n"
  }

  if {$invite(email) != ""} {
    set request($call_id,email) $invite(email)
    append payload \
      "e=$request($call_id,email)\r\n"
  }

  if {$invite(phone) != ""} {
    set request($call_id,phone) $invite(phone)
    append payload \
      "p=$request($call_id,phone)\r\n"
  }

  set request($call_id,session_addr) $invite(session_addr)
  set request($call_id,ttl) $invite(ttl)
  append payload \
    "c=IN IP4 $invite(session_addr)/$invite(ttl)\r\n"

  # time description

  # media description
  set request($call_id,media_list) $invite(media_list)
  if [llength $invite(media_list)] {
    foreach media $invite(media_list) {
      media2invite $media
      set m [lindex $media 0]
      append payload \
        "m=$m $invite($m,port) RTP/AVP $invite($m,profile_list)\r\n"
   
      if {[info exists invite($m,information)] && ($invite($m,information) != "")} {
        append payload \
          "i=$invite($m,information)\r\n"
      }  

      if {[info exists invite($m,session_addr)] && ([string compare $invite($m,session_addr) $invite(session_addr)] || [string compare $invite($m,ttl) $invite(ttl)])} {
        append payload \
          "c=IN IP4 $invite($m,session_addr)/$invite($m,ttl)\r\n"
      }
    }
  } else {
    warning "No media selected!\nPlease modify your entries."
    return 1
  }

  # create SIP request header
  set header "INVITE sip://$guest $SIP(VERSION)\r\n"

  append header \
    "VIA:$SIP(VERSION)/TCP  [idstring %i]\r\n"

  append header \
    "Call-id:$call_id\r\n"
  append header \
    "From:sip://[idstring %u@%d]\r\n"

  if {[string first @ $guest] < 0} {
    set guest $guest@[idstring "%d"]
  }
  set request($call_id,to) $guest
  append header \
    "To:sip://$guest\r\n"
  append header \
    "Content-type:application/sdp\r\n"
    
  # calculate payload length and append it to header
  append header \
    "Content-length:[string length $payload]\r\n"

  # terminate header with an empty line
  append header "\r\n"

  set request($call_id) $header
  append request($call_id) $payload
  regsub -all -- "\t" $request($call_id) " " request($call_id)
  return 0
} ;# create_sip_request



#
# send request via TCP to a SIP server
# - extract server address
# - establish TCP connection to the server
# - install timer for receiving a definite response
# - send request
#
proc send_sip_request {call_id} {
  global SIP invite request connected

  # get address of the appropriate SIP server
  # split guest into user and domain part

  set to [split $request($call_id,to) @]
  set user_name [lindex $to 0]
  set domain_part [lindex $to 1]
  if {$user_name == ""} {
    warning "No user selected!\nPlease modify your entries."
    invite_proc
    return
  }

  # get SIP server for domain_part
  set server_list [servers $domain_part]
  if {$server_list ==  ""} {
    warning "Cannot resolve host/domainname!\nPlease modify your entries."
    invite_proc
    return
  } 
  
  # try to connect to SIP server 
  set connected 0
  foreach s $server_list {
    footer "Trying to connect to [lindex $s 0]"
    #if [catch {set request($call_id,sock) [socket [lindex $s 1] $SIP(PORT)]} msg ] {
     # connection failed, try next server in server_list
     # continue
    #}
    # converting to nonblocking connection request
    set request($call_id,sock) [socket -async [lindex $s 1] $SIP(PORT)]
      proc connect {call_id } {
	  global request connected
	  set connected 1
	  fileevent $request($call_id,sock) writable ""
      }
      fileevent $request($call_id,sock) writable [list connect $call_id]
      if { $connected == 0 } {
	  vwait connected
      }
    if [catch { fconfigure $request($call_id,sock) -peername } msg] {
	# connection failed, try next server in server_list
	catch {close $request($call_id,sock)}
	set connected 0
	continue
   }
    fconfigure $request($call_id,sock) -translation binary
    fconfigure $request($call_id,sock) -buffering full
    fileevent $request($call_id,sock) readable "read_sip_response $request($call_id,sock) $call_id"
    footer "sending SIP request"

    puts $request($call_id,sock) $request($call_id)
    flush $request($call_id,sock)
    set request($call_id,timer) [after $SIP(MAX_TIME) "sip_timeout $call_id"]
    return
  }
  if {$connected == 0} {
    footer "Cannot connect to $domain_part $SIP(PORT)"
  }
} ;# send_sip_request



#
# timeout function of a SIP request
# invoked if no definite response was received during the timeout interval
#
proc sip_timeout {call_id} {
  global invite image request
    
  if [info exists request($call_id,sock)] {
    close $request($call_id,sock)
    unset request($call_id,sock)
  } 
  set index [lsearch $invite(active_invitations) $call_id]
  if {$index != -1} {
    set invite(active_invitations) [lreplace $invite(active_invitations) $index $index]

    set w .response[noperiod $call_id]
    catch {destroy $w}

    toplevel $w
      # put the windo in a fixed place
      set x [expr [winfo screenwidth .] /2 ]
      set y [expr [winfo screenheight .] /2 ]
      wm geometry $w +$x+$y
   
    frame $w.head
    pack \
      [label $w.head.im -image $image(afraid)] \
      [label $w.head.title -text "Timeout !" -font *-*-*-bold-i-*-18-*-*-*-*-*-*-*] \
      -side left -anchor w 
     
    pack \
      $w.head \
      [label $w.l -text "The call to \"$request($call_id,to)\" was unseccessful\nThe time limit expired."] \
      [button $w.b -text "Dismiss" -command "destroy $w"] \
      -side top
  }
} ;# sip_timeout


