#
# Per-member commands 'c' for member 'm' in session 's'; issued by
# remote unless noted.  We ignore 'cname' command since that must be
# set by hello to allow matching across media. If SDES information is
# empty (e.g., a note is cancelled), CNAME is used again.
#
# $Id: member.tcl,v 1.28 1998/10/09 17:20:17 dor Exp $
#

proc member {s m c args} {
  global conf member debug

  regsub -all % $s %% s_nopercent 
  set c [string tolower $c]

  # if member status doesn't exist, create it
  set mm [noperiod $m]
  if {![info exists member($m,$s,state)] && ($c != "closed")} {
    member_add $s $m
  }
    if { $debug == "pmm" } {
	puts "incoming: $conf(id)/$s/$m $c $args"
    }
    switch -glob $c {
    created {
    }
    bye {
      member_delete $s $m
      footer "$m leaves: [join [lindex $args 0]]"
    }
    cname {
    }
    enlarge {
    }
    loc   -
    name  -
    email -
    phone -
    note  -
    img   -
    tool  {
      set member($m,$c) [join [lindex $args 0]]
      if {$member($m,$c) == "" && $c == $conf(member_display)} {	  
        set member($m,display) $c
        member_label $m member($m,cname)
      } elseif {[lsearch $conf(member_display) $c] < 
          [lsearch $conf(member_display)  $member($m,display)]} {
        set member($m,display) $c
        member_label $m member($m,$c)
      }
    }
    from -
    source {
      set member($m,$s,$c) [join [lindex $args 0]]
    }
    !alive {
      set member($m,$s,alive) red
    }
    alive {
      set member($m,$s,alive) [.conference cget -background] 
    }
    active {
      if {[lindex $args 0] == 0} {
        set member($m,$s,state) r
      } else {
        set member($m,$s,state) A 
      }
	set tmp $member($m,$s,active)
	# note this member for this session as active
	set member($m,$s,active) 1
	if { $member($m,active)!=1 || $tmp!=1 } {
	    set member($m,active) 1
	    # create the basic frame for the display
	    create_entry $m $s
	}
    }
    !active {
      set member($m,$s,active) 0
      set member($m,$s,state) r
    }
    receive {
    }

    receiving {
	catch {set member($m,$s,rx) [lindex $args 0]}
    }
    statistics {
      statistics $s $m [lindex $args 0]
    }
    statistics? {
    }
    
    # fim
    vic_button1 {
      member_picture_button1 $s_nopercent $m
    }

    vic_button2 {
      member_button2 $s_nopercent $m
    }
      
    vic_button3 {
      if {$member($m,local)} {
	  member_button3_local $s_nopercent $m 0
	  member_button3_local $s_nopercent $m 1
      }
    }
      
	# vic_win_id is important for integrating the thumbnail images of vic
	# into isc. when coming from vic it contains 
	# cname winodw name  window id (the window id is the name of the window
	#                                   to steal )
	# when sent bu isc it continas
	# cname winodw name (0 or -1) 0 indicates the success of the stealing 
	#                               operation, -1 failliure
    vic_win_id {
	set ss [noperiod $s]
	
	# Save parameters
	set name [lindex [lindex $args 0] 0]
	set vic_win_name [lindex [lindex $args 0] 1]
	set vic_win_id [lindex [lindex $args 0] 2]
	if { $vic_win_id > 0 } {
	    set member($m,$s,vic_win_id) $vic_win_id
	    
	    # Destroy current label and replace it by a frame
	    # Leads to ERROR when the configure menu is invoked,
	    # because destroyed label is to be modified there  
	    
	    #update idletasks  
	    
	    # Steal  window from vic and embed it into frame
	    # (uses the xaccess function from the TkSteal extension)
	    set counter 0
	    while {[catch {xaccess eventreparent -parentwidget $member($m,picture) -windowid $vic_win_id} msg] && $counter < 10} {
		global debug
		after 500
		if { $debug == "pmm" } {
		    puts stderr "$msg"
		}
		set counter [expr $counter + 1 ]
		#update idletasks
	    }
	    if  { $counter >=10 } {
		control_send "$conf(id)/$s/$member($m,cname) vic_win_id $name $vic_win_name -1" 
	    } else {
		control_send "$conf(id)/$s/$member($m,cname) vic_win_id $name $vic_win_name 0"
	    }
	}
    }

    error {
      set msg "Error for $member($m,cname): [join [lindex $args 0]]"
      if {[winfo exists .notice]} {
        footer $msg
      } else {
        tk_dialog .notice "Error for $member($m,cname)" [join [lindex $args 0]] \
        warning 0 OK
      }
    }
     
    default {
	footer "[pid] Unknown command: member $s $m $c $args"
    }
  }
}

# check if the member is active in any of its sessions, if not remove its 
# display window
# member($m.,active) = indicate the activity of the member in all sessions
#                      the member is involved in
proc member_active { } {
    global member member_list delay
    foreach m [array names member_list] {
	# check if the member is local and if it was active
	set member($m,active) 0
	if { $member($m,local)==0 } { 
	    for {set i 0} {$i < [llength $member($m,sessions)]} {incr i} {
		set s [lindex $member($m,sessions) $i]
		set member($m,active) [expr $member($m,active) | $member($m,$s,active)]
	    }
	    if { $member($m,active) == 0 } {
		for {set l 0} {$l < [llength $member($m,sessions)]} {incr l} {
		    set s [lindex $member($m,sessions) $l]	    
		    member_delete_entry $s $m 0
		}
	    }
	}
    }
    after $delay(member_update_time) member_active
}


#
# Modify per-member display of 'm' (name, location, ...), showing 'text'.
# Member name 'm' has been stripped.
#
proc member_label {m var} {
    global member
    set mm [noperiod $m]
    if ![info exists member($m,member_frame)] return
    if { [winfo exists $member($m,member_frame).label] } {
	$member($m,member_frame).label configure -textvariable $var
    }
}

#
# Trace procedure; invoked when $member($m,$s,state)
# (A/a=active,r=recent) or $member($m,$s,rx) (0,1) changes.
#
proc member_trace {name element op} {
  global member session

    # extract member name 'm' and session 's' from element name
    set c [split $element ,]
    set m [lindex $c 0]
    set s [lindex $c 1]
    set ss [noperiod $s]
    set mm [noperiod $m]
  
    set mem 0
    if {$member($m,local)} {
	set mem 1
    }
    if {$session($s,type) == "audio" || $session($s,type) == "video"} { 
	set bitmap_file $session($s,type)_$mem$member($m,$s,state)$member($m,$s,rx)
    } else {
	set bitmap_file $session($s,type)
    }

    $member($m,container).${ss} configure \
      -bitmap $bitmap_file \
      -background $member($m,$s,alive) 
}

#
# Return 1 if 'm' is local (us).
#
proc member_islocal {m} {
  global conf

  return [expr [lsearch $conf(us) $m] >= 0]
}

# initialize some global variables for the standalone mode
set map [list 0]


#  member(m,member_frame) consists of
#      -------------
#      |  label
#      |  picture
#      |  container
#

#
#find an empty slot for the rsvp entry
#
proc get_location {} {
    global map
    if { [lsearch -exact $map 0 ] < 0 } {
	set position [llength $map]
	set map [lappend map 1]
    } else {
	set position [lsearch -exact $map 0 ]
	set map [lreplace $map $position $position 1]
    }
    return $position
}

#
# clear an entry in the map
#
proc clear_location {m} {
    global map member CONFIGURE_LEN
    set index [ expr [expr $member($m,row) * $CONFIGURE_LEN] \
		    +$member($m,column)]
    set map [lreplace  $map  $index $index  0]
    
}


#
# create the window for the new member
#
proc create_entry { m  s} {
    global member CONFIGURE_LEN Session

    set ss [noperiod $s]
    # did we already create an antry for this member
    if { [info exists member($m,member_frame)] } {
	if { ![winfo exists $member($m,container).${ss}] } {	
	    # create the control button for this session in this members entry
	    member_add_entry $s $m
	}
	return
    }
    set times12     "-*-times-small-r-normal--*-120-*"
    set position [get_location]
    set member($m,row)  [expr $position / $CONFIGURE_LEN]
    set member($m,column) [expr $position % $CONFIGURE_LEN]
    # create windows for member
    set member($m,member_frame) [frame $Session(f_members).$position \
				     -relief groove -borderwidth 2 ]
    grid configure $member($m,member_frame)  \
	-row $member($m,row) -column $member($m,column) -sticky w 
    set member($m,picture) [frame $member($m,member_frame).picture \
				-relief raised -bg white -borderwidth 2 \
				-width 80 -height 60 ]
    set member($m,container) [frame $member($m,member_frame).container \
				  -relief groove -borderwidth 2]
    if { $member($m,local) } {
	set mem_label $Session(name)
	set member($m,name) $Session(name)
	set member($m,email) $Session(email)
	set member($m,tool) $Session(tool)
    } else {
	set mem_label $member($m,cname)
    }
    
    pack \
	[label $member($m,member_frame).label -text $mem_label \
	     -padx 1 -width 10 -font $times12 -anchor w]\
	$member($m,picture) \
	$member($m,container)\
	-side top -expand yes -fill both -padx 2 -pady 1 -anchor w
    set member($m,win24) [video_8to24 [winfo id $member($m,picture)]]

    iscBalloon $member($m,member_frame).label "\
Cname:  $member($m,cname)
" iscBalloonMember $m
    iscBalloon $member($m,picture) enlarge


    # create the control button for this session in this members entry
    member_add_entry $s $m


    
}


# add the button to control the reception and sending of data as well
# as binding the needed operations to the buttons
proc member_add_entry { s m } {
    global conf member session colors 
    

    set ss [noperiod $s]
     #find out in which column to add the state button. if $s is not in 
    #conf(entry) we assume that it will be added there as the next index
    if {[lsearch -exact $conf(entry) $s] < 0} {
	set col [llength $conf(entry)]
    } else {
	set col [lsearch -exact $conf(entry) $s]
    }

    set member($m,$s,color) [expr $col % [llength $colors]]
    # Add per-session state display
    pack [button $member($m,container).${ss} -relief raised \
	      -fg [lindex $colors $member($m,$s,color)] ] \
	-side left -fill both -expand yes 
    #add help ballon
    iscBalloon $member($m,container).${ss} control_media
    # Initialize $member($m,$s,rx) so that trace works.
    set member($m,$s,rx)    1
    set member($m,$s,alive) [.conference cget -background]
    set member($m,$s,state) A
    
    # add variable write trace for member
    trace variable member($m,$s,state) w member_trace
    trace variable member($m,$s,alive) w member_trace
    trace variable member($m,$s,rx)    w member_trace
    
    # Disable talking for local member.
    set member($m,$s,rx) [expr [member_islocal $m]]
    #set member($m,$s,rx)    0
    set member($m,$s,state)    r
    set member($m,$s,enlarge)    0
    
    # disable previewing for local member
    set member($m,$s,preview)    [expr [member_islocal $m]]
    
    
    regsub -all % $s %% s_nopercent
    if {$session($s,type) == "video" || $session($s,type) == "picture"} {
	bind $member($m,container).${ss} <ButtonRelease-1> \
	    "member_button1 $s_nopercent $m  $member($m,win24)"
	# force the application to accept all incoming data
	member_button1 $s_nopercent $m  $member($m,win24)
    } else {
	bind $member($m,container).${ss} <ButtonRelease-1> \
	    "member_button1 $s_nopercent $m "
	# force the application to accept all incoming data
	member_button1 $s_nopercent $m  
    }	
    if {$session($s,type) == "netghost"} {
	bind $member($m,container).${ss} <ButtonRelease-1> \
	    "config_media $s netghost"
    }

    bind $member($m,container).${ss} <ButtonRelease-2> \
	"member_button2 $s_nopercent $m"
    
    bind $member($m,picture) <ButtonRelease-1> \
	"member_picture_button1 $s_nopercent $m"
    
    #for the local member define the 3 button to preview ones own data
    if {$member($m,local)} {
	# fim
	#bind $member($m,container).${ss} <ButtonRelease-3> \
	#    "member_button3_local $s_nopercent $m 1"
	#bind $member($m,container).${ss} <ButtonPress-3> \
	#    "member_button3_local $s_nopercent $m 0"
	
    }
}
#
# Add media session 's' to list of member 'm', using initial value 'text'
# in display.
#
proc member_add {s m {text ""}} {
    global conf member session member_list

    set ss [noperiod $s]
    set fn $conf(id)/$s/$member($m,cname)
    # check if this is first session for member
    if {![info exists member($m,sessions)]} {
 	set member($m,sessions) {}
	set member($m,display) cname
	set member($m,local)    [member_islocal $m]
    }
        
    if { $conf(conference)==1 || $member($m,local)==1 } {
	create_entry $m $s
    }
    
    # update the member_list array
    set member_list($m) $member($m,cname)

    # Check if this session already exists for member. Can occur for
    # implicit hello.
    if {[info procs $fn] == ""} {
	# Create new command.
	eval "proc $fn {c args} {member $s $m \$c \$args}"
    }
    # Add to session list for this member and sort the list alphabetically.
    if { [lsearch -exact $member($m,sessions) $s] <0 } { 
	ladd member($m,sessions) $s
    }
    # Add to member list for this session.
    if { [lsearch -exact $session($s,members) $m ] <0 } {
	ladd session($s,members) $m
    }
    # initialize the member as not active 
    set member($m,active) 0
    set member($m,$s,active) 0

}

#
# Left mouse button: toggle sending for a member 'm' of session 's'.
#
proc member_button1 {s m  {win ""} } {
    global conf member session
    set member($m,$s,rx) [expr !$member($m,$s,rx)]
    if {$member($m,local)} {
	control_send "$conf(id)/$s send $member($m,$s,rx) $win"

	#make sure that the control buttons and the control icons 
	#show thesame state for transmission
	set session($s,send) $member($m,$s,rx)
    } else {
	control_send "$conf(id)/$s/$member($m,cname) receive $member($m,$s,rx) $win "
    }
}

#
# Left mouse button: toggle showind large picture for member 'm' of session 's'
#
proc member_picture_button1 {s m} {
    global conf member session
    set member($m,$s,enlarge) [expr !$member($m,$s,enlarge)]
    
    control_send "$conf(id)/$s/$member($m,cname) enlarge $member($m,$s,enlarge)"
    #control_send "$conf(id)/$s/$member($m,old_cname) enlarge $member($m,$s,enlarge)"

}


#
# Middle mouse button on member 'm', session 's': member statistics
# for media session 's'.
#
proc member_button2 {s m} {
  global conf member

  set member($m,$s,stats) 1
  control_send "$conf(id)/$s/$member($m,cname) statistics?"
  #control_send "$conf(id)/$s/$member($m,old_cname) statistics?"
}


#
# after releasing button3 redefine conference_button to its original value
proc redefine_conference_button3 {} {
    proc conference_button3 {on} {
	global conf member
	foreach e $conf(sessions) {
	    set member([lindex $conf(us) 0],$e,rx) $on
	    control_send "$conf(id)/$e send $on"
	}
    }
}
#
# right mouse button: toggle previewiing for the local member of session 's'.
# this only has an effect in video -unless we introduce local listen in audio
#
proc member_button3_local {s m on} {
    global conf member session
    
    if {$on==0} {
	#rename conference_button3 conference_button_3.orig
	proc conference_button3 { on } {
	}
    } else {
	set member($m,$s,preview) [expr !$member($m,$s,preview)]
	if {$member($m,local)} {
	    #make sure that the control buttons and the control icons 
	    #show thesame state for the previewing 
	    if { $member($m,$s,preview) >0 } {
		set member($m,$s,state) A
		set session($s,preview) 1
	    } else {
		set member($m,$s,state) r
		set session($s,preview) 0
	    }
	    control_send "$conf(id)/$s preview $member($m,$s,preview)"
	} 
	proc conference_button3 { on } {
	    redefine_conference_button3
	}
    }
}
#
# Delete a single member 'm' from a single session 's', if it exists. 
# Delete statistics panel as well.
#
proc member_delete {s m} {
    global conf member  session member_list
    
    set mm [noperiod $m]
    set ss [noperiod $s]
    
    # update the member list
    if { [info exists member_list($m)] } {
	unset member_list($m)
    }

    if { ![info exists member($m,sessions)] } return
    if { [lsearch -exact $member($m,sessions) $s] <0 } return
    
    # Delete session from member session list.
    lremove member($m,sessions) $s
    # Delete member from session list.
    lremove session($s,members) $m
   # Remove member command.
    rename $conf(id)/$s/$member($m,cname) {}
    member_delete_entry $s $m [member_flag $m $s]
    if {$member($m,sessions) == ""} {
	foreach v [array names member *$m*] {
	    unset member($v)
	}
    }
}

# determine if we need to delete all the members frame or only a control button
# in case of conference mode only the number of session counts
# in case of teleteaching:
# local member: the same as conference mode
# remote member: any session active?
proc member_flag { m s } {
    global conf member
    if { $conf(conference) } {
	return [string compare $member($m,sessions) ""]
    }
    if { $member($m,local) ==1 } {
	return [string compare $member($m,sessions) ""]
    }
    # check if any session is active
    set member($m,active) 0
    set member($m,$s,active) 0
    for {set i 0} {$i < [llength $member($m,sessions)]} {incr i} {
	set s [lindex $member($m,sessions) $i]
	set member($m,active) [expr $member($m,active) | $member($m,$s,active)]
    }
    return $member($m,active)
}


# delete the entry for member m of session s. if flag =0 then delete the 
# entire frame and not just the control buttons

proc member_delete_entry {s m flag} {
    global conf member  session member_list Session
    
    set mm [noperiod $m]
    set ss [noperiod $s]
    
    # delete variable traces 
    trace vdelete member($m,$s,state) w member_trace
    trace vdelete member($m,$s,alive) w member_trace
    trace vdelete member($m,$s,rx)    w member_trace
    if { ![info exists member($m,member_frame)] }  {
	return
    }
    if {$session($s,type) == "video"} {
	pack forget $member($m,picture)
	destroy $member($m,picture)
	if [winfo exists $member($m,member_frame)] {
	    set member($m,picture) [frame $member($m,member_frame).picture \
					-relief raised -bg green \
					-borderwidth 2 -width 80 -height 60 ]
	    pack $member($m,picture)  $member($m,container) \
		-side bottom -expand yes -fill both -padx 2 -pady 1 -anchor w
	    set member($m,win24) [video_8to24 [winfo id $member($m,picture)]]
	}
    }


    # delete statistics panel if exist
    statistics_delete $s $m
    # (fim) Remove session display.
    #update idletasks
    if { [info exists member($m,container)] } {
	destroy $member($m,container).${ss}
    }
    if {$flag ==0} {
	if { [info exists member($m,member_frame)] } {
	    for {set i 0} {$i < [llength $member($m,sessions)]} {incr i} {
		set s [lindex $member($m,sessions) $i]	 
		# inform all sessions that the window no longer exists
		regsub -all % $s %% s_nopercent
		set  member($m,$s,rx) 1
		if {$session($s,type) == "video" || \
			$session($s,type) == "picture"} {
		    member_button1 $s_nopercent $m  $member($m,win24)
		} else {
		    member_button1 $s_nopercent $m  
		}			
	    }
	    destroy  $member($m,member_frame)
	    unset member($m,member_frame)
	    # set a 0 in theentries map for this members location
	    clear_location $m
	}
    }
    update idletasks

}
