#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/fcntl.h>
#include "/usr/include/netinet/in.h"  /* struct sockaddr_in */
#include <arpa/inet.h>   /* inet_ntoa() */
#include <netdb.h>
#include <errno.h>
#include <values.h>
#include <unistd.h>      /* select(), perror() */
#include <string.h>
#include <tk.h>
 
/* 
 * open unicast or multicast UDP connection
 *
 * session: pointer to global session information
 * sin_addr: socket address
 * proto: protocol (RTP, RTCP or conference bus protocol PMM)
 * ttl: time to live for multicast (ignored for unicast)
 * notify: create event handler for the socket descriptor
 * proto_handler: callback function if packet received on the socket
 *
 * To accomodate multi homed hosts we bind the opened sockets to the 
 * address of the active interface -this address is specified at isc-
 * For teh case of multicast we have to bind the sockets with the multicast
 * address but specify the output interface during the IP_ADD_MEMBERSHIP
 * this way the source field of the IP packets always have the right address
 * -in the usual case using just the default values leads the usage of wrong 
 * interface addresses
*/
int udp_cmd(ClientData clientData, Tcl_Interp *interp,
	    int argc, char **argv){
  
  int ttl,port;
  int sock;
  int reuse = 1;  /* reuse address */
  struct sockaddr_in sin;
  int multicast;
  extern struct in_addr host2ip(char *);
  struct in_addr sin_addr;
  char *host;

  if(argc!=4){
    sprintf(interp->result, "udp_cmd: wrong # of args: should be %s host
port ttl",argv[0]);
    return TCL_ERROR;
  }
  
  port = atoi(argv[2]);
  ttl  = atoi(argv[3]);
  host = argv[1];
  fprintf(stderr," port=%i ttl=%i host=%s \n",port,ttl,host);
    
  sin_addr = host2ip(host);
  multicast = IN_MULTICAST(htonl(sin_addr.s_addr));
  memset(&sin, 0, sizeof(sin));		      
  sin.sin_family = AF_INET;
  /* single homed host so use default values */
  sin.sin_addr.s_addr = multicast ? sin_addr.s_addr : htonl(INADDR_ANY);

  sin.sin_port = htons(port);
		      
  /* Open a UDP socket */
  
  sock = socket(PF_INET, SOCK_DGRAM, 0);
  if (sock < 0) {
    perror("can not open socket");
    return(-1);
  }

  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&reuse,
		 sizeof(reuse)) == -1){
    perror("setsockopt: reuseaddr");
    return(-1);
  }
  
  if (multicast) {
    struct ip_mreq mreq;
    int one = 1;
    u_char loop =1;
    int size, len;
    mreq.imr_multiaddr.s_addr = sin_addr.s_addr;
    mreq.imr_interface.s_addr = htonl(INADDR_ANY);
    
    if (setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
		   (char *)&mreq, sizeof(mreq)) < 0) {
      perror("IP_ADD_MEMBERSHIP");
      return(-1);
    }

    /*   if (setsockopt(sock, IPPROTO_IP, IP_MULTICAST_TTL, 
	 (char *)&ttl,sizeof(ttl)) < 0) {
      perror("IP_MULTICAST_TTL");
      return(-1);
      }*/
    
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&one,
		   sizeof(one)) == -1) {
      perror("setsockopt: reuseaddr");
      return(-1);
    }
  
    
    if (setsockopt(sock,IPPROTO_IP,
		   IP_MULTICAST_LOOP, (u_char *)&loop, sizeof(loop)) < 0) {
      perror("IP_MULTICAST_LOOP");
      return(-1);
    }
    /*
     * Make sure the send/recv buffers are of a reasonable size
     * (this is a problem under Solaris 2.x) -- the default size is 0.
     */
    size = 65536;
    len = sizeof(int);
    getsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char *)&size, &len);
    if (size < 65536) {
      size = 65536;
    }
    len = sizeof(int);
    setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char *)&size, 
	       sizeof(int));
    getsockopt(sock, SOL_SOCKET, SO_RCVBUF, (char *)&size, &len);
    if (size < 65536) {
      size = 65536;
    }
    setsockopt(sock,SOL_SOCKET,SO_RCVBUF,(char *)&size, sizeof(int));
    
    
  } /* if multicast */
  
  
  if (bind(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
    /*
     * Binding to multicast address fails if VJ's multicast mods have
     * not been installed, so we try again with INADDR_ANY.
     */
    if (errno == EADDRNOTAVAIL) {
      sin.sin_addr.s_addr = INADDR_ANY;
      if (bind(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0){
	perror("bind");
	return(-1);
      }
    }
  } /* if bind */
  /* set the pmm socket as nonblocking to allow for continous checking of 
   * the socket
   */
  fcntl(sock,F_SETFL,O_NONBLOCK);
  
  sprintf(interp->result,"%d",sock);
  return TCL_OK;
  
}


