# pmm.tcl - This module implements the pmm protocol for mpoll

# Author: Frank Emanuel (fim)

# Debug output
proc dputs {text} {
    global pmm
    if {$pmm(debug) > 0} {
	puts "$pmm(tool): $text"
    }
}

# PMM initialisation procedure 
proc pmmInit {controller} {
    global pmm
    set host [lindex $controller 0]
    set port [lindex $controller 1]
    set pmm(sock_id) [udp $host $port]
    Tcl_SetSockBuffer $pmm(sock_id)
    fconfigure $pmm(sock_id) -ttl 0  
    fconfigure $pmm(sock_id) -buffering none
    fconfigure $pmm(sock_id) -member 1
    fileevent $pmm(sock_id) readable pmmReadSock 
}

# Read PMM  message from socket
proc pmmReadSock {} {
    global pmm
    set cmd [gets $pmm(sock_id)]
    pmmEval $cmd
}

# Evaluate and execute PMM command
proc pmmEval {cmd} {
    global pmm

    dputs "<- $cmd"

    if {[string length $cmd] > 0} {
	if {[lsearch {proc exec rename} [lindex $cmd 0]] >= 0} {
	    puts stderr "$pmm(tool): Forbidden command: $cmd"
	} elseif {[string match "session *" $cmd]} {
	    set conf [lindex $cmd 1]
	    regsub -all " " $conf "%20" conf
	    if {![info exists pmm(sn)]} {
		set pmm(sn) $conf
		dputs "pmm(sn): $pmm(sn)"
		eval "proc pmm_$pmm(sn) {cmd args} {pmmSession $pmm(sn) \$cmd \$args}"
		eval "proc pmm_$pmm(sn)/$pmm(myself) {cmd args} {pmmMember $pmm(sn) $pmm(myself) \$cmd \$args}"
		pmmSendPref "created"
		#pmmSendPref "hello $pmm(myself)"
	    }
	} else {
	    if [catch {eval pmm_$cmd} msg] {
		#puts stderr "$pmm(tool): PMM error: $cmd: $msg"
	    }
	}
    }
}

# Send PMM message
proc pmmSend {msg} {
    global pmm

    puts -nonewline $pmm(sock_id) $msg
    dputs "-> $msg"
}

# Like pmmSend, but prefix our session name.
proc pmmSendPref {msg} {
    global pmm
    
    if [info exists pmm(sn)] {
	pmmSend "$pmm(sn) $msg"
    }
}


# Member related commands 
proc pmmMember {sn cname cmd args} {}

proc send_update {} {
    global pmm
    pmmSendPref " active $pmm(update_interval)"
    after $pmm(update_interval) send_update
} 

# Session related commands
proc pmmSession {sn cmd args} {
    global pmm

    switch $cmd {
	open {
	    pmmSendPref "opened"
	}


	close {
	    pmmSendPref "closed"
	    dputs "Leaving..."
	    pmmExit
	}
	
	update_interval {
	    set pmm(update_interval) $args
	    send_update
        } 
    }
}

# Ignore + command
proc pmm_+ {args} {}

# Ignore agent command
proc pmm_agent {args} {}


#---------------------------------------------
# Main part
#---------------------------------------------
global pmm
set pmm(debug) 0
set pmm(controller) "225.2.0.1 5010"
set pmm(tool) "mpoll-pmm"
set pmm(myself) [my_id]

puts ""
dputs "PMM module started..."

pmmInit $pmm(controller)
pmmSend "agent $pmm(tool) attached [pid]"

