/*
 *  cellb.c --
 *
 *       This file provides functions related to the CellB compression 
 *       algorithm
 *
 *  (C) 1995 Bernd Deffner, GMD Fokus
 *      1996 Dorgham Sisalem
 *  See the copyright file for copyright information
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <values.h>
#include <assert.h>
#ifdef XIL
#include <xil/xil.h>
#endif
#include "xil_capture.h"
#include "types.h"
#include "rtp.h"
#include "tcl.h"
#include "nevit.h"
#include "util.h"

static char rcsid[]  = "$Id: cellb.c,v 1.1.1.1 1997/11/12 17:16:01 dor Exp $";


/*
 * initialize the cellb header part of the RTP header
 */
void CELLB_fill_hdr(session_t *s, char *buf, u_int16 cellx,
		    u_int16 celly)
{
  cellb_hdr_t *cellb_hdr;

  cellb_hdr = (cellb_hdr_t *) buf;
  cellb_hdr->x = htons(cellx);
  cellb_hdr->y = htons(celly);
  cellb_hdr->w = htons(s->local->width);
  cellb_hdr->h = htons(s->local->height);
}



/* 
 * fragment CELLB frames, returns size of compressed frame
*/
int CELLB_encode(session_t *s, int level)
{
  u_int8 *data, *dataLim, *pack, *packLim;
  u_int16 cellx, celly, width;
  int pattern, num_frags;
  struct {
    rtp_hdr_t rtp_hdr;
    cellb_hdr_t cellb_hdr;
  } hdr;
    
  data = s->local->image;
  dataLim = data+s->local->rtp_s[level].nbytes; 
  cellx = 0;
  celly = 0;
  num_frags = 0;

  width = ntohs(s->local->width) >> 2;

  do {
    pack = data;
    packLim = data+s->codec.data_limit-4;
    CELLB_fill_hdr(s, (char *)&hdr.cellb_hdr, 
 		   cellx, celly); 
    while ((data < packLim) && (data < dataLim)) {
      pattern = *data;
      if (pattern >= 128) {
	data++;
	cellx += (pattern-127);
      } else {
	data += 4;
	cellx++;
      }
      
      while (cellx >= width) {
	cellx -= width;
	celly++;
      }
    }

    RTP_fill_hdr(s, (char *)&hdr.rtp_hdr,  
		  (data < dataLim) ? 0 : 1, s->local->rtp_s[level].ts, level);

    /* Send out the packet. */

    s->net.write(level, (char *)&hdr, s->codec.hdr_len, pack, 
		      data-pack);
    num_frags++;
  } while (data < dataLim);

  /*
   * increase timestamp
   */
  assert(s->grabber.frames > 0);
  s->local->rtp_s[level].ts = 
    RTP_inc_ts(s->local->rtp_s[level].ts, 90000 / s->grabber.frames);
  return(s->local->rtp_s[level].nbytes + num_frags * s->codec.hdr_len);
}


void CELLB_init(session_t *s)
{
  s->local->payload_type = RTP_PT_CELLB;
  s->codec.compressor = NEVIT_CELLB;
  s->codec.encode = CELLB_encode;
  s->codec.hdr_len = sizeof(rtp_hdr_t) + sizeof(cellb_hdr_t);
  s->codec.data_limit = NEVIT_MTU - s->codec.hdr_len;
}

