/*
 *  jpeg.c --
 *
 *       This file provides functions related to the JPEG compression algorithm
 *
 *  (C) 1995 Bernd Deffner, GMD Fokus
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <values.h>
#include <assert.h>

#include "types.h"
#include "rtp.h"
#include "tcl.h"
#include "nevit.h"
#include "util.h"

static char rcsid[]  = "$Id: jpeg.c,v 1.3 1998/04/06 08:24:18 dor Exp $";

void JPEG_fill_hdr(session_t *s, char *buf, int offset)
{
  jpeg_hdr_t *jpeg_hdr;

  jpeg_hdr = (jpeg_hdr_t *) buf;
  jpeg_hdr->mbz    = 0;
  jpeg_hdr->off    = htonl(offset);
  jpeg_hdr->type   = 1;
  jpeg_hdr->q      = s->codec.quality;
  jpeg_hdr->width  = s->local->width >> 3;
  jpeg_hdr->height = s->local->height >> 3;
}     

typedef struct {
    rtp_hdr_t rtp_hdr;
    jpeg_hdr_t jpeg_hdr;
  } hdr_t;
/* 
 * fragment JPEG frames, returns size of compressed frame
*/
int JPEG_encode(session_t *s, int level)
{
  hdr_t hdr;
  u_int8 *data;
  int i;
  int num_frags;   /* number of fragments */
  int frag_len;    /* length of the actual fragment */
  int left;        /* # payload bytes to transmit  */

  num_frags = s->local->rtp_s[level].nbytes / s->codec.data_limit + 
    ((s->local->rtp_s[level].nbytes % s->codec.data_limit) > 0); 
  
  left = s->local->rtp_s[level].nbytes;
  data = s->local->image;

  for(i = 0; i < num_frags; i++) {     
      /*
       * add RTP header and set Marker bit if this is the last
       * fragment of the JPEG frame
       */
      RTP_fill_hdr(s, (char *)&hdr.rtp_hdr, (i==num_frags-1), 
		   s->local->rtp_s[level].ts, level);
      
      /* 
       * add JPEG header 
       */
      JPEG_fill_hdr(s, (char *)&hdr.jpeg_hdr,
		    i * s->codec.data_limit);

      /* 
       * send packet 
       */
      frag_len = 
	(i == num_frags - 1) ? left : s->codec.data_limit;
      if(s->net.write(level, (char *)&hdr, s->codec.hdr_len, data,
			frag_len)<0)
	return(-1);;
      left -= s->codec.data_limit;
      data += s->codec.data_limit;
  }    
  /*
   * increase timestamp
   */
  assert(s->grabber.frames > 0);
  s->local->rtp_s[level].ts = RTP_inc_ts(s->local->rtp_s[level].ts, 90000 
					 / s->grabber.frames);
  return(s->local->rtp_s[level].nbytes + num_frags * s->codec.hdr_len);
}



int JPEG_init(session_t *s)
{
  if(s->local->payload_type == RTP_PT_JPEG)
    return 0;
  s->local->payload_type = RTP_PT_JPEG;
  s->codec.compressor = NEVIT_JPEG;
  s->codec.encode = JPEG_encode;
  s->codec.hdr_len = sizeof(rtp_hdr_t) + sizeof(jpeg_hdr_t);
  s->codec.data_limit = NEVIT_MTU - s->codec.hdr_len;
  return 1;
}

