 /*
 * nevit.h  ---
 *
 * The main .h file in nevit contains most of the exported procedures and
 * various global variables
 *
 * 1996 Dorgham Sisalem GMD Fokus
 *
 *  See the copyright file for copyright information
 */

#ifndef _nevit_h_
#define _nevit_h_
#include <sys/socket.h>
#include "/usr/include/netinet/in.h"  /* struct sockaddr_in */
#include <arpa/inet.h>   /* inet_ntoa() */
#include <netdb.h>
#include <tk.h>
#include <tcl.h> 
#include <time.h>

#define MAX_BUF_LENGTH 200000      /* maximum length of buffer */
#define NEVIT_MTU 1024             /* maximum length of network packets */
#define FILTER_PARAMETER 0.2       /* filter parameter: the smaller alpha */
                            /* the higher the influence of the old values */
#define MAX_FRAME_RATE 30          /* maximal frame rate of the sender */
#define SEG 4			   /* maximal number of shared mem segments */

#define Offset(type, field) ((int) ((char *) &((type *) 0)->field))

#define NULLIFIY(theImage,theVisual,theDisp,theGC,mainW,shminfo) \
theImage=NULL; \
theVisual=NULL; \
theDisp=NULL; \
theGC=NULL; \
mainW=NULL; \
shminfo=NULL;

#define RTP  1
#define RTCP 11
#define PMM  22
#define NUM_PROTOCOLS 23
#define NUM_LEVELS  10
#define FIRST_LEVEL 1

/* packet type and commands */

#define BM(x,b) x = (1<<b) 
typedef enum {
  BM(P_bye,               1),
  BM(P_sdes,              2),
  BM(P_traffic,           3),
  BM(P_measure,           4)
} packet_t;


union encap_hdr{
  jpeg_hdr_t jpeg_head;yuv_hdr_t yuv_head;gif_hdr_t gif_head;};

typedef struct {
        rtp_hdr_t *rtp_head;
        union encap_hdr encap_head;
}pack_header;

typedef enum {
  NEVIT_JPEG, NEVIT_CELLB, NEVIT_MPEG, NEVIT_YUV, NEVIT_GIF
} compressor_t;


/* delete flags for the delete_entry function */
enum {DELETE_THIS, DELETE_OTHERS, DELETE_ALL};

typedef struct {
  double frame_size;        /* short term average frame size */
  double frame_rate;        /* short term average frame rate */
  double bw;                /* short term average bandwidth in kbit/s */
} stats_t;

typedef struct sdes {
  int type;              /* type of information */
  char *value;           /* SDES value */ 
  int len;               /* length without \0 */
  struct sdes *next;     /* link to next item in linked list */
} sdes_t;




typedef struct {
  int frames;             /* maximum frames/s */
  int bw;                 /* maximum bandwidth in kbit/s */
  double act_frame_rate;  /* actual frame rate */
  int whichinput;
  void (*init)();
  int (*grab)(char **image);
  int (*close)();
  double pb;               /* peak bandwidth in kbit/s */
  double mb;                /* mean burst length */
} grabber_t;

typedef struct{			/* receiver data structure */
	int son_pid;		/* process id of the displaying process*/
	char **buf; 		/* image pool of the receiver */
        int counter;		/* number of image to use */
        int nbytes;		/* size of image */
        rtp_hdr_t *rtp_head;	/* rtp header of the received image*/
        union encap_hdr encap_head; /* encapsulation header of the received */
				    /* image -jpeg,cellb or YUV */
        void (*init_frame)();
} session_r;

typedef struct {
    u_int32 ts;        /* timestamp (samples) */
    u_int32 seq;       /* sequence number */
}
 timest_t;

/* a structure for handling the app measurement packet */
typedef struct {
  struct timeval last_time;  /* the time last app packet was received */
  struct timeval last_app;   /* the time indicated in the last app packet */
  u_int16  seq;           /* seq number of app packets */ 
  float bbw;           /* current estimate of the bottleneck bandwidth */
} app_meas_t;

/* define a structure for containing the bbw values */
typedef struct {
    int count;
    double average_bbw;
} bbw_classification_t;

typedef struct {
  struct sockaddr_in rmt_addr;
  u_int32 transit;        /* relative trans time for prev pkt */
  double  jitter;         /* estimated jitter */  
  u_int32 mis_order;
  u_int32 lost;
  u_int32 duplicate;
  u_int32 fraction;
  double  rr_jitter;         /* estimated jitter reported in rr*/  
  u_int32 rr_lost;           /* lost reported in rr */
  u_int32 rr_fraction;       /* fraction of loss reported in rr */
  double rr_rtt;          /* round trip time calculated from the last RR */
        
  u_int16 max_seq;        /* highest seq. number seen */
  u_int32 cycles;         /* shifted count of seq. number cycles */
  u_int32 base_seq;       /* base seq number */
  u_int32 bad_seq;        /* last 'bad' seq number + 1 */
  u_int32 probation;      /* sequ. packets till source is valid */
  u_int32 received;       /* packets received */
  u_int32 expected_prior; /* packet expected at last interval */
  u_int32 received_prior; /* packet received at last interval */

  u_int32 frames_sent;   /* overall number of sent frames*/
  u_int32 octets_sent;
  u_int32 packets_last;   /* number of packets sent just at the end of the */
			  /* last statistics update */
  int nbytes;  /* number of bytes received from a packet */
  struct timeval update_time;
  boolean active;

  u_int32 octets_last;
  char *buf;   /* buffer to save the incoming data in */

  time_t created;         /* site created */
  timest_t first; 
  timest_t This;
  timest_t prior;  /* first, current and previous packet */
  u_int32 ts;                        /* timestamp */
  u_int16 seq;                       /* current data sequence number */

  struct timeval last[3];             /* last arrival (data, control, SR) */
  
  /* activity timestamps */
  u_int32 lsr_sec;                    /* last sender report */
  u_int32 lsr_frac;                   /* last sender report */
  double frame_rate;                  /* frame rate used for this level */
  double bandwidth;                   /* bandwidth used for this level */
  app_meas_t app;                     /* strauctute for the measurement part*/  
  /* variable for calculating the bbw for this member */
  bbw_classification_t classification [5][10];
  int max_count;
  double average_bbw;
  int thumbnail;         /* flag for displaying a thumbnail image or not */
} rtp_t;
 
typedef struct {
  u_int32 ssrc;
  char *cname;
  char *name;
  char *email;
  char *loc;
  char *phone;
  char *tool;
  char *note;
  char *img;
  
  int payload_type;                  /* currently only valid for senders */
  rtp_t rtp_s[NUM_LEVELS];          /* rtp data of the seperate levels */
  char *image;
  /*int len;*/      /* number of bytes in image */

  session_r session_rec;	     /* receiver data structure */
  char* decoder;                     /* Address of memory space dedicated for*/
				     /* the decoder variables */
  char* decoder_small;               /* Address of memory space dedicated for*/
				     /* the thumbnail decoder variables */

  int (*xil_stuff)(int ,int , int);  
  void (*redecode)(char *, char* , int);
  char *(*init_decompression)(int , int ,int ,char *, int ,char*);
  char *(*delete_decompression)(char*);
  void (*decompress)(u_char *, int , int, char* , char*);
  int  (*alloc_colors)(char* ,char* , char* ,int );
  void (*draw) (int, int, int, int, char*, char*, char*, char*, char*, char*);
  void (*init_win)(char*);
  int *theImage;       /* Address of memory space dedicated for*/
                                     /* a structure of type XImage */
  int *theDisp ;       /* Address of memory space dedicated for*/
                                     /* a structure of type Display */
  int *theGC ;         /* Address of memory space dedicated for*/
                                     /* a structure of type GC */
  int *mainW;
  int *theVisual;      /* Address of memory space dedicated for*/
                                     /* a structure of type Visual*/
  int *shminfo;        /* Address of memory space dedicated for*/
                                     /* a structure of type XShmSegmentInfo */
  int decompressor;
  int preview;          /* 1=display the incoming frames */
  int width; /* width of video stream images */
  int height; /* height of video stream images */
  char* thumbnail_image; /* ximage place holder for the thumbnail images */
  int* thumbnail_window; /* Address of memory space dedicated for*/
                         /* the thumbnail window*/

  Tk_TimerToken token;  /* token for the thumbnail image */
  int active;           /* activity indication of member:
			 * 1: RTCP message seen in the last update interval
			 * 0: nothing seen
			 */
} member_t;


typedef struct {
  
  /* network I/O */
  struct {
    int (*read)(ClientData client);   /* read from network */
    int (*write)(int proto, char *buf1, int len1, char *buf2, int len2);  
	/* write to network */
    char *(*connect) (char *, int proto, char *host, 
				char *local, int port, u_int8 ttl, 
				int notify, void *);  
	/* connect to network */
    int (*close) (int proto); /* close network */
    int (*read_pmm)(ClientData client);   /* read from local bus */
    int (*write_pmm)(int proto, char *buf1, int len1, char *buf2, int len2);  
	/* write to local bus*/
    int network;			     /* network used: atm or ip */
    boolean connected[NUM_LEVELS];           /* =1 when connected */
    boolean scattered_write_possible; 
  } net;

  /*  codec_t codec;*/
  struct {
  int data_limit;         /* payload size of one packet */
  int hdr_len;            /* RTP header length + encoding header */
  int (*encode)();  /* Pointer to the actual encoding function */
  int quality;            /* compression quality */
  compressor_t compressor; /* JPEG, CELLB, ... */
  } codec;
  grabber_t grabber;
  member_t *local;       /* pointer to entry of source */
  member_t *receive;                 /* pointer to entry of receive */
  member_t active;              /* member structure for the active window*/
  Tcl_HashTable member_table;        /* hash table for members */
  int members;                       /* # of members (-1: only allocated)*/
  int senders[NUM_LEVELS];           /* # of active senders in the last */
                                     /* RTCP interval */
  int rtcp_pkt_size[NUM_LEVELS];     /* size of last RTCP packet in bytes */
  int avg_rtcp_size[NUM_LEVELS];     /* average RTCP packet size */
  boolean passive;                   /* remote control if == 1 */
  boolean send[NUM_LEVELS];          /* activate/deactivate sending */
  boolean opened;
  stats_t stats;
  u_int32 ssrc[NUM_LEVELS];          /* default stream to display*/
  int display;
  struct sockaddr_in rmt_addr[NUM_LEVELS] ;
  int control;
  float scale_factor;        /* scale_fatcor=1 -> no resizing */
                           /* scale_fatcor=2 ->    resizing */

  int delay_rtcp;  /* delay sending the RTCP packet till the next data packet
		    *  delay_rtcp = 1000  -> do not send an RTCP packet prior
		    *                        to a data packt
		    *             = !1000 -> send a RTCP packet before the next
		    *                        data packet
		    */
  int debug;
} session_t;

typedef struct {
  int data_limit;         /* payload size of one packet */
  int hdr_len;            /* RTP header length + encoding header */
  int (*encode)();  /* Pointer to the actual encoding function */
  int quality;            /* compression quality */
  compressor_t compressor; /* JPEG, CELLB, ... */
} codec_t;


/* Protocol identificator for network write and read functions */

/* !!#$@$! enum is better */

 
typedef enum {
  BM(DEBUG_PMM, 1),
  BM(DEBUG_RTP, 2),
  BM(DEBUG_RTCP, 3),
  BM(DEBUG_STATS, 4),
  BM(DEBUG_TRANSMIT, 5)
} debug_t;


/* external functions */

/* rtp.c */
extern void RTP_fill_hdr(session_t *s, char *buf, int marker, u_int32 ts, 
			 int level);
extern member_t* RTP_init(session_t *s, int level);
extern u_int32 RTP_inc_ts(u_int32 old, u_int32 add);
extern void RTP_close(session_t *s);
extern void rtp_calc_loss(rtp_t* rtp_s);
extern void RTCP_receive(session_t *s, char *buf, int len, int level);
extern member_t* RTP_receive(session_t *s, rtp_hdr_t *rtp_hdr, int level);
extern member_t *add_entry(session_t *s, u_int32 ssrc, int level);
extern void delete_entry(session_t *s, u_int32 ssrc, int flag);
extern void RTP_transmit_control(session_t *s, int type, char *reason, int level);
extern void init_seq(member_t *m, u_int16 seq, int level);
extern int display_ssrc(session_t *s, u_int32 ssrc);
extern member_t* receive_display(session_t *s, char* cname,int on);

/* rtp_handler */
extern void rtp_handler(session_t *session,char *buf,int len, int level);


extern struct in_addr host2ip(char *);

/* idstring.c */
extern char *idstring(char *fmt, int alloc);

/* xil_capture.c */
extern char* xil_init_grab(unsigned int *width_,
	      unsigned  int *height_,
	      unsigned  int quality,
			   int input,
              compressor_t compressor);

extern char* xil_open_device();
extern int xil_open_input(unsigned int input);

extern int xil_end_grab();
extern int xil_grab_image(char **image);
extern void xil_end();
extern char* create_display(char *);
extern void destroy_display();
extern void setQ(int quality);
extern void restart_TKStuff();

/* jpeg.c */
extern int JPEG_init(session_t *s);

/* cellb.c */
extern void CELLB_init(session_t *s);

/* pmm.c */
extern int control_send_cmd(ClientData clientData, Tcl_Interp *interp, 
			    int argc, char *argv[]);
extern void pmm_receive(session_t *s, char *buf, int len, int level);
extern int dotcl(Tcl_Interp *interp, char *fmt, ...);


/* nevit_main.c */
extern int Exit();
extern void message_output(char* message,int param);
/* external variables */

extern session_t session;
extern session_t *s;

/* snap and tthumbnail stuff */
extern void thumbnail_handler(ClientData member);
extern int write_snap(int w, int h,char* image, char* theVisual,char* name);
extern void write_thumbnail(member_t* member,char* image);
extern char* init_snap();
extern char* init_thumbnail(member_t *m,Window win);
extern void delete_thumbnail(member_t* m);

/* display.c */
extern char *init_stream(char *decoder,member_t *m);
extern int display_frame(char* image,int len,char * decoder);
extern member_t* receive_display(session_t * s, char* cname, int on);
extern char* create_display(char *decoder);
extern void destroy_display();


/* color definitions */
int BACKGROUND,FORGROUND; 

#endif


