/*
 *  pmm.c --
 *
 *       This file provides functions related to pmm
 *
 *  (C) 1995 Bernd Deffner, GMD Fokus
 */

#include <stdio.h>
#include <sys/types.h>
#include <stdarg.h>
#include "ansi.h"
#include <tcl.h>
#include "types.h"
#include "rtp.h"
#include "nevit.h"
#include "debug.h"


static char rcsid[]  = "$Id: pmm.c,v 1.1.1.1 1997/11/12 17:16:02 dor Exp $";

extern session_t *s;
extern Tcl_Interp *interp_;

/*
 *  creates a Tcl command to send commands to the conference control bus
*/
int control_send_cmd(ClientData clientData, Tcl_Interp *interp, 
		     int argc, char *argv[])
{
  int sent;

  if (argc != 2) {
    interp->result = "Usage: control_send cmd";
    return TCL_ERROR;
  } else {
    debug(DEBUG_PMM, "PMM: nevit-> %s", argv[1]);
    sent = s->net.write_pmm(PMM, argv[1], strlen(argv[1]), NULL, 0);
    sprintf(interp->result, "%d", sent);
    return TCL_OK;
  }
}


/*
 * invoked when a packet is arrived at the conference bus socket.
 * Lets the tcl interpreter evaluate the command
 */
void pmm_receive(session_t *s, char *buf, int len, int level)
{
  buf[len]= '\0';
  debug(DEBUG_PMM, "PMM: ->nevit %s", buf);
  if (Tcl_VarEval(interp_, "control_receive ", "{", buf, "}",
		  (char *)NULL) != TCL_OK)
    fprintf(stderr, "%s\n", interp_->result);
}


/* evaluate a tcl command */
int dotcl(Tcl_Interp *interp, char *fmt, ...)
{
  va_list args;
  char buffer[512];

  va_start(args, fmt);
  vsprintf(buffer, fmt, args);
  va_end(args);

  if (Tcl_GlobalEval(interp, buffer) != TCL_OK) {
    fprintf(stderr, "Error in dotcl: %s\n", interp->result);
    return 1;
  }
  return 0;
} /* dotcl */
