#ifdef ATM
#include <sys/file.h>
#include <sys/fcntl.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <fore_atm/fore_atm_user.h>
#include "types.h"
#include "rtp.h"
#include "tk.h"
#include "nevit.h"
#include "transmit_atm.h"
#include "udp.h"

#define SERVER_SAP  4096

static char rcsid[]  = "$Id: transmit_atm.c,v 1.1.1.1 1997/11/12 17:16:02 dor Exp $";

Atm_info info;
Atm_endpoint dst;
Atm_qos qos;
Atm_qos_sel qos_selected;
Atm_sap ssap;
Aal_type aal = aal_type_5;
Atm_dataflow dataflow =duplex;
char *device_name;
Atm_info mtu_info; 
atm_t atm;
int qlen;

  

 /*  Receive a packet
 */
static int ATM_receive(ClientData client)
{
  static char buf[NEVIT_MTU];
  int len,hdr_len,proto;
  proto = (int)client;
  if ((len=atm_recv(atm.sock[proto], buf, 2000)) < 0) {
    atm_error("atm_recv");
    Exit();
  }
  atm.proto_handler[proto](atm.session, buf, len);
}


char* atm_serevr(int proto, session_t *s, int notify, void (*proto_handler))
{
  u_int		connid,switchid, portid;
  Atm_endpoint	calling;

  ssap = SERVER_SAP;
  qlen = 1;
  if ((atm.sock[proto]= atm_open(device_name, O_RDWR, &info)) < 0) { 
    atm_error("atm_open ");
    return(NULL);
  }
  if (atm_bind(atm.sock[proto], ssap, &ssap, qlen) < 0) {
    atm_error("atm_bind");
    return(NULL);
  }
  printf("SAP server assigned=%d\n", ssap);
  if (atm_listen(atm.sock[proto], &connid, &calling, &qos, &aal) < 0) {
      atm_error("atm_listen");
      return(NULL);
  }
  /*
   * Extract the switch id and port id from the NSAP.
   */
  GET_SWITCH(switchid, calling.nsap);
  GET_PORT(portid, calling.nsap);
 
  printf("Server calling switch=%d, port=%d, sap=%d, aal=%d\n",
        switchid, portid, calling.asap, aal);
 
  printf("qos target peak=%d, mean=%d, burst=%d\n",
        qos.peak_bandwidth.target,
        qos.mean_bandwidth.target,
        qos.mean_burst.target);
 
  printf("Server qos minimum peak=%d, mean=%d, burst=%d\n",
        qos.peak_bandwidth.minimum,
        qos.mean_bandwidth.minimum,
        qos.mean_burst.minimum);
 
  printf("Server connect conn_id=%d\n", connid);

  /*
   * Accept the connection request.
   */
  if (atm_accept(atm.sock[proto],atm.sock[proto], connid, &qos, 
		     dataflow)< 0) {
    atm_error("atm_accept");
    return(NULL);
  }
  atm.proto_handler[proto] = proto_handler;

  if (notify) {
#ifdef NEW
    atm.sock_handler[proto] = atm.sock[proto];
#else
    atm.sock_handler[proto] = Tcl_GetFile((ClientData)atm.sock[proto],
					  TCL_UNIX_FD);
#endif
    Tcl_CreateFileHandler(atm.sock_handler[proto],TK_READABLE,
			  (Tcl_FileProc *)ATM_receive, (ClientData) proto);
  }
  
  return((char *)&(atm.sock[proto]));
}

  /*
   * Request some quality of service.
   */
ATM_qos_init(int tag, char *parameter)
{
  switch ( tag ) {
  case DEVICE: 
    device_name= (char *)malloc(strlen("/dev/") + strlen(parameter) + 1);
    sprintf(device_name, "/dev/%s", parameter);
    break;
  case MEAN_BURST_MINIMUM: 
    qos.mean_burst.minimum		= atoi(parameter);
    break;
  case MEAN_BURST_TARGET: 
    qos.mean_burst.target		= atoi(parameter);
    break;
  case PEAK_BANDWIDTH_MINIMUM: 
    qos.peak_bandwidth.minimum		= atoi(parameter);
    break;
  case PEAK_BANDWIDTH_TARGET: 
    qos.peak_bandwidth.target		= atoi(parameter);
    break;
  case MEAN_BANDWIDTH_MINIMUM: 
    qos.mean_bandwidth.minimum		= atoi(parameter);
    break;
  case MEAN_BANDWIDTH_TARGET: 
    qos.mean_bandwidth.target		= atoi(parameter);
    break;
  default :
    perror(" ATM_qos_init: Unknown parameter ");
  }
}


char*  ATM_connect (session_t *s,int proto,char *destination, char *local, 
		 int port, u_int8 ttl, int notify, void (*proto_handler))
{

 
  if ((atm.sock[proto]= atm_open(device_name, O_RDWR, &info)) < 0) { 
    perror("atm_open ");
    return(NULL);
  }
  /*
   * Bind to a well known sap and set pending
   * connect request queue length.
   */
  ssap = 0;
  qlen = 0;
  if (atm_bind(atm.sock[proto], ssap, &ssap, qlen) < 0) {
    atm_error("atm_bind");
    return(NULL);
  }
  printf("SAP assigned=%d\n", ssap);
  /*
   * Initialize the server's data link address.
   */
  if (atm_gethostbyname(destination, &dst.nsap) < 0) {
    atm_error(" atm_gethostbyname failed ");
    return(NULL);
  }
  fprintf(stderr," nasp=%x \n",dst.nsap.addr);
  dst.asap = SERVER_SAP;
  if (atm_connect(atm.sock[proto], &dst, &qos, &qos_selected, aal,
			      dataflow) < 0) {
      atm_error("atm_connect");
      atm_close(atm.sock[proto]);
      return (atm_serevr(proto, s, notify, proto_handler));
  }
  printf("selected qos peak=%d, mean=%d, burst=%d\n",
	     qos_selected.peak_bandwidth,
	     qos_selected.mean_bandwidth,
	     qos_selected.mean_burst);
      
 
  atm.proto_handler[proto] = proto_handler;

  if (notify) 
    Tk_CreateFileHandler(atm.sock[proto],TK_READABLE,
			 (Tk_FileProc *)ATM_receive, (ClientData) proto);
  return((char *)&(atm.sock[proto]));

   /*    This part used to the connection establishment part for PVC

	 if(!conn){
    	if(atm_bind_pvc(atm.sock[proto],info.vpvc,aal,&(info.qos))<0){
       		 atm_error("atm_bind");
       		 return(-1);
    	}
    }
    else{
    	if(atm_connect_pvc(atm.sock[proto],info.vpvc,aal,&(info.qos))<0){
       		 atm_error("atm_connect");
       		 return(-1);
    	}
    }*/
}


int ATM_write(int proto,char *hdr,int hdr_len,char *start,int length)
{
  int len;
  if (length > 0) {
    memcpy(start-hdr_len,hdr,hdr_len);
    if ((len=atm_send(atm.sock[proto], start-hdr_len, length+hdr_len))< 0) {
      atm_error("atm_send");
      return(-1);
    }
  } else {
    if ((len=atm_send(atm.sock[proto], hdr, hdr_len)) < 0) {
      atm_error("atm_send");
      return(-1);
    }
  }
  return(len);
}


ATM_close(int proto)
{
	atm_close(atm.sock[proto]);
}


/* initialize the atm related network procedures */
ATM_init(session_t *session)
{
  atm.session = session;     /* provide pointer to session structure */

  /* set generic function pointers to the UDP functions */
  session->net.read       = ATM_receive;
  session->net.write      = ATM_write;
  session->net.connect    = ATM_connect;
  session->net.close      = ATM_close;
  session->net.read_pmm   = UDP_receive;
  session->net.write_pmm  = UDP_write;
}
#endif
  
