#include <tcl.h>
#include <tk.h>
#include "x11_grab.h"

/**************************************/
static int lowbitnum(ul)
     unsigned long ul;
{
  /* returns position of lowest set bit in 'ul' as an integer (0-31),
   or -1 if none */

  int i;
  for (i=0; ((ul&1) == 0) && i<32;  i++, ul>>=1);
  if (i==32) i = -1;
  return i;
}
/**************************************/
/* following code snarfed from 'xwd.c' */
/**************************************/

#define lowbit(x) ((x) & (~(x) + 1))


int getxcolors(win_info, colors,theDisp)
     XWindowAttributes *win_info;
     XColor **colors;
     Display* theDisp;
{
  int i, ncolors;

  *colors = (XColor *) NULL;

  if (win_info->visual->class == TrueColor) {
    return 0;
  }

  else if (!win_info->colormap) {
    fprintf(stderr,"no colormap associated with window\n");
    return 0;
  }
  ncolors = win_info->visual->map_entries;

  if (!(*colors = (XColor *) malloc (sizeof(XColor) * ncolors))) {
    fprintf(stderr,"malloc failed in getxcolors()");
    return -1;
  }


  if (win_info->visual->class == DirectColor) {
    Pixel red, green, blue, red1, green1, blue1;


    red = green = blue = 0;
    red1   = lowbit(win_info->visual->red_mask);
    green1 = lowbit(win_info->visual->green_mask);
    blue1  = lowbit(win_info->visual->blue_mask);
    for (i=0; i<ncolors; i++) {
      (*colors)[i].pixel = red|green|blue;
      (*colors)[i].pad = 0;
      red += red1;
      if (red > win_info->visual->red_mask)     red = 0;
      green += green1;
      if (green > win_info->visual->green_mask) green = 0;
      blue += blue1;
      if (blue > win_info->visual->blue_mask)   blue = 0;
    }
  }
  else {
    for (i=0; i<ncolors; i++) {
      (*colors)[i].pixel = i;
      (*colors)[i].pad = 0;
    }
  }

  XQueryColors(theDisp, win_info->colormap, *colors, ncolors);

  return(ncolors);
}
    
union swapun {
  CARD32 l;
  CARD16 s;
  CARD8  b[sizeof(CARD32)];
};
/**************************************/
int convertImage(image, colors, ncolors, xwap,theVisual)
     XImage *image;
     XColor *colors;
     int     ncolors;
     XWindowAttributes *xwap;
     Visual* theVisual;
{
  /* attempts to conver the image from whatever weird-ass format it might
     be in into something E-Z to deal with (either an 8-bit colormapped
     image, or a 24-bit image).  Returns '1' on success. */

  /* this code owes a lot to 'xwdtopnm.c', part of the pbmplus package,
     written by Jef Poskanzer */

  int             i, j;
  CARD8          *bptr, tmpbyte;
  CARD16         *sptr, sval;
  CARD32         *lptr, lval;
  CARD8          *pptr, *lineptr;
  int            bits_used, bits_per_item, bit_shift, bit_order;
  int            bits_per_pixel, byte_order;
  CARD32         pixvalue, pixmask, rmask, gmask, bmask;
  int            rshift=0,gshift=0, bshift=0, r8shift=0, g8shift=0, b8shift=0;
  CARD32         rval, gval, bval;
  union swapun   sw;
  int            isLsbMachine, flipBytes;
  Visual         *visual;
  char            errstr[256];


  /* quiet compiler warnings */
  sval = 0;
  lval = 0;
  bit_shift = 0;
  pixvalue  = 0;
  rmask  = gmask  = bmask = 0;
  rshift = gshift = bshift = 0;

  grabPic=NULL;

  /* determine byte order of the machine we're running on */
  sw.l = 1;
  isLsbMachine = (sw.b[0]) ? 1 : 0;

  if (xwap && xwap->visual) visual = xwap->visual;
                       else visual = theVisual;

  if (image->bitmap_unit != 8 && image->bitmap_unit != 16 &&
      image->bitmap_unit != 32) {
    sprintf(errstr, "%s\nReturned image bitmap_unit (%d) non-standard.",
	    "Can't deal with this display.", image->bitmap_unit);
    fprintf(stderr," %s \n",errstr);
    return 0;
  }

  if (!ncolors && visual->class != TrueColor) {
    sprintf(errstr, "%s\nOnly TrueColor displays can have no colormap.",
	    "Can't deal with this display.");
    fprintf(stderr," %s \n",errstr);
    return 0;
  }


  /* build the 'global' grabPic stuff */
  gWIDE = image->width;  gHIGH = image->height;

  if (visual->class == TrueColor || visual->class == DirectColor ||
      ncolors > 256) {
    grabPic = (unsigned char *) malloc((size_t) gWIDE * gHIGH * 3);
    if(grabPic==NULL) {
      perror(" could not malloc ");
      return 0;
    }
    gbits = 24;
  }
  else {
    grabPic = (unsigned char *) malloc((size_t) gWIDE * gHIGH);
    if(grabPic==NULL) {
      perror(" could not malloc ");
      return 0;
    }
    gbits = 8;

    /* load up the colormap */
    for (i=0; i<ncolors; i++) {
      grabmapR[i] = colors[i].red   >> 8;
      grabmapG[i] = colors[i].green >> 8;
      grabmapB[i] = colors[i].blue  >> 8;
    }
  }
  
  if (!grabPic) fprintf(stderr,"unable to malloc grabPic in convertImage()");
  pptr = grabPic;


  if (visual->class == TrueColor || visual->class == DirectColor) {
    unsigned int tmp;

    /* compute various shifty constants we'll need */
    rmask = image->red_mask;
    gmask = image->green_mask;
    bmask = image->blue_mask;

    rshift = lowbitnum((unsigned long) rmask);
    gshift = lowbitnum((unsigned long) gmask);
    bshift = lowbitnum((unsigned long) bmask);

    r8shift = 0;  tmp = rmask >> rshift;
    while (tmp >= 256) { tmp >>= 1;  r8shift -= 1; }
    while (tmp < 128)  { tmp <<= 1;  r8shift += 1; }

    g8shift = 0;  tmp = gmask >> gshift;
    while (tmp >= 256) { tmp >>= 1;  g8shift -= 1; }
    while (tmp < 128)  { tmp <<= 1;  g8shift += 1; }

    b8shift = 0;  tmp = bmask >> bshift;
    while (tmp >= 256) { tmp >>= 1;  b8shift -= 1; }
    while (tmp < 128)  { tmp <<= 1;  b8shift += 1; }

  }


  bits_per_item = image->bitmap_unit;
  bits_used = bits_per_item;
  bits_per_pixel = image->bits_per_pixel;

  if (bits_per_pixel == 32) pixmask = 0xffffffff;
  else pixmask = (((CARD32) 1) << bits_per_pixel) - 1;

  bit_order = image->bitmap_bit_order;
  byte_order = image->byte_order;

  /* if we're on an lsbfirst machine, or the image came from an lsbfirst
     machine, we should flip the bytes around.  NOTE:  if we're on an
     lsbfirst machine *and* the image came from an lsbfirst machine, 
     *don't* flip bytes, as it should work out */

  /* pity we don't have a logical exclusive-or */
  flipBytes = ( isLsbMachine && byte_order != LSBFirst) ||
              (!isLsbMachine && byte_order == LSBFirst);

  for (i=0; i<image->height; i++) {
    lineptr = (unsigned char *) image->data + (i * image->bytes_per_line);
    bptr = ((CARD8  *) lineptr) - 1;
    sptr = ((CARD16 *) lineptr) - 1;
    lptr = ((CARD32 *) lineptr) - 1;
    bits_used = bits_per_item;

    for (j=0; j<image->width; j++) {
      
      /* get the next pixel value from the image data */

      if (bits_used == bits_per_item) {  /* time to move on to next b/s/l */
	switch (bits_per_item) {
	case 8:  bptr++;  break;
	case 16: sptr++;  sval = *sptr;
	         if (flipBytes) {   /* swap CARD16 */
		   sw.s = sval;
		   tmpbyte = sw.b[0];
		   sw.b[0] = sw.b[1];
		   sw.b[1] = tmpbyte;
		   sval = sw.s;
		 }
	         break;
	case 32: lptr++;  lval = *lptr;
	         if (flipBytes) {   /* swap CARD32 */
		   sw.l = lval;
		   tmpbyte = sw.b[0];
		   sw.b[0] = sw.b[3];
		   sw.b[3] = tmpbyte;
		   tmpbyte = sw.b[1];
		   sw.b[1] = sw.b[2];
		   sw.b[2] = tmpbyte;
		   lval = sw.l;
		 }
	         break;
	}
		   
	bits_used = 0;
	if (bit_order == MSBFirst) bit_shift = bits_per_item - bits_per_pixel;
	                      else bit_shift = 0;
      }

      switch (bits_per_item) {
      case 8:  pixvalue = (*bptr >> bit_shift) & pixmask;  break;
      case 16: pixvalue = ( sval >> bit_shift) & pixmask;  break;
      case 32: pixvalue = ( lval >> bit_shift) & pixmask;  break;
      }

      if (bit_order == MSBFirst) bit_shift -= bits_per_pixel;
                            else bit_shift += bits_per_pixel;
      bits_used += bits_per_pixel;

      
      /* okay, we've got the next pixel value in 'pixvalue' */
      
      if (visual->class == TrueColor || visual->class == DirectColor) {
	/* in either case, we have to take the pixvalue and 
	   break it out into individual r,g,b components */
	rval = (pixvalue & rmask) >> rshift;
	gval = (pixvalue & gmask) >> gshift;
	bval = (pixvalue & bmask) >> bshift;

	if (visual->class == DirectColor) {
	  /* use rval, gval, bval as indicies into colors array */

	  *pptr++ = (rval < ncolors) ? (colors[rval].red   >> 8) : 0;
	  *pptr++ = (gval < ncolors) ? (colors[gval].green >> 8) : 0;
	  *pptr++ = (bval < ncolors) ? (colors[bval].blue  >> 8) : 0;
	}

	else {   /* TrueColor */
	  /* have to map rval,gval,bval into 0-255 range */
	  *pptr++ = (r8shift >= 0) ? (rval << r8shift) : (rval >> (-r8shift));
	  *pptr++ = (g8shift >= 0) ? (gval << g8shift) : (gval >> (-g8shift));
	  *pptr++ = (b8shift >= 0) ? (bval << b8shift) : (bval >> (-b8shift));
	}
      }

      else { /* all others: StaticGray, StaticColor, GrayScale, PseudoColor */
	/* use pixel value as an index into colors array */

	if (pixvalue >= ncolors) {
	  fprintf(stderr,"convertImage(): pixvalue >= ncolors");
	}

	if (gbits == 24) {   /* too many colors for 8-bit colormap */
	  *pptr++ = (colors[pixvalue].red)   >> 8;
	  *pptr++ = (colors[pixvalue].green) >> 8;
	  *pptr++ = (colors[pixvalue].blue)  >> 8;
	}
	else *pptr++ = pixvalue & 0xff;

      }
    }
  }

  return 1;
}
static Window TryChildren PARM((Display *, Window, Atom));

/* Find a window with WM_STATE, else return '0' */

static Window xvClientWindow (dpy, win)
    Display *dpy;
    Window win;
{
    Atom WM_STATE;
    Atom type = None;
    int format;
    unsigned long nitems, after;
    unsigned char *data;
    Window inf;

    WM_STATE = XInternAtom(dpy, "WM_STATE", True);
    if (!WM_STATE) return win;

    XGetWindowProperty(dpy, win, WM_STATE, 0L, 0L, False, AnyPropertyType,
		       &type, &format, &nitems, &after, &data);
    if (type) return win;

    inf = TryChildren(dpy, win, WM_STATE);

    return inf;
}

static Window TryChildren (dpy, win, WM_STATE)
    Display *dpy;
    Window win;
    Atom WM_STATE;
{
    Window root, parent;
    Window *children;
    unsigned int nchildren;
    unsigned int i;
    Atom type = None;
    int format;
    unsigned long nitems, after;
    unsigned char *data;
    Window inf = 0;

    if (!XQueryTree(dpy, win, &root, &parent, &children, &nchildren))
	return 0;

    for (i = 0; !inf && (i < nchildren); i++) {
	XGetWindowProperty(dpy, children[i], WM_STATE, 0L, 0L, False,
			   AnyPropertyType, &type, &format, &nitems,
			   &after, &data);
	if (type)
	  inf = children[i];
    }

    for (i = 0; !inf && (i < nchildren); i++)
      inf = TryChildren(dpy, children[i], WM_STATE);

    if (children) XFree((char *)children);
    return inf;
}

/*
 * determine the region to grab and its dimensions 
 */
int x11_grab_region(theDisp, win, rootW, x, y, w, h )
     Display* theDisp;
     Window win,rootW;
     int  *x,*y,*w,*h;
     
{
  int ix=0,iy=0,ih=0,iw=0;
  Window rW, cW;
  GC xorGC;
  int rx,ry,x1,y1; 
  unsigned int rootx, rooty;
  XEvent event;
  int maskno=0;
  u_long        xorMasks[8];

  xorGC = XCreateGC(theDisp, rootW, 0, NULL);
  XSetSubwindowMode(theDisp, xorGC, IncludeInferiors);
  XSetForeground(theDisp, xorGC, -1);
  XSetFunction(theDisp, xorGC, GXxor);
  XMaskEvent(theDisp, ButtonPressMask, &event);
  rootx = event.xbutton.x_root;
  rooty = event.xbutton.y_root;

  fprintf(stderr," x=%i y=%i \n",*x,*y);
  while (1) {  /* wait for button to be released, if clicked */
    unsigned int mask;

    if (XQueryPointer(theDisp,rootW,&rW,&cW,&rx,&ry,&x1,&y1,&mask)) {
      if (!(mask & Button1Mask)) break;
    }
    ix = (*x<rx) ? *x : rx;
    iy = (*y<ry) ? *y : ry;
    iw=abs(x1-*x);
    ih=abs(y1-*y);
    maskno = (maskno + 1) & 6;
    XSetPlaneMask(theDisp, xorGC, xorMasks[maskno]);
    XDrawRectangle(theDisp, rootW, xorGC, ix,iy,iw,ih);
    XFlush(theDisp); 
    
    maskno=0;
    XSetPlaneMask(theDisp, xorGC, xorMasks[maskno]);
    XDrawRectangle(theDisp,rootW,xorGC,ix,iy,iw,ih);
    XFlush(theDisp); 
  }

  XFreeGC(theDisp,xorGC);
  fprintf(stderr," x1=%i y1=%i \n",x1,y1);
  *w=iw;
  *h=ih;
  *x=ix;
  *y=iy;
  return rootW;
  
}

void ungrabX(Display* theDisp) {
  XUngrabServer(theDisp);
  XUngrabPointer(theDisp,CurrentTime);
}

/*
 * determine the window to grab and its dimensions 
 */
int x11_grab_frame(theDisp, win, rootW, x, y, w, h )
     Display* theDisp;
     Window win,rootW;
     int  *x,*y,*w,*h;
     
{
  int ix,iy,ih,iw;
  Window clickWin;
  Window rW, cW;
  int rx,ry,x1,y1; 

  while (1) {  /* wait for button to be released, if clicked */
    unsigned int mask;
    if (XQueryPointer(theDisp,rootW,&rW,&cW,&rx,&ry,&x1,&y1,&mask)) {
      if (!(mask & Button2Mask)) break;
    }
  }
  if (!win || win == rootW) clickWin = rootW;
  else {
    int xr, yr;    Window chwin=None;
    XTranslateCoordinates(theDisp, rW, cW, rx, ry, &xr, &yr, &chwin);
    if (chwin != None) {
      XWindowAttributes clickxwa, parentxwa;

      clickWin = xvClientWindow(theDisp, chwin);
      
      /* decide if we want to just grab clickWin, or cW.
	 basically, if they're different in any important way 
	 (depth, visual, colormap), grab 'clickWin' only, 
	 as it's the important part */
      if (!clickWin || 
	  (XGetWindowAttributes(theDisp, clickWin, &clickxwa)  &&
	   XGetWindowAttributes(theDisp, cW,       &parentxwa) &&
           clickxwa.visual->class == parentxwa.visual->class   &&
	   clickxwa.colormap      == parentxwa.colormap        &&
	   clickxwa.depth         == parentxwa.depth)
	  )
	clickWin = cW;   	  /* close enough! */
    }
    else clickWin = cW;
    
  }
  
  if (clickWin == rootW) {   /* grab entire screen */
    ix = iy = 0;  iw = dispWIDE;  ih = dispHIGH;
  }
  else {
    int x,y;  Window win;   unsigned int rw,rh,rb,rd;
    if (XGetGeometry(theDisp,clickWin,&rW, &x, &y, &rw, &rh, &rb, &rd)) {
      iw = (int) rw;  ih = (int) rh;
      XTranslateCoordinates(theDisp, clickWin, rootW, 0, 0, &ix,&iy, &win);
      
    }
    else {
      ix = iy = 0;  iw = dispWIDE;  ih = dispHIGH;  clickWin = rootW;
    }
  }
  /* always take the 0,0 coordinates for the frame */
  *x=0;
  *y=0;
  /* give the width and height up */
  *w=(ix+iw>dispWIDE)?dispWIDE-ix-10:iw;
  *h=(iy+ih>dispHIGH)?dispHIGH-iy-10:ih;
  return(clickWin);
}


/* 
 * initialize the needed window information such as the root window name, 
 * the display and grab the server.
 * button 1 grab a region
 * button 2 grab a window
 * button 3 cancel
 *
 * after the grabing get the right colors and write t oa gif file 
 */


int x11_grab(char *name,int ssrc)
{
  XImage *image=NULL;
  XWindowAttributes xwa;
  XColor *colors;
  Display* theDisp;
  Tk_Window mainWindow;
  int ncolors;
  Window mainXW, clickWin,rootW;
  char str[100];
  int           ix, iy, iw, ih;
  FILE *fp;
  char Name[100];


  Window rW,cW;
  unsigned int mask;
   
  Cursor cursor;

  /* initilize the name */
  sprintf(Name,"/tmp/%s",name);
  /* initialize the diplay and main window values */ 
  if((theDisp=XOpenDisplay(NULL))==NULL){
    perror("could not open display: x11_grab \n");
    return 0;
  }


  /*  mainXW=Tk_WindowId(mainWindow);*/
  rootW=DefaultRootWindow(theDisp);


  /* set the cursur */
  cursor = XCreateFontCursor(theDisp, XC_cross);
  if (XGrabPointer(theDisp, rootW, False, ButtonPressMask, GrabModeAsync,
		   GrabModeAsync, rootW, cursor, CurrentTime)!=GrabSuccess) {
    perror(" could not grab mouse \n");
    XCloseDisplay(theDisp);
    return (-1);
  }
  /* get the dimensions of the screen */
  dispWIDE=DisplayWidth(theDisp,DefaultScreen(theDisp));
  dispHIGH=DisplayHeight(theDisp,DefaultScreen(theDisp));


  XGrabServer(theDisp);	 /* until we've done the grabImage */
  while (1) {      /* wait for a button action  */
    if (XQueryPointer(theDisp,rootW,&rW,&cW,&ix,&iy,&ix,&iy,&mask)) {
      if ((mask & (Button1Mask | Button2Mask | Button3Mask))) break;
    }
  }
  if (mask & Button3Mask) {
    ungrabX(theDisp);
    XCloseDisplay(theDisp);
    return 0;
  }
  if (mask & Button2Mask) { /* grab a window */
    clickWin=x11_grab_frame(theDisp,cW,rootW,&ix,&iy,&iw,&ih);
  }
  if (mask & Button1Mask) { /* grab a region*/
    clickWin=x11_grab_region(theDisp,cW,rootW,&ix,&iy,&iw,&ih);
  }
  if(iw!=0||ih!=0) {
    if((image=XGetImage(theDisp, clickWin, ix, iy, (u_int) iw, (u_int) ih, 
			AllPlanes, ZPixmap))==NULL) {
      perror("can not get image");
      Exit();
    }
  } else {
    ungrabX(theDisp);
    XCloseDisplay(theDisp);
    return 0;
  }

 
  if (!XGetWindowAttributes(theDisp, clickWin, &xwa)) {
    sprintf(str,"Unable to get window attributes for clicked-on window\n");
    fprintf(stderr,"%s That Sucks!\n",str);
    XCloseDisplay(theDisp);
    return 0;
  }
  ungrabX(theDisp);
  
  if((ncolors = getxcolors(&xwa, &colors,theDisp))<0){
    XCloseDisplay(theDisp);
    return 0;
  }

  XBell(theDisp, 0);    /* beep twice at end of grab */
  XBell(theDisp, 0);

  
  if(convertImage(image, colors, ncolors, &xwa,
		  DefaultVisual(theDisp,DefaultScreen(theDisp)))==0){
    fprintf(stderr," could not convert Image \n");
    XCloseDisplay(theDisp);
    return 0;
  }

  /* DO *NOT* use xvDestroyImage(), as the 'data' field was alloc'd by X, not
     necessarily through 'malloc() / free()' */
  XDestroyImage(image);   
  
  if (colors) free((char *) colors);
  if ((fp = fopen(Name, "w"))==NULL){
    XCloseDisplay(theDisp);
   return 0;
  }
  fprintf(stderr," w=%i h=%i \n",gWIDE,gHIGH);
  WriteGIF(fp,grabPic,gbits,gWIDE,gHIGH,grabmapR,grabmapG,grabmapB,256,8,"");
  fclose(fp);
  if(grabPic!=NULL) free(grabPic); grabPic=NULL;
  {
    static char file[100000];
    FILE* fp;
    static size_t length;
    
    fp =fopen(Name,"r"); 
    length=fread(file,(size_t)1,(size_t)100000,fp);
    fclose(fp);
  
    fprintf(stderr," len=%i \n",length);
    GIF_encode(file,length,ssrc);
  }
  XCloseDisplay(theDisp);
  return 1;
}


