#ifndef __tsap_h
#define __tsap_h
#include "types.h"
#include <sys/socket.h>
#include <netinet/in.h>

#define TSAP_SIN(t)  ((struct sockaddr_in *)t.addr.buf)
#define TSAP_SUN(t)  ((struct sockaddr_un *)t.addr.buf)
#define TSAP_SINP(t) ((struct sockaddr_in *)(t)->addr.buf)
#define TSAP_SUNP(t) ((struct sockaddr_un *)(t)->addr.buf)
#define TSAP_SAP(t)  ((struct sockaddr *)(t)->addr.buf)

#if !defined(AF_OSI) && defined(AF_CCITT)
#define AF_OSI AF_CCITT
#endif
/* if undefined, it doesn't matter */
#if !defined(AF_OSI) && defined(AF_UNSPEC)
#define AF_OSI 9999
#endif

#ifndef _nettli_tiuser_h
struct netbuf {
  unsigned int maxlen;
  unsigned int len;
  char *buf;
};
#endif

typedef struct {
  enum {T_unknown, T_sockaddr = AF_INET, T_unix = AF_UNIX, T_osi = AF_OSI} family;   /* address type */
  struct netbuf addr;                                 /* pointer to address buffer */
} tsap_t;

extern void tsap_alloc (tsap_t *, int);
extern void tsap_free  (tsap_t *);

extern void    tsap_cpy   (tsap_t *, tsap_t *);
extern int     tsap_equ   (tsap_t *, tsap_t *);
extern int     tsap_equ_a (tsap_t *, tsap_t *);
extern char   *tsap_ntoa  (tsap_t *);
extern char   *tsap_gethostbyaddr (tsap_t *);
extern void    tsap_gethostaddress (int family, tsap_t *tsap);
extern void    tsap_setport (tsap_t *, u_int32);
extern int     tsap_getport (tsap_t *);
extern int     tsap_isbroadcast (tsap_t *);
extern int     tsap_ismulticast (tsap_t *);
extern int     tsap_gethostbyname (char *name, tsap_t adrlist[], int listlen);
extern int     tsap_pullup (char *buf, tsap_t *, int *len);
extern u_int32 tsap_ip4(tsap_t *tsap);
extern char   *tsap_sprintf(char *buf, char *fmt, tsap_t *t);
#endif
