/*
crypt  ---  generate key and perform encryption/decryption.

Each block is coded independently to avoid error propagation through
lost packets.  On a SPARCstation II, the encryption and decryption
each take about 4.1 ms per 200 bytes.  Use stream cipher.

Copyright 1993 by AT&T Bell Laboratories; all rights reserved
*/

#include "types.h"      /* u_int32 */
#include <stdio.h>      /* sscanf() */
#include <ctype.h>
#include <string.h>     /* strlen(), memcpy() */
#include <ansi.h>

#define ENCRYPT (1)
#define DECRYPT (-1)



/*
* The key string may contain the usual C backslash-escaped characters.
* Each byte in a DES key has to have odd parity. We enforce that
* silently by toggling the 8th (msb) bit as necessary.
* The key schedule is written to the matrix ks.
* Returns 0 on success, -1 on error (bad checksum).
*/
int crypt_key(char *key, u_int32 ks[16][4])
{
  u_char inkey[8];
  int v;
  int i, j;
  int parity;

  for (i = 0; i < 8; i++) inkey[i] = 0x80; /* odd parity! */

  for (i = 0; *key && i < 8; key++) {
    if (*key == '\\') {
      key++;
      if (*key == 'n') inkey[i++] = '\n';
      else if (*key == 'r') inkey[i++] = '\r';
      else if (*key == 't') inkey[i++] = '\t';
      else if (*key == 'v') inkey[i++] = '\v';
      else if (*key == 'b') inkey[i++] = '\b';
      else if (*key == 'f') inkey[i++] = '\f';
      else if (*key == '\\') inkey[i++] = '\\';
      else if (*key == '\?') inkey[i++] = '\?';
      else if (*key == '\'') inkey[i++] = '\'';
      else if (*key == '\"') inkey[i++] = '\"';
      else if (isdigit(*key)) {
        sscanf(key, "%o", &v);
        inkey[i++] = v;
        v = strlen(key);
        key += (v >= 3) ? 3 : v;
      } 
      else inkey[i++] = *key;
    }
    else inkey[i++] = *key;
  }

  /* enforce parity */
  for (i = 0; i < 8; i++) {
    v = inkey[i];
    for (j = 0, parity = 0; j < 8; j++) {  /* 8 bits in a DES byte */
      if (v & 01) parity++;
      v >>= 1;  /* shift one bit */
    }
    if ((parity & 01) == 0) inkey[i] ^= 0x80;
  }

#if DES
  if (ksmake(inkey, ks) != 0) return -1;
#endif
  return 0;
} /* crypt_key */


/*
* If in and out point to the same place, encryption and decryption are
* done in-place.
*
* The buffer must be a multiple of 8 bytes. If not, the leftover bytes
* are not encrypted.
*/
void crypt(int encrypt, u_int32 ks[16][4], char *in, char *out, int len)
{
#if DES
  int i;
#endif

  if (in != out) memcpy((void *)out, (void *)in, (size_t)len);
  len &= 0xfffffff8;
#if DES
  if (encrypt) {
    for (i = 0; i < len; i += 8) do_des_inplace(&out[i], ks, ENCRYPT);
  }
  else {
    for (i = 0; i < len; i += 8) do_des_inplace(&out[i], ks, DECRYPT);
  }
#endif
} /* crypt */
