/*
l16  --  Convert between L16 and other audio formats

Copyright 1993 by AT&T Bell Laboratories; all rights reserved
*/

#include "types.h"
#include "audio_generic.h"
#include "pcmu_l.h"
#include "l16.h"



/*
* Initialize L16 codec state. Dummy function.
*/
void *l16_init(void *v, double frame)
{
  return 0;
} /* l16_init */


/*
* Translate from L16 to output format specified by 'ad'.
* Does not handle stereo/mono conversion.
*/
int l16_decode(void *in_buf, int in_size, audio_descr_t *ad, 
  void *out_buf, int *out_size, void *state)
{
  int16 *in_s   = (int16 *)in_buf;
  u_int8 *out_c = (u_int8 *)out_buf;
  int i;

  switch(ad->encoding) {  /* input format */
    case AE_PCMU:
      in_size /= 2;
      for (i = 0; i < in_size; i++) out_c[i] = l16_pcmu(in_s[i]);
      *out_size = in_size;
      break;

    default:
      return -1;
  }

  return 0;
} /* l16_decode */


/*
* Translate from input format described by 'ad' to L16.
* Does not handle stereo/mono conversion.
*/
int l16_encode(void *in_buf, int in_size, audio_descr_t *ad,
  void *out_buf, int *out_size, void *state, int header_flag)
{
 int i;
 u_int8 *in_c  = (u_int8 *)in_buf;
 int16  *out_s = (int16 *)out_buf;

  switch(ad->encoding) {  /* input format */
    case AE_PCMU:
      for (i = 0; i < in_size; i++) out_s[i] = pcmu_l16(in_c[i]);
      *out_size = in_size * 2;
      break;

    default:
      return -1;
  }
  return 0;
} /*  l16_encode */
