/*
cmd  --  parse commands

Copyright 1995 by Henning Schulzrinne; all rights reserved
*/
#include <tk.h>
#include <tcl.h>
#include "audio_generic.h"
#include "ahw.h"      /* AS_fillpoint */
#include "nevot.h"
#include "control.h"
#include "list.h"     /* list_*() */
#include "debug.h"
#include "util.h"     /* open_file(), tcl() */
#include <stdlib.h>   /* atoi(), malloc() */
#include <string.h>   /* strcpy() */
#include <memory.h>   /* memcpy() */
#include <math.h>     /* exp() */
#include <stdarg.h>   /* va_start */
#include <ansi.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <signal.h>



extern int errno;
extern int optind;
extern char *optarg;
extern char *sys_errlist[];
static char *version =
#include "version.h"
;

static struct sockaddr_in my_addr, serv_addr;
extern struct in_addr host2ip(char *);
static Audio_session_t session;    /* add similarity to nevit_main */

/*
 * get the local address of the calling host, or the address of the interface
 * from which the user whiches to send its data.
 */

int get_myaddr(struct sockaddr_in* lmy_addr,char* local)
{
  my_addr.sin_family = AF_INET;
  if(local!=NULL&&*local!=0) {  
    struct hostent *name;
    name= gethostbyname(local);
    if (name == NULL) {
      fprintf(stderr, "Can't resolve host using default address\n");
    } 
    else {
      memcpy((char *)&(my_addr.sin_addr),name->h_addr,name->h_length);
      Audio_local(local);    
      return 1;
    }
  }
  {
    /* no interface was specified so use the default one */    
    char *name;
    name=idstring("%d",100);
    my_addr.sin_addr=host2ip(name);
    return 1;
  }
}

extern Tcl_Interp *interp2;

extern int getoptions(int argc, char *argv[], char *opt);
extern int cmd_audioctl (ClientData client, Tcl_Interp *interp, int argc, char *argv[]);

static list_map_t rt_proto_map[] = {
  {"VAT",  R_vat},
  {"RTP",  R_rtp},
  {0,      -1}
};
static list_map_t encoding_map[] = {
  {"PCMU", AE_PCMU},
  {"PCMA", AE_PCMA},
  {"G721", AE_G721},
  {"DVI4", AE_DVI4},
  {"G723", AE_G723},
  {"GSM",  AE_GSM},
  {"1016", AE_1016},
  {"LPC",  AE_LPC},
  {"L8",   AE_L8},
  {"L16",  AE_L16},
  {"L24",  AE_L24},
  {"L32",  AE_L32},
  {0,      -1}
};


static char *duplex_list[] =
  {"duplex", "net", "local", 0};


/*
* Return printable representation of audio setting.
*/
static char *audio_name(audio_descr_t *ad)
{
  static char buf[80];

  sprintf(buf, "%s %d %d",
    list_value2key(encoding_map, ad->encoding), ad->channels, ad->sample_rate);
  return buf;
} /* audio_name */


/*
* Trace handler for general Nevot variables.
* Also modifies all variables that depend on them.
*/
static char *cmd_general(ClientData client, Tcl_Interp *interp,
  char *name1, char *name2, int flags)
{
  char *value;                /* Tcl variable value */
  int v;                      /* value */
  int i, j;
  int mask;
  char **listArgv;
  int listArgc;

  static char *r_list[] = {"end_node", "gateway", 0};
  static char *a_list[] = {"on_demand", "idle", "idle_demand", "never", 0};
  static struct {
    char *name;
    int  mask;
  } v_list[] = {
    {"rx",    DEBUG_RX},
    {"tx",    DEBUG_TX},
    {"audio", DEBUG_AUDIO},
    {"sd",    DEBUG_SD},
    {"ts",    DEBUG_TS},
    {"cntl",  DEBUG_CNTL},
    {0,0}
  };

  value = Tcl_GetVar2(interp, name1, name2,
    (flags & TCL_GLOBAL_ONLY) | TCL_LEAVE_ERR_MSG);
  if (!value) return interp->result;
  v = atoi(value);
  switch((long)client) {
    case 't':  /* audio_chunk */
      nevot.audio.tick = atof(value);
      audio_chunk();
      break;

    case 'a':  /* audio_priority */
      nevot.audio_priority = list_to_index(a_list, value);
      break;

    case 'd': /* debug output file */
      debug_file(value);
      break;

    case 'r':  /* role */
      nevot.role = list_to_index(r_list, value);
      break;

    case 'v': /* debug */
      if (Tcl_SplitList(interp, value, &listArgc, &listArgv) != TCL_OK) {
        return "can't split list";
      }
      mask = 0;
      for (i = 0; i < listArgc; i++) {
        for (j = 0; v_list[j].name; j++) {
          if (strcmp(v_list[j].name, listArgv[i]) == 0) {
            mask |= v_list[j].mask;
            break;
          }
        }
      }
      free((char *)listArgv);
      debug_mask(mask);
      break;
  }
  return TCL_OK;
} /* cmd_general */


/*
* play file_name
* play -pause
* play -stop
* play
* Any command can take the flags:
*   [-loop]    [-noloop]
*   [-monitor] [-nomonitor]
*/
static int cmd_play(ClientData ch, Tcl_Interp *interp, int argc, char *argv[])
{
  int fd;     /* file descriptor for audio file */
  int result = TCL_OK;
  int i;      /* command line loop */

  for (i = 1; i < argc; i++) {
    if (*argv[i] == '-') {
      if      (strcmp(argv[i], "-loop")      == 0) tx_file_loop(1);
      else if (strcmp(argv[i], "-noloop")    == 0) tx_file_loop(0);
      else if (strcmp(argv[i], "-monitor")   == 0) tx_file_monitor(1);
      else if (strcmp(argv[i], "-nomonitor") == 0) tx_file_monitor(0);
      else if (strcmp(argv[i], "-pause")     == 0) tx_file_pause(1);
      else if (strcmp(argv[i], "-stop")      == 0) tx_file_stop();
      else if (strcmp(argv[i], "-resume")    == 0) tx_file_pause(0);
    }
    else if (strlen(argv[i]) > 0) break;
  }

  /* file name: play */
  if (i == argc-1) {
    tx_file_stop();
    fd = open_file((char *)NULL, argv[i], ".au", "r", (char **)0);
    if (fd >= 0) {
      audio_file_t af;

      if (tx_file(fd, &af) == Audio_OK) {
        sprintf(interp->result, "%s (%1.1f)",
          af.info ? af.info : "",
          audio_bytes_to_seconds(&af.ad, af.data_size));
      }
      else {
        sprintf(interp->result, "Cannot play audio without session");
        result = TCL_ERROR;
      }
    }
    else {
      sprintf(interp->result, "play %s: %s", argv[i], sys_errlist[errno]);
      result = TCL_ERROR;
    }
  }
  else if (i < argc-1) {
    sprintf(interp->result, "play: only one file name allowed - %s",
      argv[i]);
    result = TCL_ERROR;
  }
  return result;
} /* cmd_play */


/*
* record
* record [-append] [-info "info string"] file_name
* record [-stop|-pause|-resume]
*
* Called recursively. Returns TCL_OK
*
* ch:      in: identifies command
* interp:  in: TCL interpreter
* argc:    in: number of words
* argv:    in: command words
*/
static int cmd_record(ClientData ch, Tcl_Interp *interp,
  int argc, char *argv[])
{
  int fd;   /* file descriptor for audio file */
  int result = TCL_OK;
  int append = FALSE;
  int c;
  char *info = 0;
  char *chosen;

  optind = 1;
  while ((c = getoptions(argc, argv, "ai:spr")) != -1) {
    switch(c) {
    case 'a':
      append = TRUE;
      break;
    case 'i':
      info = optarg;
      break;
    case 's':
      rx_file_stop();
      return result;
    case 'r':
      rx_file_pause(0);
      return result;
    case 'p':
      rx_file_pause(1);
      return result;
    }
  }

  if (optind == argc) rx_file_pause(0);
  /* one file name: start recording */
  else if (optind == argc-1) {
    rx_file_stop();
    fd = open_file(".", argv[optind], ".au", append ? "a" : "w", &chosen);
    if (fd >= 0) rx_file(fd, info, chosen);
    else {
      sprintf(interp->result, "%s: %s", argv[0], sys_errlist[errno]);
      result = TCL_ERROR;
    }
  }
  else {
    sprintf(interp->result, "record: only one file name allowed - %s",
      argv[0]);
    result = TCL_ERROR;
  }
  return result;
} /* cmd_record */

/*
* Set session parameters of new or existing session.
*   0   1                2
*   s.t transport        RTP|VAT
*   s.t network          UDP|TCP|ST-II
*   s.t key              [key]
*   s.t ttl              value
*   s.t media            encoding channels sampling_rate
*   s.t frame            frame duration
*   s.t min_delay        minimum delay, in seconds
*   s.t max_delay        maximum delay, in seconds
*   s.t initial_delay    initial delay, in seconds
*   s.t delay_multiplier variance multiplier
*   s.t delay_smooth     delay variance smoothing
*   s.t confid           #   (vat only)
*   s.t port             #
*   s.t cname            CNAME to be sent
*   s.t name             NAME to be sent
*   s.t email            EMAIL to be sent
*   s.t phone            PHONE to be sent
*   s.t loc              LOC to be sent
*   s.t img              IMG to be sent (file name or URL)
*   s.t note             NOTE to be sent
*   s.t encrypt          0|1
*   s.t mute_new         0|1
*   s.t receive          0|1
*   s.t send             0|1
*   s.t open             [host1 host2 ...]
*   s.t play             [-options] filename to be played
*   s.t close
*   s.t statistics?
*   s.t error       op (is this needed?)
*
*  Ignore others (our own).
*  [#] not changeable once session is established
*/
static int cmd_session(ClientData client, Tcl_Interp *interp,
  int argc, char *argv[])
{
  audio_descr_t audio;
  Audio_session_t s = (Audio_session_t)client;
  Audio_sp_t *sp = &s->sp;
  int status = TCL_OK;
  int value;
  int i;

  session=s;

  if (argc < 2) {
    sprintf(interp->result, "wrong # args: should be \"%s Op\"",
      argv[0]);
    return TCL_ERROR;
  }

  if (strcmp(argv[1], "cname") == 0) {
    char *cname=idstring("%u@%i",50);
    status = Audio_SessionSetSDES(s, SDES_CNAME, cname);
  }
  if (strcmp(argv[1], "local") == 0) { /* decide which address to use for the 
					  communication in case of multihomed 
					  stations*/
    get_myaddr(&(my_addr), argv[2]);
  }
  else if (strcmp(argv[1], "name") == 0) {
    status = Audio_SessionSetSDES(s, SDES_NAME, argv[2]);
  }
  else if (strcmp(argv[1], "email") == 0) {
    status = Audio_SessionSetSDES(s, SDES_EMAIL, argv[2]);
  }
  else if (strcmp(argv[1], "phone") == 0) {
    status = Audio_SessionSetSDES(s, SDES_PHONE, argv[2]);
  }
  else if (strcmp(argv[1], "loc") == 0) {
    status = Audio_SessionSetSDES(s, SDES_LOC, argv[2]);
  }
  else if (strcmp(argv[1], "tool") == 0) {
    status = Audio_SessionSetSDES(s, SDES_TOOL, argv[2]);
  }
  else if (strcmp(argv[1], "note") == 0) {
    status = Audio_SessionSetSDES(s, SDES_NOTE, argv[2]);
  }
  else if (strcmp(argv[1], "img") == 0) {
    status = Audio_SessionSetSDES(s, SDES_IMG, argv[2]);
  }
  else if (strcmp(argv[1], "ttl") == 0) {
    if (argc != 3) goto args;
    status = Audio_SessionSetTTL(s, atoi(argv[2]));
  }
  else if (strcmp(argv[1], "port") == 0) {
    if (argc != 3) goto args;
    value = atoi(argv[2]);
    sp->port = value;
    serv_addr.sin_port = htons(value);
    my_addr.sin_port= htons(value);
  }
  else if (strcmp(argv[1], "frame") == 0) {
    double v;
    if (argc != 3) goto args;

    v = atof(argv[2]);
    status = Audio_SessionSetAudio(s, 0, 0, v);
    if (status < 0) {
      sprintf(interp->result, "cannot set frame duration to %f", v);
      return TCL_ERROR;
    }
  }
  else if (strcmp(argv[1], "media") == 0) {
    char enc[20];

    /* single list */
    if (argc == 3) {
      sscanf(argv[2], "%s %d %d", enc, &audio.channels, &audio.sample_rate);
    }
    /* individual arguments */
    else if (argc >= 5) {
      strcpy(enc, argv[2]);
      audio.channels    = atoi(argv[3]);
      audio.sample_rate = atoi(argv[4]);
    }
    else goto args;
	   
    audio.encoding    = list_key2value(encoding_map, enc, 1);
    status = Audio_SessionSetAudio(s, &audio, 1, 0.);
    if (status < 0) {
      sprintf(interp->result, "cannot use %s encoding (%s) for %s protocol",
        audio_ntoa(&audio), Audio_strerror(status), 
        list_value2key(rt_proto_map, sp->rt_proto));
      return TCL_ERROR;
    }
  }
  else if (strcmp(argv[1], "min_delay") == 0) {
    if (argc != 3) goto args;
    Tcl_GetDouble(interp, argv[2], &s->sp.delay.min);
    s->force_ts = 1;
  }
  else if (strcmp(argv[1], "max_delay") == 0) {
    if (argc != 3) goto args;
    Tcl_GetDouble(interp, argv[2], &s->sp.delay.max);
    s->force_ts = 1;
  }
  else if (strcmp(argv[1], "initial_delay") == 0) {
    if (argc != 3) goto args;
    Tcl_GetDouble(interp, argv[2], &s->sp.delay.initial);
  }	 
  else if (strcmp(argv[1], "delay_multiplier") == 0) {
    if (argc != 3) goto args;
    Tcl_GetDouble(interp, argv[2], &s->sp.delay.multiplier);
    s->force_ts = 1;
  }

  else if (strcmp(argv[1], "delay_smooth") == 0) {
    double smooth;

    if (argc != 3) goto args;
    Tcl_GetDouble(interp, argv[2], &smooth);
    s->sp.delay.dvar_x = exp(-smooth);
    s->sp.delay.dvar_y = 1 - s->sp.delay.dvar_x;
    s->force_ts = 1;
  }
  else if (strcmp(argv[1], "late_ts") == 0) {
    if (argc != 3) goto args;
    Tcl_GetBoolean(interp, argv[2], &value);
    s->sp.delay.late_ts = value;
    s->force_ts = 1;
  }
  else if (strcmp(argv[1], "immediate") == 0) {
    if (argc != 3) goto args;
    Tcl_GetBoolean(interp, argv[2], &value);
    s->sp.delay.immediate = value;
  }
  else if (strcmp(argv[1], "encrypt") == 0) {
    if (argc != 3) goto args;
    Tcl_GetBoolean(interp, argv[2], &value);
    s->encrypt = value;
  }
  else if (strcmp(argv[1], "mute_new") == 0) {
    if (argc != 3) goto args;
    Tcl_GetBoolean(interp, argv[2], &value);
  }
  else if (strcmp(argv[1], "priority") == 0) {
    if (argc != 3) goto args;
    status = Audio_SessionSetDuplex(s, list_to_index(duplex_list, argv[2]));
  }
  else if (strcmp(argv[1], "key") == 0) {
    status = Audio_SessionSetKey(s, argc == 2 ? 0 : argv[2]);
  }
  else if (strcmp(argv[1], "network") == 0) {
    if (argc != 3) goto args;
    if (strcmp(argv[2], "UDP") == 0) sp->net_proto = P_UDP;
    else {
      sprintf(interp->result,
        "%s %s: undefined network protocol %s", argv[0], argv[1], argv[2]);
      return TCL_ERROR;
    }
  }
  else if (strcmp(argv[1], "transport") == 0) {
    if (argc != 3) goto args;
    value = list_key2value(rt_proto_map, argv[2], 1);
    if (value < 0) {
      status = Audio_EPROTONOSUPPORT;
    }
    else {
      sp->rt_proto = value;
      if      (sp->rt_proto == R_vat) vat_init(&s->rt);
      else if (sp->rt_proto == R_rtp) rtp_init(&s->rt);
      (*s->rt.change_audio)(&s->rt, &s->audio);
    }
  }
  else if (strcmp(argv[1], "confid") == 0) {
    if (argc != 3) goto args;
    sp->confid = atoi(argv[2]);
  }
  else if (strcmp(argv[1], "receive") == 0) {
    if (argc != 3) goto args;
    status = Audio_SessionListen(s, atoi(argv[2]));
  }
  else if (strcmp(argv[1], "send") == 0) {
    if (argc != 3) goto args;
    status = Audio_SessionTalk(s, atoi(argv[2]));
  }
  else if (strcmp(argv[1], "statistics?") == 0) {
    if (argc != 2) goto args;
    sprintf(interp->result, "{created %d} {psent %u %u} {bsent %u}",
      (int)s->local->created,
      (unsigned int)s->packets[0], (unsigned int)s->packets[1],
      (unsigned int)s->bytes);
    tcl(interp, "control_send {%s statistics %s}", s->name,
      interp->result);
  }
  else if (strcmp(argv[1], "open") == 0) {
    if (argc < 2) goto args;
    for (i = 2; i < argc; i++) {
      if ((status = Audio_SessionOpen(s, argv[i])) != Audio_OK) {
        sprintf(interp->result, "cannot connect to %s: %s",
          argv[i], Audio_strerror(status));
        return TCL_ERROR;
      }
    }
  }
  else if (strcmp(argv[1], "close") == 0) {
    if (argc != 2) goto args;
    status = Audio_SessionDelete(s);
    Tcl_DeleteCommand(interp, argv[0]);
    tcl(interp, "control_send {- %s}", argv[0]);
  }
  else if (strcmp(argv[1], "play") == 0) {
    if (cmd_play(0, interp, argc-1, argv+1) != TCL_OK) 
      status = Audio_ECANTREAD;
  }
  else if (strcmp(argv[1], "record") == 0) {
    if (cmd_record(0, interp, argc-1, argv+1) != TCL_OK) 
      status = Audio_ECANTWRITE;
  }
  /* ignore our own packets */

  if (status < 0) {
    sprintf(interp->result, errno ? "%s (%s)" : "%s",
      Audio_strerror(status), sys_errlist[errno]);
  }
  return (status == Audio_OK) ? TCL_OK : TCL_ERROR;

args:
  sprintf(interp->result,
    "wrong # of args (%d): should be \"%s %s ?args?\"",argc, argv[0], argv[1]);
  return TCL_ERROR;
} /* cmd_session */

/*
* Evaluate command for all sessions.
*/
static int cmd_session_all(ClientData client, Tcl_Interp *interp,
  int argc, char *argv[])
{
  Audio_session_t s;

  ALL_SESSIONS(s) {
    cmd_session((ClientData)s, interp, argc, argv);
  }
  return TCL_OK;
} /* cmd_session_all */

/*
* Create media session; create new command with session name.
* Enable pmm reception for that session.
*   0       1
*   session $s.audio.i
*/
static int cmd_session_audio(ClientData client, Tcl_Interp *interp,
  int argc, char *argv[])
{
  Audio_session_t s;
/*printf(" nevot: arguments in cmd_session_audio: %s , %s\n",argv[0],argv[1]);*/
 /* inserted by fim */
  if (argc != 2) {
    sprintf(interp->result, "wrong # of args: should be \"%s Name\"", argv[0]);
    return TCL_ERROR;
  }
  if (strstr(argv[1],"audio") == NULL) return 0; /* inserted by fim */
  s = Audio_SessionFind(argv[1]);
  if (!s) {
    char *slash, *name;

    if (Audio_SessionAlloc(argv[1], 0, &s) < 0) {
      sprintf(interp->result, "%s: cannot create session %s", argv[0], argv[1]);
      return TCL_ERROR;
    }
    s->local_audio_channel = 0;
    Tcl_CreateCommand(interp, s->name, cmd_session, (ClientData)s, 0);
    /* also create aliases for conference commands */
    name = strdup(s->name);
    slash = strchr(name, '/');
    *(slash+1) = '\0';
    Tcl_CreateCommand(interp, name, cmd_session, (ClientData)s, 0);
    *slash = '\0';
    Tcl_CreateCommand(interp, name, cmd_session, (ClientData)s, 0);
    free(name);
    tcl(interp, "control_send {+ %s}", argv[1]);
    tcl(interp, "control_send {+ %s/*}", argv[1]);
  }
  /* whether existing or not, report it */
  (*nevot.event)(s, 0, Audio_created);
  return TCL_OK;
} /* cmd_session_audio */

/*
* Member-specific commands.
*  0     1          2
*  s/t/m receive    0|1
*  s/t/m statistics?
*  s/t/m error      ??     (ignored, should not happen)
*
* Other commands (including our own echo) are ignored.
*/
int cmd_member(ClientData client, Tcl_Interp *interp, int argc, char *argv[])
{
  stats_t *stats;
  Audio_member_t m;

  m = (ClientData)client;
  if (argc < 2) {
    sprintf(interp->result, "wrong # args: should be \"%s Op\"", argv[0]);
    return TCL_ERROR;
  }
  if (strcmp(argv[1], "receive") == 0) {
    if (argc != 3) {
      sprintf(interp->result, "wrong # args: should be %s receive 0|1",
        argv[0]);
      return TCL_ERROR;
    }
    Audio_MemberListen(m, atoi(argv[2]));
  }
  else if (strcmp(argv[1], "statistics?") == 0) {
    char buf[256];
    audio_descr_t *ad = &nevot.rx.ad;

    if (argc != 2) {
      sprintf(interp->result, "wrong # args: should be %s", argv[0]);
      return TCL_ERROR;
    }
    stats = &m->stats;
    stats->expected = (*m->s->rt.packets_exp)(m);

    /*
     * Don't count duplicate packets in loss statistic, as we
     * otherwise get negative loss.
     */ 
    stats->lost = (stats->expected > stats->received - stats->dup) ?
      stats->expected - (stats->received - stats->dup): 0;

    sprintf(buf,
    "{expected %u} {received %u} {marker %u} {bad_hdr %u} {bad_fmt %u}",
      (unsigned int)stats->expected,
      (unsigned int)stats->received,
      (unsigned int)stats->marker,
      (unsigned int)stats->bad_header,
      (unsigned int)stats->bad_fmt);
    Tcl_AppendResult(interp, buf, NULL);

    sprintf(buf,
    " {late %u} {jitter %0.3f} {dup %u} {order %u} {lost %u} {good %u} {samples %u}",
      (unsigned int)stats->late,
      (unsigned int)stats->jitter / (double)(ad->sample_rate),
      (unsigned int)stats->dup,
      (unsigned int)stats->misorder,
      (unsigned int)stats->lost,
      (unsigned int)stats->good,
      (unsigned int)stats->samples);
    Tcl_AppendResult(interp, buf, NULL);

    sprintf(buf, " {control %d} {created %d} {last {%d %d}}",
      (int)stats->control, (int)m->created,
      (int)m->last[0].tv_sec, (int)m->last[1].tv_sec);
    Tcl_AppendResult(interp, buf, NULL);

    /* playout delay in seconds */
    if (stats->pit) {
      sprintf(buf, " {delay %0.3f}",
        stats->playout_delay / (double)(ad->sample_rate * stats->pit));
      stats->playout_delay = 0;
      stats->pit = 0;
      Tcl_AppendResult(interp, buf, NULL);
    }

    /* packets sent to this site */
    if (stats->sent) {
      sprintf(buf, " {sent %u}", (unsigned int)stats->sent);
      Tcl_AppendResult(interp, buf, NULL);
    }

    /* audio format */
    sprintf(buf, " {media {%s}}", audio_name(&m->audio));
    Tcl_AppendResult(interp, buf, NULL);

    /* frame size */
    sprintf(buf, " {frame %0.3f}", m->covers / (double)m->audio.sample_rate);
    Tcl_AppendResult(interp, buf, NULL);

    /* end of list */
    tcl(interp, "control_send {%s/%s statistics %s}",
      m->s->name, m->id, interp->result);
  }
  /* ignore our own echo */
  return TCL_OK;
} /* cmd_member */

/*
* Process simple command and options.
* Return TCL_OK.
*/
static int cmd_simple(ClientData client, Tcl_Interp *interp,
  int argc, char *argv[])
{
  int ch = (long)client;  /* selector */
  int status = TCL_OK;

  argv++; argc--;  /* skip command name */
  switch(ch) {
  case 'i':  /* program info; with audio: return list of capabilities */
    if (argc > 0) {
      if (strcmp(argv[0], "version") == 0) {
        Tcl_SetResult(interp, version, TCL_STATIC);
      }
      else if (strcmp(argv[0], "date") == 0) {
        Tcl_SetResult(interp, __DATE__, TCL_STATIC);
      }
      else {
        sprintf(interp->result, "unsupported info option '%s'",
          argv[0]);
        status = TCL_ERROR;
      }
    }
    break;
  } /* switch */
  return status;
} /* cmd_simple */


/*
* Called on exiting through Tcl routine 'nevot_exit'.
*/
static int cmd_terminate(ClientData client, Tcl_Interp *interp, int argc,
  char *argv[])
{
  static boolean exit_pending;

  if (exit_pending) return TCL_OK;  /* avoid infinite loops */
  exit_pending = TRUE;
  Audio_exit();   /* do actual clean-up */
  return TCL_OK;  /* proceed with exit processing */
} /* cmd_terminate */


/*
* Event handler; generates control messages (Tcl specific).
*/
static void event(Audio_session_t s, Audio_member_t m,
  Audio_event_t e, ...)
{
  va_list args;
  int iarg;
  float farg;
  int eval = TRUE;   /* by default, evaluate */
  Tcl_DString cmd, id;
  extern void Tk_BackgroundError(Tcl_Interp *interp);
  extern void vu_show(int i, float value);
  extern void vu_show_max(int i, float value);

  va_start(args, e);

  Tcl_DStringInit(&cmd);
  Tcl_DStringAppendElement(&cmd, "control_send");
  Tcl_DStringStartSublist(&cmd);

  /* create member identifier s/m */
  Tcl_DStringInit(&id);
  if (s) {
    Tcl_DStringAppend(&id, s->name, -1);
  }
  if (m && m->id && (e != Audio_hello)) {
    Tcl_DStringAppend(&id, "/", 1);
    Tcl_DStringAppend(&id, m->id, -1);
  }
  Tcl_DStringAppendElement(&cmd, Tcl_DStringValue(&id));

  switch(e) {
  case Audio_receive:
    Tcl_DStringAppendElement(&cmd, "receiving");
    Tcl_DStringAppendElement(&cmd, itoa(va_arg(args, int)));
    break;

  case Audio_created:
    Tcl_DStringAppendElement(&cmd, "created");
    break;

  case Audio_opened:
    Tcl_DStringAppendElement(&cmd, "opened");
    Tcl_DStringAppendElement(&cmd, va_arg(args, char *));
    break;

  case Audio_closed:
    Tcl_DStringAppendElement(&cmd, "closed");
    Tcl_DStringAppendElement(&cmd, va_arg(args, char *));
    break;

  case Audio_bye:
    Tcl_DStringAppendElement(&cmd, "bye");
    Tcl_DStringAppendElement(&cmd, va_arg(args, char *));
    Tcl_DeleteCommand(interp2, Tcl_DStringValue(&id));
    break;

  case Audio_alive:
    /*
     * If m->id is zero, the site has been closed (after BYE); we
     * should not send alive messages in that case.
     */
    if (m->id)
      Tcl_DStringAppendElement(&cmd, va_arg(args, int) ? "alive" : "!alive");
    break;

  case Audio_hello:
    Tcl_DStringAppendElement(&cmd, "hello");
    Tcl_DStringAppendElement(&cmd, m->id);
    Tcl_DStringAppend(&id, "/", 1);
    Tcl_DStringAppend(&id, m->id, -1);
    Tcl_CreateCommand(interp2, Tcl_DStringValue(&id), cmd_member, (ClientData)m, 0);
    break;

  case Audio_active:
    Tcl_DStringAppendElement(&cmd, va_arg(args, int) ? "active" : "!active");
    break;

  case Audio_vu:
    iarg = va_arg(args, int);
    farg = va_arg(args, double);
    vu_show(iarg, farg);
    eval = FALSE;
    break;

  case Audio_vu_max:
    iarg = va_arg(args, int);
    farg = va_arg(args, double);
    vu_show_max(iarg, farg);
    eval = FALSE;
    break;

  case Audio_error:
    iarg = va_arg(args, int);
    /* handle errors */
    switch (iarg) {
      /*
       * If we get an audio packet with the wrong encoding, we
       * don't want to show errors continuously, so we silence
       * that site.
       */
      case Audio_EENC:
      case Audio_ECHANNELS:
      case Audio_ERATE:
        if (m) Audio_MemberListen(m, FALSE);
        break;

      /*
	   * If we try to open the audio device because packets are
	   * arriving and fail (because somebody else has the device), we
	   * want to avoid getting the same error message again and again.
	   * Thus, we wait for the user to enable listening, trying to
	   * acquire the device again.
       */
      case Audio_EAUDIO_BUSY:
        Audio_SessionListen(s, FALSE);
        break;
      default:
        break;
    }
    Tcl_DStringAppendElement(&cmd, "error");
    Tcl_DStringAppendElement(&cmd, Audio_strerror(iarg));
    break;

  case Audio_door:
    Tcl_DStringAppendElement(&cmd, "door");
    Tcl_DStringAppendElement(&cmd, va_arg(args, char *));
    break;

  case Audio_cname:
    Tcl_DStringAppendElement(&cmd, "cname");
    Tcl_DStringAppendElement(&cmd, va_arg(args, char *));
    break;

  case Audio_name:
    Tcl_DStringAppendElement(&cmd, "name");
    Tcl_DStringAppendElement(&cmd, va_arg(args, char *));
    break;

  case Audio_email:
    Tcl_DStringAppendElement(&cmd, "email");
    Tcl_DStringAppendElement(&cmd, va_arg(args, char *));
    break;

  case Audio_phone:
    Tcl_DStringAppendElement(&cmd, "phone");
    Tcl_DStringAppendElement(&cmd, va_arg(args, char *));
    break;

  case Audio_loc:
    Tcl_DStringAppendElement(&cmd, "loc");
    Tcl_DStringAppendElement(&cmd, va_arg(args, char *));
    break;

  case Audio_img:
    Tcl_DStringAppendElement(&cmd, "img");
    Tcl_DStringAppendElement(&cmd, va_arg(args, char *));
    break;

  case Audio_note:
    Tcl_DStringAppendElement(&cmd, "note");
    Tcl_DStringAppendElement(&cmd, va_arg(args, char *));
    break;

  case Audio_tool:
    Tcl_DStringAppendElement(&cmd, "tool");
    Tcl_DStringAppendElement(&cmd, va_arg(args, char *));
    break;

  case Audio_source:
    Tcl_DStringAppendElement(&cmd, "source");
    Tcl_DStringAppendElement(&cmd, va_arg(args, char *));
    break;

  case Audio_from:
    Tcl_DStringAppendElement(&cmd, "from");
    Tcl_DStringAppendElement(&cmd, va_arg(args, char *));
    break;

  default:
    break;
  }

  /* evaluate control command, if any */
  Tcl_DStringEndSublist(&cmd);
  if (eval && (Tcl_Eval(interp2, Tcl_DStringValue(&cmd)) != TCL_OK))
    Tk_BackgroundError(interp2);
  va_end(args);
  Tcl_DStringFree(&cmd);
  Tcl_DStringFree(&id);
} /* event */

/*
 * signal that all sessions are still active
 */
static int cmd_active_all(ClientData client, Tcl_Interp *interp,
  int argc, char *argv[])
{
  Audio_session_t s;

  ALL_SESSIONS(s) {
    tcl(interp, "control_send {%s active}", s->name);
  }
  return TCL_OK;
} /* cmd_active_all */


/*
* Initialize Tcl command interpreter. Must be called before any
* Tcl_*() calls. Ensures that only one copy is running.
*/
int cmd_init(Tcl_Interp *interp)
{
  extern void rx_config(Tcl_Interp *);
  extern void tx_config(Tcl_Interp *);
  extern int onlyone(char *, int);
  int result;

  /* assure that only one copy is running for this user and host */
  result = onlyone("nevot", 0);
  if (result == -1) {
    sprintf(interp->result, "error in locking file %s", strerror(errno));
    return TCL_ERROR;
  }
  /* lock file exists, but process doesn't; live dangerously */
  else if (result < 0) {
  }
  /* another process running */
  else if (result > 0) {
    fprintf(stderr, "Another copy of NeVoT running (pid %d)\n", result);
    tcl(interp, "control_send agent NeVoT%s attached %d", version, result);
    exit(1);
  }

  /* set stdin to be unbuffered to avoid problems if input is from a pipe */
  setbuf(stdin, (char *)NULL);

  /* variable traces */
#define C(txt, cmd, c)  Tcl_TraceVar(interp, txt, TCL_TRACE_WRITES, cmd, (ClientData)c);
  C("audio_priority", cmd_general,  'a');
  C("audio_tick",     cmd_general,  't');
  C("role",           cmd_general,  'r');
  C("debug",          cmd_general,  'v');
  C("debugfile",      cmd_general,  'd');
#undef C

  Audio_init(event);
  tx_config(interp);
  rx_config(interp);

  /* commands */
  Tcl_CreateCommand(interp, "session",         cmd_session_audio,
    0, 0);
  Tcl_CreateCommand(interp, "all/all/all",     cmd_session_all,
    0, 0);
  Tcl_CreateCommand(interp, "active_all",      cmd_active_all,
    0, 0);
  Tcl_CreateCommand(interp, "stats",           cmd_simple,
    (ClientData)'?', 0);
  Tcl_CreateCommand(interp, "infon",           cmd_simple,
    (ClientData)'i', 0);
  Tcl_CreateCommand(interp, "terminate",       cmd_terminate,
    0, 0);
  Tcl_CreateCommand(interp, "audioctl",        cmd_audioctl,
    0, 0);
  return TCL_OK;
} /* cmd_init */


