# Control panel for IBM AIX workstations using the Ultimedia Audio Adapter
#
# $Id: ibm_audioctl.tcl,v 1.3 1998/07/30 11:48:10 dor Exp $
#
# $Log: ibm_audioctl.tcl,v $
# Revision 1.3  1998/07/30 11:48:10  dor
# *** empty log message ***
#
# Revision 1.1  1995/08/17  14:00:37  hgs
# Initial revision
#
# Revision 1.4  1995/06/27  15:24:25  czahl
# - changes were needed because we also want to switch the internal
#   speaker on and off.
#
# Revision 1.3  1995/06/27  13:11:07  czahl
# - bug: aux1 radiobutton set the wrong value
#
# Revision 1.2  1995/06/27  10:38:05  czahl
# - First "official" version for setting the AIX audio parameters.
#   The style of this panel is very different from to the others,
#   keep this in mind...
#
#

proc MakeScale {w From To o Label Side} {
	frame $w
	  frame $w.s
	    frame $w.s.l
	      label $w.s.l.t -text $To
	      label $w.s.l.f -text $From
	      pack $w.s.l.t -side top
	      pack $w.s.l.f -side bottom
	      set S [scale $w.s.s -from $From -to $To -orient $o -showvalue 0 \
		-to $From -from $To ]
	    if {$Side == "left"} {
	      pack $w.s.l -side left -expand true -fill both
	      pack $w.s.s -side right
	    } else {
	      pack $w.s.l -side right -expand true -fill both
	      pack $w.s.s -side left
	    }
	  label $w.l -text $Label
	  pack $w.s -side top
	  if {$Side == "left"} {
	    pack $w.l -side top -anchor e
	  } else {
	    pack $w.l -side top -anchor w
	  }
	return $S
}

proc audioctl_panel {{w .audioctl} {title "Audio Control"}} {
	if {[winfo exists $w]} {destroy $w}

	toplevel $w
	set i [frame $w.i -bd 1 -relief raised]
	  frame $i.f1 -bd 1 -relief raised
	    label $i.f1.l -text "Input Settings"
	    pack $i.f1.l -fill x -expand true
	  frame $i.f
	    frame $i.f.c -bd 1 -relief raised
	      radiobutton $i.f.c.hgm -text "high gain mic" -relief flat \
		-variable AudioCtl(Input) -value highgainmike \
		-command {audioctl -input $AudioCtl(Input)}
	      radiobutton $i.f.c.lgm -text "low gain mic" -relief flat \
		-variable AudioCtl(Input) -value lowgainmike \
		-command {audioctl -input $AudioCtl(Input)}
	      radiobutton $i.f.c.lin -text "line 1" -relief flat \
		-variable AudioCtl(Input) -value line1 \
		-command {audioctl -input $AudioCtl(Input)}
	      radiobutton $i.f.c.aux -text "aux 1" -relief flat \
		-variable AudioCtl(Input) -value line2 \
		-command {audioctl -input $AudioCtl(Input)}
	      pack $i.f.c.hgm $i.f.c.lgm $i.f.c.lin $i.f.c.aux \
		-side top -anchor w -expand true -padx 2m
	    frame $i.f.g -bd 1 -relief raised
	      label $i.f.g.l -text "Gain"
	      frame $i.f.g.s
	        set s [MakeScale $i.f.g.s.l 0 100 vertical "L" left]
		$s config -command "audioctl -leftgain "
	        set s [MakeScale $i.f.g.s.r 0 100 vertical "R" right]
		$s config -command "audioctl -rightgain "
	        pack $i.f.g.s.l $i.f.g.s.r -side left
	      pack $i.f.g.l $i.f.g.s -side top -padx 2m
	    pack $i.f.c $i.f.g -side left -expand true -fill y
	  pack $i.f1 $i.f -side top -expand true -fill x
	set o [frame $w.o -bd 1 -relief raised]
	  frame $o.f1 -bd 1 -relief raised
	    label $o.f1.l -text "Output Settings"
	    pack $o.f1.l -fill x -expand true
	  frame $o.f2
	    frame $o.f2.c -bd 0 -relief raised
	      checkbutton $o.f2.c.isp -text "internal speaker" -relief raised \
		-bd 1 -variable AudioCtl(IntSpeaker) \
		-command {
			audioctl -intspeaker $AudioCtl(IntSpeaker)
			audioctl -output line1
		}
	      frame $o.f2.c.f -bd 1 -relief raised
	        label $o.f2.c.f.l -text Balance
	        scale $o.f2.c.f.s -from -100 -to 100 -showvalue false \
			-orient horizontal \
			-command "audioctl -balance "
	        frame $o.f2.c.f.f
	          label $o.f2.c.f.f.l -text L -anchor w
	          label $o.f2.c.f.f.m -text M
	          label $o.f2.c.f.f.r -text R -anchor e
	          pack $o.f2.c.f.f.l $o.f2.c.f.f.m $o.f2.c.f.f.r \
			-side left -expand true -fill x
	        pack $o.f2.c.f.l $o.f2.c.f.s $o.f2.c.f.f -side top -expand true -fill x -padx 2m
	      pack $o.f2.c.isp $o.f2.c.f \
		-anchor w -expand true -fill both -padx 0m -side top
	    frame $o.f2.v -bd 1 -relief raised
	      label $o.f2.v.l -text "Volume"
	      scale $o.f2.v.s -from 100 -to 0 -orient vertical \
		-command "audioctl -volume "
	      pack $o.f2.v.l $o.f2.v.s -side top -fill both -pady 2m
	    pack $o.f2.c $o.f2.v -side left -expand true -fill both
	  pack $o.f1 $o.f2 -side top -fill x -expand true
	set b [frame $w.b -bd 1 -relief raised]
	  frame $b.f -bd 1 -relief raised
	    button $b.f.b -text "Dismiss" -command "destroy $w"
	    pack $b.f.b -fill x -padx 3m -pady 3m
	  pack $b.f -expand true -fill both
	pack $i $o $b -side top -expand true -fill x
}

#audioctl_panel

