#
# silence detection
#
# $Id: setup.tcl,v 1.3 1998/07/30 11:48:12 dor Exp $
#
proc silence_panel {{w .silence} {title "Silence detection"}} {
  if {[winfo exists $w]} {
    return
  }
  toplevel $w
  wm title $w $title

  set lw 30
  pack \
  [hscale $w.before_spurt "Before talkspurt"  $lw 0 10 1 pkt tx(before)] \
  [hscale $w.after_spurt  "After talkspurt"   $lw 0 10 1 pkt tx(after)] \
  [hscale $w.sd_increase  "Threshold incr."   $lw   0  1 0.01 dB sd(increase)] \
  [hscale $w.sd_min       "Minimum threshold" $lw -60  0 0.5 dB sd(minimum)] \
  [hscale $w.sd_max       "Maximum threshold" $lw -60  0 0.5 dB sd(maximum)] \
  [hscale $w.sd_hyst      "Hysteresis"        $lw   0 10 0.1 dB sd(hysteresis)] \
  [hscale $w.sd_echo      "Echo suppression"  $lw   0 10 0.1 dB sd(echo)] \
    -side top -anchor nw
  pack \
    [checkbutton $w.sd -text "enable" -variable sd(enable) -relief flat] \
    [button $w.ok -text "Dismiss" -command "destroy $w"] \
    -side top
}


#
# AGC parameters
#
proc agc_panel {{w .agc} {title "Automatic gain control"}} {
  if {[winfo exists $w]} {
    return
  }
  toplevel $w
  wm title $w $title

  set lw 30
  set pos {top frame nw}
  pack \
   [hscale $w.agc_tc   "Time constant"  $lw 0.1  10 0.2 sec agc(tc)] \
   [hscale $w.agc_low  "Low volume"     $lw -60 -10 1 dB agc(low)] \
   [hscale $w.agc_high "High volume"    $lw -60   0 1 dB agc(high)] \
   -side top -anchor nw
  pack \
   [checkbutton $w.agc -text "enable" -variable agc(enable) -relief flat] \
   [button $w.ok -text "Dismiss" -command "destroy $w"] \
   -side top
}


#
# VU parameters
#
proc vu_panel {{w .vu} {title "Volume meter"}} {
  if {[winfo exists $w]} {
    return
  }
  toplevel $w
  wm title $w $title

  set lw 30
  pack \
    [hscale $w.vu_interval "Update interval"  $lw 1 20  1 pkt vu(interval)] \
    [hscale $w.vu_hyst     "Update threshold" $lw 0  5  0.5 dB vu(hyst)] \
    -side top -anchor nw
  pack \
    [checkbutton $w.vu -text "enable" -variable vu(enable) -relief flat] \
    [button $w.ok -text "Dismiss" -command "destroy $w"] \
    -side top
}


#
# Delay adjustment parameters
#
proc del_panel {{w .del} {title "Delay adjustment"}} {
  if {[winfo exists $w]} {
    return
  }
  toplevel $w
  wm title $w $title

  set lw 30
  pack \
  pack \
    [button $w.ok -text "Dismiss" -command "destroy $w"] \
    -side top
}


#
# incoming and outgoing audio
#
proc audio_panel {{w .audio} {title "Local audio"}} {
  if {[winfo exists $w]} {
    return
  }
  toplevel $w
  wm title $w $title

  set lw 30
  set pos {top frame nw}
  pack \
  [blist $w.audio_priority "Release audio" $lw 2 \
   {{on_demand   on_demand    "on_demand"}
    {idle        idle         "idle"}
    {idle_demand idle_demand  "idle and demand"}
    {never       never        "never"}} audio_priority] \
  [checkbutton $w.monitor -text "Monitor audio" -variable tx(monitor) -relief flat ] \
    -side top -anchor nw
  pack [button $w.ok -text "Dismiss" -command "destroy $w"] \
     -side bottom
}


#
# timeout values
#
proc timeout_panel {{w .timeout} {title "Timeouts"}} {
  if {[winfo exists $w]} {
    return
  }
  toplevel $w
  wm title $w $title

  set lw 30
  pack \
    [hscale $w.soft_to "Soft timeout" $lw 2 50 1 intervals rx(soft_timeout)] \
    [hscale $w.hard_to "Hard timeout" $lw 2 200 1 intervals rx(hard_timeout)] \
    -side top -anchor nw
  pack \
    [button $w.ok -text "Dismiss" -command "destroy $w"] -side bottom
}


#
# echo suppression
#
proc echo_panel {{w .echo} {title "Echo Suppression"}} {
  if {[winfo exists $w]} {
    return
  }
  toplevel $w
  wm title $w $title

  pack \
    [checkbutton $w.echo -text "Echo suppression" -variable echo_supp -relief flat ] \
    -side top -anchor w
  pack \
    [button $w.ok -text "Dismiss" -command "destroy $w"] -side bottom
}


#
#
#
proc debug_accept {} {
  global debug dbg

  set t ""
  foreach d [array names dbg] {
    if {$dbg($d)} {
      lappend t $d
    }
  }
  set debug $t
}


#
# debug functions
#
proc debug_panel {{w .debug} {title "Debugging"}} {
  global debug dbg

  if {[winfo exists $w]} {
    return
  }
  toplevel $w
  wm title $w $title

  set lw 20
  set pos {top frame nw}

  # this should probably be in a trace
  # checkbuttons should allow updating lists
  if {[info exists dbg]} {
    foreach e [array names dbg] {
      set dbg($e) 0
    }
  }
  foreach e $debug {
    set dbg($e) 1
  }

  pack \
    [checkbutton $w.tx    -text "Transmit"          -var dbg(tx)    -rel flat] \
    [checkbutton $w.rx    -text "Receive"           -var dbg(rx)    -rel flat] \
    [checkbutton $w.sd    -text "Silence Detection" -var dbg(sd)    -rel flat] \
    [checkbutton $w.audio -text "Audio"             -var dbg(audio) -rel flat] \
    [checkbutton $w.cntl  -text "Control"           -var dbg(cntl)  -rel flat] \
    [checkbutton $w.ts    -text "Talkspurts"        -var dbg(ts)    -rel flat] \
    [checkbutton $w.pmm   -text "PMM messages"      -var dbg(pmm)   -rel flat] \
    [field_str $w.debugfile  "Debug output" $lw 20 "" debugfile] \
    -side top -anchor nw

  pack [frame $w.action]
  pack \
    [button $w.action.dismiss -text "Dismiss" -command "destroy $w"] \
    [button $w.action.accept  -text "Accept" \
     -command "debug_accept; destroy $w"] \
    -padx 10 -side left
}
