#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "/usr/include/netinet/in.h"  /* struct sockaddr_in */
#include <arpa/inet.h>   /* inet_ntoa() */
#include <string.h>
#include <tcl.h>
#include <tk.h>


Tcl_Interp *interp_;          /* global interpreter for this application */

#include "pmm-rsvp.c"

int Tcl_AppInit(Tcl_Interp *interp)
{
  
#if STANDALONE
  if (Tcl_InitStandAlone(interp) == TCL_ERROR) {
    return TCL_ERROR;
  }
  if (Tk_InitStandAlone(interp) == TCL_ERROR) {
    return TCL_ERROR;
  }
#else
  if (Tcl_Init(interp) == TCL_ERROR) {
    return TCL_ERROR;
  }
  if (Tk_Init(interp) == TCL_ERROR) {
    return TCL_ERROR;
  }
#endif
  
  interp_ = interp;     /* set global pointer to tcl interpreter */

  
  if (Udp_Init(interp) == TCL_ERROR) {
    return TCL_ERROR;
  }
  /* set up commands */
  
  Tcl_CreateCommand(interp, "idstring", Tcl_idstring, 0, 0);

  /* execute inlined tcl script */
  if (Tcl_Eval(interp, pmm-rsvp_tcl) != TCL_OK) {
    fprintf(stderr, interp->result);
    Exit();
  }
    
  while (1) { /* start event handler */
    /*
     * Tk doesn't give priority to the X server so file handlers
     * (i.e., packets from the network) can starve the window
     * system (technically, tk lets one X event through per file
     * event, but a file event can trigger more than one X event
     * so the X server loses to a high-rate video source).
     * The solution is to let the X server go first by only
     * processing file handlers when their are no pending
     * X events.
     */
    if (Tcl_DoOneEvent(TCL_WINDOW_EVENTS|TCL_IDLE_EVENTS|TCL_DONT_WAIT)==0)
      Tcl_DoOneEvent(0);
    }

  return TCL_OK;
}


