/*
 * RSVP Release 4.2a3
 * Linux version (Linux 2.0.34 or Redhat 5.1)
 * Date:	August 15, 1998
 * By Phil Wang, Ph.D
 * Computer Science Department, Columbia University
 * Email:	yhwang@cs.columbia.edu
 * URL:	http://www.cs.columbia.edu/~yhwang/cworks.html
 * FTP:	http://www.cs.columbia.edu/~yhwang/ftp/qos/rsvp
 */
/*
 *	rsvp_rcv.c
 *	C Program for Internet RSVP-based Socket:	Receiver/Server
 *		Connectint to a socket via RSVP
 *		(1)	Socket initialized as a server;
 *		(2)	RSVP session as a receiver;
 *		(3)	Listen & select I/O input;
 *		(4)	Read a message from the socket
 */

#include <sys/types.h>
#include <sys/signal.h>
#include <sys/socket.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/time.h>
#include <stdio.h>
#include <sys/un.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <sys/signal.h>

#include <rapi_lib.h>
#include <rapi_err.h>
#include <rsvp_intserv.h>

#ifdef  LINUX
#include <huge_val.h>
#define INFINITY32f     ((float32_t)HUGE_VAL)
#endif  /* LINUX */

#define	MAX_CLIENTS	32

fd_set	fd_masks;
fd_set	rcv_select_fd_mask;
fd_set	last_select_fd_mask;
fd_set	readable_fd_masks;
fd_set	writeable_fd_masks;
int	fd_to_client[MAX_CLIENTS];
int	num_client;

struct timeval wait_time;
int	rsvpfd, sockfd;
int	ssnid;
void	end_socket();

struct sockaddr_in	snd, rcv;
int 	port_number, proto_type, sock_type;

int
main(argc,argv)
	int argc;
	char *argv[];
{
	int	retval;
	char	sndhost[128], rcvhost[128];
	int	select_fd;
	struct timeval wait_time;
	struct hostent *hst_hp, *gethostbyname();
	int	udp_connected=0;

	/*
	 *	1. Inputing/Parameters
	 */
	/*	Input handler	*/
	if((argc>1) && (argv[1][0]=='-'))
	{	/*	Usage of arguments	*/
		printf("Usage:	rsvp_rcv [ sndhost protocol portnumber ]\n");
		exit(0);
	}

	if(argc<4)
	{	/*	Input from stdin	*/
		/*	Sender Hostname	*/
		printf("Sender Hostname [local]: ");
		scanf("%s", sndhost);
		if((sndhost[0]=='l') || (sndhost[0]=='L'))
		gethostname(sndhost, 128);

		/*	Protocol	*/
		proto_type=-1;
		while( (proto_type<0) || (proto_type>1) )
		{
	 	       	printf("Protocol [0: tcp; 1: udp]:    ");        
  		      	scanf("%d", &proto_type);
		}

		/*	Port number	*/
		printf("Port number:	");
		scanf("%d", &port_number);
	}
	else
	{	/*	Parse the arguments	*/
		strcpy(sndhost, argv[1]);
		proto_type = atoi(argv[2]);
		port_number= atoi(argv[3]);
	}

	/*	Convert to real protocol & socket type	*/
	switch(proto_type)
	{
		case	0:
			printf("rsvp_rcv:	RSVP reservation with TCP.\n");
			proto_type = IPPROTO_TCP;
			sock_type  = SOCK_STREAM;
			break;
		case	1:
			printf("rsvp_rcv:	RSVP reservation with UDP.\n");
			proto_type = IPPROTO_UDP;
			sock_type  = SOCK_DGRAM;
			break;
	}

	/*	Printing	*/
	printf("rsvp_rcv:	proto_type=%d port_number=%d\n", proto_type, port_number);

	/*
	 *	2. Initialization
	 */
	/*	--- Receiver ---	*/
	/*	Host Info:	Local only	*/
	gethostname(rcvhost, 128);
	hst_hp=gethostbyname(rcvhost);

	/*	Socket Address	*/
	memset(&rcv, 0, sizeof(struct sockaddr_in));
	rcv.sin_family=AF_INET;	/*	IPV4	*/
/*
	rcv.sin_addr.s_addr=htonl(INADDR_ANY);
*/
	memcpy((char *)&rcv.sin_addr, (char *)hst_hp->h_addr, hst_hp->h_length);
	rcv.sin_port=htons(port_number);	/*	 one port number	*/
	printf("rcv.sin_addr.s_addr=%x:%d\n", rcv.sin_addr.s_addr, rcv.sin_port);

	/*	--- Sender ---	*/
	/*	Host Info	*/
	hst_hp=gethostbyname(sndhost);

	/*	Socket Address	*/
	memset(&snd, 0, sizeof(struct sockaddr_in));
	snd.sin_family=AF_INET;	/*	IPV4	*/
	memcpy((char *)&snd.sin_addr, (char *)hst_hp->h_addr, hst_hp->h_length);
	if( snd.sin_addr.s_addr==rcv.sin_addr.s_addr)	/*	Same host	*/
		snd.sin_port=htons(port_number+1);	/*	different from reserver : used by bind	*/
	else	/*	Two different hosts	*/
		snd.sin_port=rcv.sin_port;
	printf("snd.sin_addr.s_addr=%x:%d\n", snd.sin_addr.s_addr, snd.sin_port);

	/*
	 *	3. Socket
	 */
	/*	open	*/
	sockfd=socket(AF_INET, sock_type, 0);
	if(sockfd<0)
	{
		perror("socket() ");
		exit(1);
	}
	else
		printf("socket() is OK: sockfd=%d\n", sockfd);

 	/*	Socket setting:	TCP/UDP only	*/
{
	int	level=SOL_SOCKET, optname=SO_REUSEADDR;
	int	optval=1;
	int	optlen=sizeof(int);

	setsockopt(sockfd, level, optname, (char *)&optval, optlen);
}

	/*	Bind to socket server	*/
	if(bind(sockfd, (struct sockaddr *)&rcv, sizeof(rcv))<0)
	{
		perror("bind ");
		exit(1);
		
	}
	else
		printf("bind() is OK.\n");

	/*	check the sock parameters	*/
#ifdef	ADDbyGET
	memset(&snd, 0, sizeof(struct sockaddr_in));
	retval=sizeof(struct sockaddr);
	if(getsockname(sockfd, (struct sockaddr *)&snd, &retval)<0)
	{
		perror("getsockname");
		exit(1);
	}
	else
		printf("getsockname():	IP=%x portno=%d\n", snd.sin_addr.s_addr,snd.sin_port);
#endif	ADDbyGET

	/*	Set the connection non-blocking     */
        fcntl (sockfd, F_SETFL, O_NONBLOCK);

	/*
	 *	4. RSVP moving to rsvp_client
	 */
	rsvpfd=0;

	/*	Open a RSVP Session based on IP & portnumber of receiver	*/
	ssnid = rapi_session((struct sockaddr *)&rcv,
				proto_type, 0, 0, 0, &retval);

	if(ssnid <1)
	{
		printf("rapi_session() is failed: retval=%d", retval);
		exit(1);
	}
	else
		printf("rapi_session() is OK: ssnid=%d\n", ssnid);


	/*
	 *	3. Listen and select the I/O input
	 */
	/*	Handling ^C	*/
	signal(SIGINT, end_socket);

	/*	Listen 9 clients at most	*/
	if(sock_type == SOCK_STREAM)	/*	TCP only	*/
		listen(sockfd,9);

	/*	The first client: itself	*/
	num_client=0;
	fd_to_client[sockfd]=num_client;	/*	server or listener	*/
	num_client++;

	/*	Set the file mask	*/
	FD_ZERO(&rcv_select_fd_mask);
	FD_SET(sockfd, &rcv_select_fd_mask);

	/*
	 *	5. Listen and select I/O 
	 */
	for(;;)
	{
		/*	Write to all existing clients: keep in pace
		 *	==> accept() returns no sockaddr
		 */
		sleep(2);
		write_clients();

		/*
		 *	Check the socket stream
		 */
		memcpy(&last_select_fd_mask, &rcv_select_fd_mask, sizeof(fd_set));
		wait_time.tv_sec= 5;	/*	1 second	*/
		wait_time.tv_usec=0;
		/*	select_fd = the order number of ready I/O channel	*/
		select_fd = select (FD_SETSIZE, (fd_set *) &last_select_fd_mask, (fd_set *)NULL, (fd_set *) NULL, &wait_time);
		printf("select_fd2=%d.\n", select_fd);

		if(select_fd<=0)	/* An error or timeout occurred */
		{
            		if (select_fd < 0)
                    			printf("select(): errno=%d\n", errno);
			continue;
	        }

		if(FD_ISSET(rsvpfd, &last_select_fd_mask))	/*	RSVP deamon Event	*/
			rapi_dispatch();
		else
		if(FD_ISSET(sockfd, &last_select_fd_mask))
		{	/*	socket connection	*/
			int	acceptfd;

			acceptfd=accept_new_client(sock_type);

			/*	make a reservation	*/
			if(sock_type == SOCK_STREAM)	/*	TCP	*/
				rsvp_client(acceptfd);
			else	/*	UDP	*/
			{
				/*	make a reservation	*/
				if(!udp_connected)
				{
					rsvp_client(sockfd);
					udp_connected =1;
				}
				read_from_client();
			}
		}
		else	/*	TCP client	*/
			read_from_client();

	}
	
}


/*
 *	Accepting for Client's connection: TCP & UDP
 */
accept_new_client(sock_type)
int	sock_type;
{
	int	cur_fd, len;
        struct sockaddr_in      remotehost;

	if(!FD_ISSET(sockfd, &last_select_fd_mask))
		return 0;

	memset(&remotehost, 0, sizeof(struct sockaddr_in));
	if(sock_type == SOCK_STREAM)
	{	/*	TCP	*/
		len=sizeof(struct sockaddr);
		if((cur_fd=accept(sockfd, (struct sockaddr *)&remotehost, &len))<0)
			perror("accept");
		printf("accept_new_client():	TCP connecting from IP=%x:%d\n", remotehost.sin_addr.s_addr, remotehost.sin_port);
	}
	else	/*	UDP	*/
	{
		if(num_client>1)	/*	only one client	*/
			return;
		else
			cur_fd=sockfd;
	}

	printf("accept_new_client():	cur_fd=%d\n", cur_fd);
	if(cur_fd>0)
	{
		FD_SET(cur_fd,&rcv_select_fd_mask);
		fd_to_client[cur_fd]=num_client;
		printf("accept_new_client():	client=%d cur_fd=%d\n", num_client, cur_fd);
		num_client ++;
	}

	return cur_fd;
}

int
rsvp_client(int acceptfd)
	/*	Reserve a session:	receiver	*/
{
        struct sockaddr_in      rhost/* locate receiver */,
                                shost/* remote sender   */;
	int	retval, len;
	rapi_sid_t	sid= ssnid;
		/*	Session ID	*/
	int	flags = RAPI_REQ_CONFIRM;
/*
	int	flags = 0;
		/*	Flag:	request confirmation of reservation	*/
	struct sockaddr *	recvhost = (struct sockaddr *)&rhost;
/*	struct sockaddr *	recvhost = (struct sockaddr *)&rcv;
		/*	Receiver host: IP address and port	*/
/*
	rapi_styleid_t	styleid = RAPI_RSTYLE_WILDCARD;	/*	OK for Solaris & Linux	*/
	rapi_styleid_t	styleid = RAPI_RSTYLE_FIXED;	/*	OK for Solaris	*/
		/*	Receiver style: filter_spec_list & flow_spec_list	*/
	rapi_stylex_t *	style_extension = (rapi_stylex_t * ) NULL;
		/*	Receiver style extension: optional	*/
	rapi_policy_t *	rcvr_policy = (rapi_policy_t * ) NULL;
		/*	Receiver policy data: optional	*/
	int	n_filter;
		/*	No of RAPI filter spec objects	*/
	rapi_filter_t	filter_spec_list;
		/*	List of RAPI filter spec objects	*/
	int	n_flow;
		/*	No of RAPI flow spec objects (0: no reserve)	*/
	rapi_flowspec_t	flow_spec_list;
		/*	List of RAPI flow spec objects	*/
	char	spec_str[256];

        printf("rsvp_client:     acceptfd=%d\n", acceptfd);
#ifdef	ADDbyGET	/*	OK for TCP not UDP	*/
        /*
         *      Get two counterpart of the socket
         */
        len=sizeof(struct sockaddr_in);
        getsockname(acceptfd, (struct sockaddr *)&rhost, &len);
 
	/*	UDP does not work with getpeername()	*/
        len=sizeof(struct sockaddr_in);
        getpeername(acceptfd, (struct sockaddr *)&shost, &len);
	printf("rsvp_client:     rhost0.sin_addr.s_addr=%x sin_port=%d\n", rhost.sin_addr.s_addr, rhost.sin_port);
	printf("rsvp_client:     shost0.sin_addr.s_addr=%x sin_port=%d\n", shost.sin_addr.s_addr, shost.sin_port);
#else	/*	OK for TCP and UDP	*/
	shost=snd;	rhost =rcv;
#endif	ADDbyGET

	printf("rsvp_client:     rhost.sin_addr.s_addr=%x sin_port=%d\n", rhost.sin_addr.s_addr, rhost.sin_port);
	printf("rsvp_client:     shost.sin_addr.s_addr=%x sin_port=%d\n", shost.sin_addr.s_addr, shost.sin_port);
	/*	Filter spec	*/
	n_filter = 1;
	filter_spec_list.len = sizeof(rapi_hdr_t) + sizeof(rapi_filter_base_t);
	filter_spec_list.form = RAPI_FILTERFORM_BASE;
	filter_spec_list.filt_u.base.sender = shost;
	rapi_fmt_filtspec((rapi_filter_t *) &filter_spec_list, spec_str, 256);
	printf("rsvp_client:	Filter Spec=%s\n", spec_str);

	/*	Flow spec	*/
	n_flow = 1;
	flow_spec_list.len = sizeof(rapi_hdr_t) + sizeof(qos_flowspecx_t);
	flow_spec_list.form = RAPI_FLOWSTYPE_Simplified;
/*
	flow_spec_list.specbody_qosx.spec_type = QOS_GUARANTEED;
*/
	flow_spec_list.specbody_qosx.spec_type = QOS_CNTR_LOAD;
	flow_spec_list.specbody_qosx.xspec_r = 63000;	/* Token bucket avg rate (Bps)*/
	flow_spec_list.specbody_qosx.xspec_b = 3000; /* 2 * flow_spec_list.specbody_qosx.xspec_r;	/* Token bucket depth (B) */
/*
	flow_spec_list.specbody_qosx.xspec_p = 5000; /*	Peak data rate	*/
	flow_spec_list.specbody_qosx.xspec_p = INFINITY32f; /*	Peak data rate	*/
	flow_spec_list.specbody_qosx.xspec_m = 0;
	flow_spec_list.specbody_qosx.xspec_M = 63000;	/*	Max pkt size ( bytes )	*/
	flow_spec_list.specbody_qosx.xspec_R = 0;	/*	Rate ( bytes )	*/
	flow_spec_list.specbody_qosx.xspec_S = 0;	/*	Slack term (microsecs )	*/
	rapi_fmt_flowspec((rapi_filter_t *) &flow_spec_list, spec_str, 256);
	printf("rapi_reserve:	Flow Spec=%s\n", spec_str);

	retval = rapi_reserve(sid, flags, recvhost, styleid, style_extension, rcvr_policy, n_filter, &filter_spec_list, n_flow, &flow_spec_list);
	if(retval != 0)
	{
		printf("rapi_reserve() is failed: retval=%d\n", retval);
		exit(1);
	}
	else
		printf("rapi_reserve() is OK: retval=%d\n", retval);

	/*	Get the Socket file descriptor	*/
	rsvpfd=rapi_getfd(ssnid);
	if(rsvpfd <1)
	{
		printf("rapi_getfd() is failed: rsvpfd=%d\n", rsvpfd);
		exit(1);
	}
	else
		printf("rapi_getfd() is OK: rsvpfd=%d\n", rsvpfd);

	FD_SET(rsvpfd, &rcv_select_fd_mask);

}

/*
 *	Read from Client
 */
read_from_client()
{
	int	i, cur_client;
	int	read_fd_bit, select_fd;

	printf("read_from_client():	called\n");
	/*	Find which client to read	*/
	readable_fd_masks = last_select_fd_mask;
	for(i=0; i<MAX_CLIENTS; i++)
	{
		cur_client=fd_to_client[i];
		if(!cur_client)
			continue;

		if(FD_ISSET(i, &readable_fd_masks))
		{
			printf("read_from_client():	client=%d fd = %d\n", cur_client, i);
			/*	read the message	*/
			read_client_message(i);
		}
        }
}

read_client_message(fd)
int	fd;
{
	char	request[256];
	int	cur_client;

	request[0]=0;
	if(sock_type == SOCK_STREAM)	/*	TCP	*/
	{
		if(read(fd, request, 256)<1)
		{	
			perror("read_client_message");
			check_client_connection();
			return(-1);
		}
	}
	else	/*	UDP	*/
	{
		int	addrlen;

		addrlen = sizeof(snd);
        	if(recvfrom(fd, (char*)request, 256, 0, (struct sockaddr *)&snd, &addrlen)<1)
		{	
			perror("read_client_message");
			check_client_connection();
			return(-1);
		}
		printf("read_client_message():	snd.sin_addr.s_addr=%x:%d\n", snd.sin_addr.s_addr, snd.sin_port);
	}
	
	printf("read_client_message():	request=%s\n", request);
	if(!strncasecmp(request, "END", 3))
	{
		cur_client=fd_to_client[fd];
		close(fd);
		FD_CLR(fd, &rcv_select_fd_mask);
		fd_to_client[fd]=0;	num_client--;
		printf("read_client_message(): End from Client %d\n", cur_client);
		return 0;
	}
}

/*
 *	Checking if some clients have died
 */
check_client_connection()
{
	int	i;
	int	cur_fd, cur_client, select_fd;

	wait_time.tv_sec=1;
	wait_time.tv_usec=0;
	printf("Checking for dead clients ...\n");
	for(i=0; i<MAX_CLIENTS; i++)
	{
		cur_client=fd_to_client[i];
		if(!cur_client)
			continue;

		FD_ZERO(&fd_masks);
		FD_SET(i,&fd_masks);
		
		select_fd = select (FD_SETSIZE, (fd_set *)&fd_masks, (fd_set *)NULL, (fd_set *)NULL, &wait_time);
		printf("check_client_connection():	client_fd=%d select_fd= %d\n",i, select_fd);
		if(select_fd<0)
		{
			printf("Client %d has died.\n", cur_client);
			close(i);
			fd_to_client[i] = 0;
			FD_CLR(i, &rcv_select_fd_mask);
		}
	}

}

int 
write_clients()
{
	char	msg[256];
	static	int	write_times=0, i;
	int	cur_client;

	sprintf(msg, "rsvp_rcv:	msg %d", write_times++);
	printf("write_clients: msg=%s.\n", msg);

	for(i=0; i<MAX_CLIENTS; i++)
	{
		cur_client=fd_to_client[i];
		if(!cur_client)
			continue;

		printf("write_clients: client=%d, fd=%d\n", cur_client, i);
		if(sock_type == SOCK_STREAM)
		{	/*	TCP only	*/
			if(write(i, (char*)&msg,sizeof(msg))<1)
				perror("write_clients");
		}
		else	/*	UDP	*/
		{
        		if(sendto(i, msg, sizeof(msg),0, (struct sockaddr *)&snd, sizeof (snd))<1)
				perror("write_clients");
		}
        }

}

void 
end_socket()
{
	int	i, cur_client;

	printf("end_socket() is called.\n");

	for(i=0; i<MAX_CLIENTS; i++)
	{
		cur_client=fd_to_client[i];
		if(!cur_client)
			continue;

		close(i);
        }

	rapi_release(ssnid);
	close(sockfd);
	exit(0);
}
