/*
 * RSVP Release 4.2a3
 * Linux version (Linux 2.0.34 or Redhat 5.1)
 * Date:	August 15, 1998
 * By Phil Wang, Ph.D
 * Computer Science Department, Columbia University
 * Email:	yhwang@cs.columbia.edu
 * URL:	http://www.cs.columbia.edu/~yhwang/cworks.html
 * FTP:	http://www.cs.columbia.edu/~yhwang/ftp/qos/rsvp
 */
/*
 *	rsvp_snd.c
 *	C Program for Internet RSVP-based Socket:	Sender/Client
 *		Connectint to a socket via RSVP
 *		(1)	Socket initialized as a client;
 *		(2)	RSVP session as a sender;
 *		(3)	Select I/O input;
 *		(4)	write a message from the socket
 */

#include <sys/types.h>
#include <unistd.h>
#include <sys/fcntl.h>
#include <sys/signal.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <stdio.h>
#include <netinet/in.h>
#include <errno.h>
#include <netdb.h>

#include <rapi_lib.h>
#include <rapi_err.h>
#include <rsvp_intserv.h>

#ifdef  LINUX
#include <huge_val.h>
#define INFINITY32f     ((float32_t)HUGE_VAL)
#endif  /* LINUX */

fd_set	fd_masks;
fd_set	snd_select_fd_mask;
fd_set	last_select_fd_mask;
fd_set	readable_fd_masks;
fd_set	writeable_fd_masks;
struct timeval wait_time;

int	rsvpfd, sockfd;
int	ssnid;
void	end_socket();

void	*
main(argc,argv)
	int argc;
	char *argv[];
{
	
	int	i, retval;
	struct sockaddr_in snd, rcv, sockname;
	int 	port_number, proto_type, sock_type;
	char	sndhost[128], rcvhost[128];
	int	select_fd;
	struct hostent *hst_hp, *gethostbyname();

	/*
	 *	1. Inputing/Parameters
	 */
	/*	Input handler	*/
	if((argc>1) && (argv[1][0]=='-'))
	{	/*	Usage of arguments	*/
		printf("Usage:	rsvp_snd [ rcvhost protocol portnumber ]\n");
		exit(0);
	}

	if(argc<4)
	{	/*	Input from stdin	*/
		/*	Receiver Hostname	*/
		printf("Receiver Hostname [local]: ");
		scanf("%s", rcvhost);
		if((rcvhost[0]=='l') || (rcvhost[0]=='L'))
		gethostname(rcvhost, 128);
	
		/*	Protocol	*/
		proto_type=-1;
		while( (proto_type<0) || (proto_type>1) )
		{
	 	       	printf("Protocol [0: tcp; 1: udp]:    ");        
  		      	scanf("%d", &proto_type);
		}

		/*	Port number	*/
		printf("Port number:	");
		scanf("%d", &port_number);
	}
	else
	{	/*	Parse the arguments	*/
		strcpy(rcvhost, argv[1]);
		proto_type = atoi(argv[2]);
		port_number= atoi(argv[3]);
	}

	/*	Convert to real protocol & socket type	*/
	switch(proto_type)
	{
		case	0:
			printf("rsvp_snd:	RSVP Sender with TCP.\n");
			proto_type = IPPROTO_TCP;
			sock_type  = SOCK_STREAM;
			break;
		case	1:
			printf("rsvp_snd:	RSVP Sender with UDP.\n");
			proto_type = IPPROTO_UDP;
			sock_type  = SOCK_DGRAM;
			break;
	}

	/*	Printing	*/
	printf("rsvp_snd:	sndhost=%s rcvhost=%s\n", sndhost, rcvhost);
	printf("rsvp_snd:	proto_type=%d port_number=%d\n", proto_type, port_number);

	/*
	 *	2. Initialization
	 */

	/*	--- Receiver ---	*/
	/*	Host Info	*/
	hst_hp=gethostbyname(rcvhost);

	/*	Socket Address	*/
	memset(&rcv, 0, sizeof(struct sockaddr_in));
	rcv.sin_family=AF_INET;	/*	IPV4	*/
	memcpy((char *)&rcv.sin_addr, (char *)hst_hp->h_addr, hst_hp->h_length);
	rcv.sin_port=htons(port_number);	/*	 one port number	*/
	printf("rcv.sin_addr.s_addr=%x:%d\n", rcv.sin_addr.s_addr, rcv.sin_port);

	/*	--- Sender ---	*/
	/*	Host Info:	Local only	*/
	gethostname(sndhost, 128);
	hst_hp=gethostbyname(sndhost);

	/*	Socket Address	*/
	memset(&snd, 0, sizeof(struct sockaddr_in));
	snd.sin_family=AF_INET;	/*	IPV4	*/
	memcpy((char *)&snd.sin_addr, (char *)hst_hp->h_addr, hst_hp->h_length);
	if( snd.sin_addr.s_addr==rcv.sin_addr.s_addr)   /*      Same host       */
		snd.sin_port=htons(port_number+1);	/*	different from reserver : used by bind	*/
        else    /*      Two different hosts     */ 
                snd.sin_port=rcv.sin_port;
	printf("snd.sin_addr.s_addr=%x:%d\n", snd.sin_addr.s_addr, snd.sin_port);
	/*
	 *	3. RSVP
	 */
	/*	Open a RSVP Session:	server	*/
{
	struct sockaddr	*dst= (struct sockaddr	*)&rcv;
		/*	Destination host: Receiver's IP address and port	*/
	int	protoid=proto_type;
		/*	IP protocol ID for the session. 0: UDP(17)	*/
	int	flags=0;
		/*	Flag: RAPI_GPI_SESSION or RAPI_USE_INTSERV	*/
	int	(*event_rtn)()= (int *)NULL;

		/*	Upcall routine to notify RSVP errors and state change events	*/
	void	*event_arg=NULL;
		/*	Arguments passed to the above upcall routine	*/
	int	*err=&retval;
		/*	Return the RSVP error	*/

	ssnid = rapi_session(dst, protoid, flags, event_rtn, event_arg, err);
}
	if(ssnid <1)
	{
		printf("rapi_session() is failed: retval=%d\n", retval);
		exit(1);
	}
	else
		printf("rapi_session() is OK: ssnid=%d\n", ssnid);

	/*	Become a sender	*/
{
        rapi_sid_t	sid=ssnid;
		/*	Session ID	*/
        int	flags=0;
		/*	No flag	*/
        struct sockaddr *	localhost = (struct sockaddr *)&snd;
		/*	Sender host: IP address and port	*/
        rapi_filter_t *	sender_template = (rapi_filter_t *) NULL;
		/*	RAPI filter spec of the data packets to be sent: optional	*/
        rapi_tspec_t	sender_tspec;
		/*	a Tspec defining traffic sepc created by sender: required	*/
        rapi_adspec_t	sender_adspec;
		/*	an Adspec: optional/unsupported	*/
        rapi_policy_t *	sender_policy = (rapi_policy_t *) NULL;
		/*	Sender policy data struct: optional/unsupported	*/
        int	ttl=0;
		/*	Data TTL (optional) */
	char	spec_str[256];

	/*
	 *	Sender Traffic spec
	 */
	sender_tspec.len = sizeof(rapi_hdr_t) + sizeof(qos_tspecx_t);;
	sender_tspec.form = RAPI_TSPECTYPE_Simplified;
	sender_tspec.tspecbody_qosx.spec_type = QOS_TSPEC;
	sender_tspec.tspecbody_qosx.xtspec_r = 63000;
	sender_tspec.tspecbody_qosx.xtspec_b = 3000; /* 2*sender_tspec.tspecbody_qosx.xtspec_r;
/*
	sender_tspec.tspecbody_qosx.xtspec_p = 5000;
*/
	sender_tspec.tspecbody_qosx.xtspec_p = INFINITY32f;
	sender_tspec.tspecbody_qosx.xtspec_m = 0;
	sender_tspec.tspecbody_qosx.xtspec_M = 63000;
	rapi_fmt_tspec(&sender_tspec, spec_str, 256);
	printf("rapi_sender:	Traffic Sepc=%s\n", spec_str);

	/*
	 *	Sender Advertising spec
	 */
	sender_adspec.len = sizeof(rapi_adspec_t);
  	sender_adspec.form = RAPI_ADSTYPE_Simplified;

        /*      General path characterization parameters */
#ifdef	LINUX
  	sender_adspec.adspecbody_qosx.xaspec_flags = XASPEC_FLG_BRK;
#else
  	sender_adspec.adspecbody_qosx.xaspec_flags = XASPEC_FLG_PARM;
#endif	LINUX
  	sender_adspec.adspecbody_qosx.xaspec_hopcnt = 10;
  	sender_adspec.adspecbody_qosx.xaspec_path_bw = 1000000;
  	sender_adspec.adspecbody_qosx.xaspec_min_latency = 1000;
  	sender_adspec.adspecbody_qosx.xaspec_composed_MTU = 1500;

        /*      Controlled-Load service Adspec parameters */
#ifdef	LINUX
  	sender_adspec.adspecbody_qosx.xaspec_flags = XASPEC_FLG_BRK;
#else
  	sender_adspec.adspecbody_qosx.xaspec_flags = XASPEC_FLG_PARM;
#endif	LINUX
  	sender_adspec.adspecbody_qosx.xClaspec_override = 1;
  	sender_adspec.adspecbody_qosx.xClaspec_hopcnt = 10;
  	sender_adspec.adspecbody_qosx.xClaspec_path_bw = 1000000;
  	sender_adspec.adspecbody_qosx.xClaspec_min_latency = 1000;
  	sender_adspec.adspecbody_qosx.xClaspec_composed_MTU = 1500;

	/*
	 *	Make a Sender
	retval = rapi_sender(sid, flags, localhost, sender_template, &sender_tspec, &sender_adspec, sender_policy, ttl);
	 */
	retval = rapi_sender(sid, flags, localhost, sender_template, &sender_tspec, 0, sender_policy, ttl);
}
	if(retval != 0)
	{
		printf("rapi_sender() is failed: retval=%d\n", retval);
		perror("rapi_sender");
		exit(1);
	}
	else
		printf("rapi_sender() is OK: retval=%d\n", retval);


	/*	Get the Socket file descriptor	*/
	rsvpfd=rapi_getfd(ssnid);
	if(rsvpfd <1)
	{
		printf("rapi_getfd() is failed: rsvpfd=%d\n", rsvpfd);
		exit(1);
	}
	else
		printf("rapi_getfd() is OK: rsvpfd=%d\n", rsvpfd);

	/*
	 *	4. Socket
	 */
	/*	open	*/
	sockfd=socket(AF_INET, sock_type, 0);
	if(sockfd<0)
	{
		perror("socket() ");
		exit(1);
	}
	else
		printf("socket() is OK: sockfd=%d\n", sockfd);

	/*	Bind to a socket address	*/
	if(bind(sockfd, (struct sockaddr *)&snd, sizeof(snd))<0)
	{
		perror("bind ");
		exit(1);
		
	}
	else
		printf("bind() is OK.\n");

	/*	connect to socket server	*/
	if(connect(sockfd, (struct sockaddr *)&rcv, sizeof(rcv))<0)
	{
		perror("connecting ");
		exit(1);
		
	}
	else
		printf("connect() is OK.\n");

	/*	check the sock parameters	*/
	memset(&sockname, 0, sizeof(struct sockaddr_in));
	retval=sizeof(struct sockaddr);
	if(getsockname(sockfd, (struct sockaddr *)&sockname, &retval)<0)
	{
		perror("getsockname");
		exit(1);
	}
	else
		printf("getsockname():	IP=%x:%d\n", sockname.sin_addr.s_addr, sockname.sin_port);

        /* Set the connection non-blocking     */
        fcntl (sockfd, F_SETFL, O_NONBLOCK);

	/*
	 *	5. Listen and select I/O 
	 */
	/*	Handling ^C	*/
	signal(SIGINT, end_socket);

	/*	Set the file mask	*/
	FD_ZERO(&snd_select_fd_mask);
	FD_SET(rsvpfd, &snd_select_fd_mask);
	FD_SET(sockfd, &snd_select_fd_mask);

	/*	Loop for input & client connection	*/
	for(;;)
	{
		/*	Write to the server	*/
		write_to_socket(sockfd);

		memcpy(&last_select_fd_mask, &snd_select_fd_mask, sizeof(fd_set));

		/*
		 *	Check the socket stream
		 */
		wait_time.tv_sec= 15;
		wait_time.tv_usec=0;
		/*	select_fd = the order number of ready I/O channel	*/
		select_fd = select (FD_SETSIZE, (fd_set *) &last_select_fd_mask, (fd_set *)NULL, (fd_set *) NULL, &wait_time);
		printf("select_fd2=%d.\n", select_fd);

		if(select_fd<=0)	/* An error or timeout occurred */
		{
            		if (select_fd < 0)
                    			printf("select(): errno=%d\n", errno);
			continue;
	        }

		if(FD_ISSET(rsvpfd, &last_select_fd_mask))	/*	RSVP deamon Event	*/
			rapi_dispatch();
		else
		if(FD_ISSET(sockfd, &last_select_fd_mask))	/*	Data incoming	*/
			read_from_socket(sockfd);

	}
}

write_to_socket(fd)
int	fd;
{
	char	msg[256];
	static	int	write_times=0;

	sprintf(msg, "rsvp_snd:	msg %d", write_times++);
	printf("write_to_socket: msg=%s.\n", msg);

	if(write(fd, (char*)&msg, sizeof(msg))<1)
		perror("write_to_socket");

}

read_from_socket(fd)
int	fd;
{
	char	msg[256];
	
	if(read(fd, msg, 256)<1)
	{	
		perror("read_from_socket");
		printf("read_from_socket():	errno=%d\n", errno);
		return(-1);
	}
	else
		printf("read_from_socket: msg=%s\n", msg);
}

void 
end_socket()
{
	int	i, cur_client;
	char	msg[256];

	printf("end_socket() is called.\n");
	strcpy(msg, "End");
	write(sockfd, msg, sizeof(msg));
	sleep(2);
	close(sockfd);
	rapi_release(ssnid);
	exit(0);
}


