/**********************************************************************************
 *
 * procedures invoked by timers
 *
 * Copyright 1997 by Stefan Hoffmann
 * GMD Fokus
 *
 ***********************************************************************************/



#include <pwd.h>
#include <stdio.h>
#include <syslog.h>
#include <unistd.h>
#include "debug.h"
#include "defines.h"
#include "list.h"
#include "misc.h"
#include "notify.h"
#include "multimer.h"
#include "isc.h"
#include "reply_codes.h"
#include "response.h"
#include "udp.h"



extern char          myAddr[17];
extern int           udp_server_socket;
extern inv_list_t    *invitation_list;
extern int           isc_connections;
extern int           isc_socket;
extern int           sip_port;



/* close_isc_socket ******************************************************************
 *
 *  closes local multicast socket to isc
 *  invoked if isc_connections <= 0
 *
 ***********************************************************************************/
Notify_func close_isc_socket (Notify_client client)
{ 
  static char  func[] = "close_isc_socket";


  DEBUG (2, syslog (LOG_DEBUG, "%s",func);)

  close(isc_socket);
} /* close_isc_socket */



/* request_timeout ******************************************************************
 *
 *  invoked if no definite  response from client was received
 *
 *  in:  
 *  - client: pointer to invitation_list entry to which request was sent to
 *
 ***********************************************************************************/
Notify_func request_timeout (Notify_client client)
{
  static char  func[] = "request_timeout";
  inv_list_t  *inv;
  struct timeval time;


  DEBUG (2, syslog (LOG_DEBUG, "%s",func);)

  inv = (inv_list_t *)client; 
  
  /***********************************************************************************
   * decrease counter for outstanding responses because the daemon no longer waits
   * for a response 
   ***********************************************************************************/
  --inv->locations;

  if (inv->transport_to_server == SOCK_STREAM) { 
    /* kill event handler of  tcp socket to the server */
    notify_set_input_func((Notify_client)inv, NOTIFY_FUNC_INPUT_NULL, inv->socket_to_server);   } 

  reply(inv, NOT_CURRENTLY_HERE, "Not Currently Here"); 
} /* request_timeout */



/* isc_request_timeout **************************************************************
 *
 *  sends a reply (ALTERNATIVE_ADRESS or NOT_CURRENTLY_HERE) to client
 *  invoked if no definite response from isc
 *
 ***********************************************************************************/
Notify_func isc_request_timeout (Notify_client client)
{
  static char     func[] = "isc_request_timeout";
  loc_list_t      *loc;
  inv_list_t      *inv;
  struct timeval  time;


  DEBUG (2, syslog (LOG_DEBUG, "%s",func);)

  inv = (inv_list_t *)client;
  loc = loc_list_search(inv, (char *)myAddr);
  
  /***********************************************************************************
   * check if there are still other connection to isc
   ***********************************************************************************/
  if (--isc_connections == 0) {
    /* kill event handler of isc_socket */
    notify_set_input_func(0, NOTIFY_FUNC_INPUT_NULL, isc_socket);
    time.tv_sec = 1; 
    time.tv_usec = 0;
    /* close isc_socket */
    timer_set(&time, (Notify_func)close_isc_socket, (Notify_client)0, 1);
  }
    
  /***********************************************************************************  
   * decrease counter for outstanding responses because the daemon no longer  waits
   * for a response from isc
   ***********************************************************************************/
  --inv->locations;
  
  /***********************************************************************************
   * set response_received flag to avoid decreasing the locations counter if a 
   * response from this location will be received in the future
   ***********************************************************************************/  
  if (loc != NULL)
    loc->response_received = TRUE;

  if (inv->state != PARALLEL_SEARCH && inv->locations > 1) {
    /***********************************************************************************
     * user is probably located at several locations but request was only
     * forwarded to isc because this is not the first proxy
     * isc doesn't reply, so send all alternative locations to the client
     ***********************************************************************************/
    inv->locations = 0;
    reply(inv, ALTERNATIVE_ADDRESS, "Alternative Address");
  }
  else 
    reply(inv, NOT_CURRENTLY_HERE, "Not Currently Here");
} /* isc_request_timeout */



/* start_isc_timeout *******************************************************************
 *
 *  invoked if isc was started by isc_timeout and didn't respond
 *  sends appropriate reply to client
 *
 *  in:  
 *  - client: pointer to invitation_list entry 
 *
 ***********************************************************************************/
Notify_func start_isc_timeout (Notify_client client)
{
  static char     func[] = "start_isc_timeout";
  inv_list_t      *inv;
  loc_list_t      *loc;
  struct timeval  time;

 
  DEBUG (2, syslog (LOG_DEBUG, "%s",func);)
  
  inv = (inv_list_t *)client;
  loc = loc_list_search(inv, (char *)myAddr);
  
  /***********************************************************************************
   * check if there are still other connection to isc
   ***********************************************************************************/
  if (--isc_connections == 0) {
    /* kill event handler of isc_socket */
    notify_set_input_func(0, NOTIFY_FUNC_INPUT_NULL, isc_socket);
    time.tv_sec = 1; 
    time.tv_usec = 0;
    /* close isc_socket */
    timer_set(&time, (Notify_func)close_isc_socket, (Notify_client)0, 1);
  }
    
  /***********************************************************************************  
   * decrease counter for outstanding responses because the daemon no longer  waits
   * for a response from isc
   ***********************************************************************************/
  --inv->locations;
  
  /***********************************************************************************
   * set response_received flag to avoid decreasing the locations counter if a 
   * response from this location will be received in the future
   ***********************************************************************************/  
  if (loc != NULL)
    loc->response_received = TRUE;

  if (inv->state != PARALLEL_SEARCH && inv->locations > 1) {
    /***********************************************************************************
     * user is probably located at several locations but request was only
     * forwarded to isc because this is not the first proxy
     * isc doesn't reply, so send all alternative locations to the client
     ***********************************************************************************/
    inv->locations = 0;
    reply(inv, ALTERNATIVE_ADDRESS, "Alternative Address");
  }
  else 
    reply(inv, SERVER_INTERNAL_ERROR, "Server Internal Error");
} /* start_isc_timeout */



/* isc_timeout ***********************************************************************
 *
 *  invoked if isc doesn't respond to the request with a "Trying" response
 *  starts isc and retransmits request to isc
 *
 *  in:  
 *  - client: pointer to invitation_list entry 
 *
 ***********************************************************************************/
Notify_func isc_timeout(Notify_client client)
{
  static char     func[] = "isc_timeout";
  inv_list_t      *inv;
  struct timeval  time;
 

  DEBUG (2, syslog (LOG_DEBUG, "%s",func);)

  inv = (inv_list_t *)client;

  /***********************************************************************************
   * isc doesn't reply, so try to start it and retransmit request 
   ***********************************************************************************/
  if (start_isc((struct passwd *)inv->userEntry, inv->call_id)) {
    start_isc_timeout((Notify_client)inv->call_id);
    return;
  }

  /* start T_start_isc timer for a "+ call_id" response from isc */
  time.tv_sec = 20;  
  time.tv_usec = 0;
  timer_set(&time, (Notify_func)start_isc_timeout, (Notify_client)inv, 1);

} /* isc_timeout */



/* retransmit_timeout ****************************************************************
 *
 *  periodically invoked if request was sent to next server via UDP and no final 
 *  response was received
 *  retransmits request to next server
 *
 *  in:  
 *  -client: pointer to location_list entry to which request was sent
 *
 ***********************************************************************************/
Notify_func retransmit_timeout (Notify_client client)
{
  static char     func[] = "retransmit_timeout";
  loc_list_t      *loc;
  inv_list_t      *inv;
  char            new_line[MAX_LINE_LENGTH];
  char            *p;
  struct timeval  retransmit_time;


  DEBUG (2, syslog (LOG_DEBUG, "%s",func);)

  loc = (loc_list_t *)client;
  inv = loc->inv_parent;
  if (++loc->retransmissions > MAX_RETRANSMISSIONS - get_via_header_no(inv->request)) {
    /***********************************************************************************
     * request was too often retransmitted and no final response was received
     ***********************************************************************************/

    --inv->locations;
    loc->response_received = TRUE;
    reply(inv, NOT_CURRENTLY_HERE, "Not Currently Here"); 
  }
  else {
    /***********************************************************************************
     * retransmit request 
     ***********************************************************************************/
    /* create via header */
    sprintf(new_line, "v:%s/UDP %s", SIP_VERSION, myAddr);
    p = str_search(inv->request, "\nv")+1;
    insert_SIP_line(p, new_line);
    UDP_send (loc->addr, sip_port, udp_server_socket, 1, inv->request); 

    /* start T_retransmit timer */
    retransmit_time.tv_sec = RETRANSMIT_TIME;
    retransmit_time.tv_usec = 0;
    timer_set(&retransmit_time, (Notify_func)retransmit_timeout, (Notify_client)loc, 1); 
     
    /* rebuild original request */
    delete_SIP_line(p);
  }
} /* retransmit_timeout */



/* close_invitation ****************************************************************
 *
 *  close sockets and delete entry in invitation_list
 *
 *  in:  
 *  - client: pointer to invitation_list entry 
 *
 ***********************************************************************************/
Notify_func close_invitation (Notify_client client)
{
  static char    func[] ="close_invitation";
  inv_list_t     *inv;

 
  DEBUG (2, syslog (LOG_DEBUG, "%s",func);)

  inv=(inv_list_t *)client;
  if (inv->transport_to_client == SOCK_STREAM) {
    shutdown(inv->socket_to_client, 2);
    close(inv->socket_to_client);
  }

  if (inv->transport_to_server == SOCK_STREAM) {
    shutdown(inv->socket_to_server, 2);
    close(inv->socket_to_server);
  }
  
  LIST_DEL(inv->location_list);
  if (inv->userEntry != NULL)
    {
      free (inv->userEntry->pw_name);
      free (inv->userEntry->pw_passwd);
      free (inv->userEntry->pw_gecos);
      free (inv->userEntry->pw_dir);
      free (inv->userEntry->pw_shell);
      free (inv->userEntry);
    }
  inv_list_remove(inv->call_id);
} /* close_invitation */



/* close_event_handler *************************************************************
 *
 *  delete event handlers 
 *
 *  in:  
 *  - client: pointer to invitation_list entry 
 *
 ***********************************************************************************/
Notify_func close_event_handler (Notify_client client)
{
  static char     func[] ="close_event_handler";
  loc_list_t      *loc;
  inv_list_t      *inv;  
  struct timeval  time;


  DEBUG (2, syslog (LOG_DEBUG, "%s",func);)
    
  inv=(inv_list_t *)client;
    
  /***********************************************************************************
   * kill all timers 
   ***********************************************************************************/
  for (loc = inv->location_list; loc != NULL; loc = loc->next) { 
    timer_set(NULL, (Notify_func)retransmit_timeout, (Notify_client)loc, 1);   
  }

  timer_set(NULL, (Notify_func)request_timeout, (Notify_client)inv, 1);   

  if (inv->transport_to_server == SOCK_STREAM)  
    notify_set_input_func((Notify_client)inv, NOTIFY_FUNC_INPUT_NULL, inv->socket_to_server);     
  if (inv->transport_to_client == SOCK_STREAM)  
    notify_set_input_func((Notify_client)inv, NOTIFY_FUNC_INPUT_NULL, inv->socket_to_client);   
  time.tv_sec = 1; 
  time.tv_usec = 100;  
  timer_set(&time, (Notify_func)close_invitation, (Notify_client)inv, 1);     
} /* close_event_handler */



