#
# pmm.tcl - this file implements the pmm protcol for vic
#
# Initial Version written by K. Samp 
# Modified and enhanced version written by F. Emanuel
#
# $Id: cf-pmm.tcl,v 1.2 1998/07/20 22:39:39 koepsel Exp $
# 
# ----------------------------------------------------------------------

#
# Initialisation of pmm - invoke pmm_module and open_menu
# after vic_main has started

set pmm(tool) ""
set c ""
set myself ""
set first 1
set srclist {}

#
# Called from the pmm_module function.
#
proc init_pmm {} {
    global pmm  
    

#    rename adios pmm_adios
#    proc adios {} {
#	global pmm myself 
#	set src [srctab local]
#	set src $pmm($myself)
#	set cname [$src sdes cname]
#	if [info exists pmm(sn)] {
#	    pmm_send_ " closed"
#	}
#	puts "\n$pmm(tool): Exiting...\n"
#	pmm_adios
#    }

    set pmm(debug) 1
    set pmm(controller) "225.2.0.1 5010"
    set pmm(tool) "vat"
    
    pmm_init_udp $pmm(controller)


    # Issue "attached" message
    pmm_send "agent $pmm(tool) attached [pid] audio"

}



#
# Open UDP connection 
#
proc pmm_init_udp {controller} {
   global pmm c
    
    set host [lindex $controller 0]
    set port [lindex $controller 1]
    set c [udp $host $port]
    #Tcl_SetSockBuffer $c
    fconfigure $c -ttl 0  
    fconfigure $c -buffering none
    fconfigure $c -member 1
    fconfigure $c -buffersize 128000
    fileevent $c readable {pmm_read_udp $c}
}    

#
# Receive UDP packet
#
proc pmm_read_udp {sock} {
    pmm_eval [gets $sock]
}

#
# Source and destination port on all multicast packets is 5010
# (sending port uses bind(), so that we have no way of knowing whether
# we see our own transmissions or those from another process.  
#
proc pmm_eval {cmd} {
    global pmm myself meterDisable mute cnamelist env

    set result [catch {exec id} id]
    set id2 $id
    #set user [string trim [string range $id2 [string first ( $id] [string first ) $id]] "()"] 
 

    if [info exists env(USER) ] {
       set user $env(USER)
    } else {
       set user "unknown"
    }
    if [info exists env(HOSTNAME) ] {
       set host $env(HOSTNAME)
    } else {
       set host "unknown"
    }
    set myself "$user@$host"


    if {[string length $cmd] > 0} {
	if {[lsearch {proc exec rename} [lindex $cmd 0]] >= 0} {
	    puts stderr "[pid] $pmm(tool): forbidden command: $cmd"
	} elseif {[string match "session *" $cmd]} {
	    # Set to generic pattern on first invocation, but then only
	    # allow repeat of first 'session' command.  Must not answer
	    # other video session requests.
	    if {![info exists pmm(sn)]} {

		set textvar1 $cmd
		set end1 [string first / $textvar1]
		incr end1 -1
		set sessiontitle [string trim [string range $textvar1 [string first " " $textvar1] $end1]]
		set pmm(sn) $sessiontitle/audio/*
	    }
	    if {[string match $pmm(sn) [lindex $cmd 1]]} {
		set pmm(sn) [lindex $cmd 1]
		pmm_send_ " created"
		#wm geometry . 90x0+0+0
		wm iconify .

		set mute($myself) 0
	    }
	} else {
	    if {[info exists pmm(sn)]&&[string match $pmm(sn)* $cmd]} {
		pmm_eval_ $cmd 
	    }
	}
    }
    
}


#
# Send PMM message
#
proc pmm_send {msg} {
    global pmm c
    
    puts -nonewline $c $msg 
}


#
# Like pmm_send, but prefix our session name.
#
proc pmm_send_ {msg} {
    global pmm
    
    # To send messages to ISC, we have to wait
    # until the sessiontitle was received.
    # Any pending transmission must be retransmitted 
    # This is achieved by an "after" command
    # condition:  pmm(sn) exists

    if [info exists pmm(sn)] {
	pmm_send "$pmm(sn)$msg"
    } else {
	after 100 pmm_send_ "\"$msg\""
    }	

}

#
# ignore pmm "+" message
#
proc pmm_+ {args} {
}
 
#
# ignore pmm "agent" message
#
proc pmm_agent {args} {
}

proc send_update {} {
    global pmm

    pmm_send_ " active $pmm(update_interval)"
    after $pmm(update_interval) send_update

}



#
# Invoked when a pmm session command was received.
# Vic-specific since it invokes buttons.
#
proc pmm_eval_ {msg} {
    global pmm V first mute srclist cnamelist myself stat_method src_objects src_nickname localsrc localcname sending_pmm receiving_pmm

    set c [lindex $msg 1]
    string trim $c
    switch $c {
	hello {
	}
	update_interval {
            set pmm(update_interval) [lindex $msg 2]
            send_update
        }
	cname {
	}
	open {
	    set net $V(data-net)
	    pmm_send_ " opened [$net addr]"
	}
	opened {
	}
	close {
	    pmm_send_ " closed {}"
	    adios
	}
	closed {
	}
	media {
	    set type [lindex $msg 2]
	    switch $type {
		"PCMU 1 8000" {
		    select_format pcm 2
		}
		"DVI4 1 8000" {
		    select_format dvi 4
		}
	        "L16 1 44100" {

		}
		"LPC 1 8000" {
		    select_format lpc 4
		}
	    }
	}
	transport {
	}
	send {

	    set value [lindex $msg 2]
	    switch $value {
		1 { 
		    if ![audio have] {
			audio_demand
		    }
		    if [audio halfduplex] {
			audio output mute
			audio input unmute
		    }
		  }
		0 {  
		    if [audio have] {
			audio input mute
			if [audio halfduplex] {
			    audio output unmute
			}
		    }
		  }
	    }

	}
	receive {

	    set textvar2 [lindex $msg 0]
	    set start1 [string last "/" $textvar2]
	    incr start1
	    set cname [string trim [string range $textvar2 $start1 [string length $textvar2]]] 
	    set src $src_objects($cname)

	    set value [lindex $msg 2]
	    if {$value != $receiving_pmm($src)} {
		toggle_mute $src
		set receiving_pmm($src) $value
	    } 

	}
	statistics {
	}
	statistics? {
	    set textvar2 [lindex $msg 0]
	    set start1 [string last "/" $textvar2]
	    incr start1
	    set cname [string trim [string range $textvar2 $start1 [string length $textvar2]]] 

	    set w $src_objects($cname)
	    set stats [eval $w stats_pmm] 
	    pmm_send_ "/$cname statistics $stats"
	}

    }
}


#
# This function returns the dotted notation of another host
# by resolving the given name from a hello cmd
# e.g.: user@host.domain.net -> 192.168.0.1
#       user@192.168.0.1     -> 192.168.0.1
#
proc get_host_from_cname {user_host} {
    set  addrstr $user_host
    set  start  [string first "@" $addrstr]
    incr start
    set  end    [string length $addrstr]
    set  addr   [string trim [string range $addrstr $start $end]]

    set  dotted [getdottedaddr $addr]

    return $dotted

}

#
# This function returns the user from a cname string
# received by a hello cmd
#
proc get_user_from_cname {user_host} {
    set  userstr $user_host
    set  end  [string first "@" $userstr]
    incr end -1
    set  user   [string trim [string range $userstr 0 $end]]

    return $user

}                              

proc register_pmm src {
    global localsrc muteNewSites muting_pmm src_nickname src_info src_name src_registered_isc src_objects src_cname

    # first object is always local srcobject 
        if ![info exists localsrc] {    
	    set localsrc $src
	}

        if ![info exists src_cname($src)] {

	    if {[$src sdes cname] == ""} {
		after 100 register_pmm $src
	    } else {
		
		set src_cname($src) [$src sdes cname]
		set cname $src_cname($src)
		
		if ![info exists src_registered_isc($src)] {
		    pmm_send_ " hello $cname"
		    set src_registered_isc($src) "1"
		    audio input mute
		}
		
		if ![info exists src_objects($cname)] {
		    set src_objects($cname) $src
		}
	    }
	}
        
}

proc unregister_pmm src {
    global localsrc muteNewSites muting_pmm src_nickname src_info src_name src_registered_isc src_cname

    if [info exists src_cname($src)] {
	pmm_send_ "/$src_cname($src) bye"
	unset src_registeres_isc($src)
    }

}


proc set_busy_pmm src {
    global pmm src_nickname myself cnamelist localcname localsrc src_cname

    if [info exists src_cname($src)] {
	pmm_send_ "/$src_cname($src) active"
    } else {
	# if we failed, retry cmd after 2sec.
	after 2000 set_busy_pmm "$src"
    }

}

proc clear_busy_pmm src {
    global pmm src_nickname myself localcname localsrc src_cname

    if [info exists src_cname($src)] {
	pmm_send_ "/$src_cname($src) !active"
    } else {
	after 2000 clear_busy_pmm "$src"
    }

}

