/*
 *   ac.c --
 *
 *   This file contains the application control procedures. It is designed as 
 *   a stand alone file with no dependencies on the rest of nevit.
 *   When choosing to contol the sending source a hash table containing all
 *   the incoming streams with their information about the error and 
 *   transmission rates.
 *
 *  (C) 1995 Dorgham Sisalem, Bernd Deffner, GMD Fokus
 *
 *  See the copyright file for copyright information
 */

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <tcl.h>
#include <sys/types.h>
#include <math.h>
#include "ac.h"


static char rcsid[]  = "$Id: ac.cc,v 1.2 1998/08/13 11:17:00 stb Exp $";

static int DEBUG=0;

static AC ac;

AC::AC():TclObject("ac"){
}

AC::~AC(){
  ac_delete_entry(0);
}
/*
 * add an entry to the table
 */
mem_t* AC::ac_add_entry(u_int32_t Ssrc)
{
  int newval;
  Tcl_HashEntry *entryPtr;
  mem_t *m=NULL;
  entryPtr = Tcl_CreateHashEntry(&member_table, (char*)(long)Ssrc, &newval);
  m = (mem_t *)malloc(sizeof(mem_t));
  Tcl_SetHashValue(entryPtr, m);
  members++;
  memset(m, 0, sizeof(mem_t));       /* initialize member structure */
  m->ssrc=Ssrc;
  m->loss=0;
  m->filtered_loss=0;
  m->current_Q=0;
  m->max_Q=-1;
  m->TCP_Rate=0;
  m->time=0;
  m->min_rtt=100; /* set the initial minumum rtt to some very big value*/
  
  m->bbw=-1;
  return m;
}

/*
 * get the time in usecs
 */
u_int32_t AC::ac_time()
{
  struct timeval now;
  gettimeofday(&now,0);
  return((now.tv_sec*1000000)+now.tv_usec);
}



/*
 * convenience function for searching an entry in the hash table
 * returns pointer to member or NULL 
*/
mem_t* AC::ac_find_entry(u_int32_t Ssrc)
{
  Tcl_HashEntry *entryPtr;
  
  entryPtr = Tcl_FindHashEntry(&member_table, (char*)(long)Ssrc);
  if (entryPtr == NULL)
    return NULL;
  return (mem_t *)Tcl_GetHashValue(entryPtr);
}

int AC::ac_max_bandwidth()
{
 return (int)max_bandwidth;
}

/* 
 * deletes an entry from the hash table and frees the allocated 
 * memory for the member entry
 */

void AC::ac_delete_entry(u_int32_t Ssrc)
{
  Tcl_HashEntry *entryPtr;
  mem_t *m=NULL;
  Tcl_HashSearch search;
  if((int)(Ssrc)>0) {
    entryPtr = Tcl_FindHashEntry(&member_table, (char*)(long unsigned )Ssrc);
    if (entryPtr != NULL) {
      m = (mem_t *)Tcl_GetHashValue(entryPtr);
      Tcl_DeleteHashEntry(entryPtr);
      members--;
      free ((char *)m);
      m=NULL;
    }
    return;
  }
  /* if ssrc<0 delete all members */
  if(&member_table) {
    for (entryPtr = Tcl_FirstHashEntry(&member_table, &search); 
	 entryPtr != NULL;
	 entryPtr = Tcl_NextHashEntry(&search)) {
      m = (mem_t *) Tcl_GetHashValue(entryPtr);
      Tcl_DeleteHashEntry(entryPtr);
      members--;
      free ((char *)m);
      m=NULL;
    }
  }
}


/*
*  Initialisation of all ac related parameters
  

void ac_init_all(u_int32_t Ssrc, int algorithm, double alpha, int max_bandwidth,
		 double multiplicative_decrease,
		 double additive_increase,
		 double loss_lower_threshold,
		 double loss_upper_threshold,
		 double congested_recv_fraction,
		 double loaded_recv_fraction)
{
  
  Tcl_InitHashTable(&ac_p->member_table, TCL_ONE_WORD_KEYS);
  ac_p->control			= 1;
  ac_p->algorithm		= algorithm;
  ac_p->alpha			= alpha;
  ac_p->max_bandwidth		= 0.9*max_bandwidth;
  ac_p->min_bandwidth		= 0;
  ac_p->multiplicative_decrease	= multiplicative_decrease/100.0;
  ac_p->additive_increase	= additive_increase;
  ac_p->loss_lower_threshold	= loss_lower_threshold/100.0;
  ac_p->loss_upper_threshold	= loss_upper_threshold/100.0;
  ac_p->congested_recv_fraction	= congested_recv_fraction/100.0;
  ac_p->loaded_recv_fraction	= loaded_recv_fraction/100.0; 
  ac_p->members			= 1;
  ac_p->ssrc			= ssrc;
  ac_p->time                    = 0;
  ac_p->aif                     = additive_increase;
}
*/
/* just init the ssrc and member table */
void AC::ac_init(u_int32_t Ssrc)
{
  
  Tcl_InitHashTable(&member_table, TCL_ONE_WORD_KEYS);
  control		= 1;
  members		= 1;
  ssrc			= Ssrc;
  time                  = 0;
}

/* 
 * update the control information of the sending source
 */

int AC::command(int argc, const char*const* argv)
{
  Tcl& tcl = Tcl::instance();
  char* wrk = tcl.buffer();
  if(argc == 6) {
    if(!strcasecmp(argv[1], "ac_qos")){
      sprintf(wrk,"%d",
	      ac_qos_control(atoi(argv[2]),atof(argv[3]),
			     atof(argv[4]),atoi(argv[5])));
       tcl.result(wrk);
       //cerr<<"new BW: "<<wrk<<" SSRC: "<<argv[2]<<" RR-Loss: "<<argv[3]<<endl;
       return (TCL_OK);
    }
  }


  if(argc == 4) {
    if(!strcasecmp(argv[1], "control")){
      control=1;
      ac_init(atoi(argv[3])); //Local SSRC
      if(DEBUG)
	cout<<"Controll initalized "<<argv[3]<<endl;
      return(TCL_OK);
    }
  }

  if (argc == 3) {
    if(!strcasecmp(argv[1], "control")){
      control=0;
      ac_init(0);
      if(DEBUG)
	cout<<"Control called"<<endl;
      return(TCL_OK);
    }

    if (!strcasecmp(argv[1], "algorithm")){
      algorithm=atoi(argv[2]);
      if(DEBUG)
	cout<<"Algorithm "<<algorithm<<endl;
      return(TCL_OK);
    }

    if (!strcasecmp(argv[1], "delete_entry")){
      ac_delete_entry(atoi(argv[2]));
      if(DEBUG)
	cout<<"Delete Entry "<<argv[2]<<endl;
      return(TCL_OK);
    }
    
    if (!strcasecmp(argv[1], "max_bandwidth")){
      max_bandwidth=0.9*atoi(argv[2]);
      min_bandwidth=10;
      if(DEBUG)
	cout<<"Max Bandwidth "<<max_bandwidth<<endl;
      return(TCL_OK);
    }
    
    if (!strcasecmp(argv[1], "alpha")){
      alpha=atof(argv[2]);
      if(DEBUG)
	cout<<"Alpha "<<alpha<<endl;
      return(TCL_OK);
    }

    if (!strcasecmp(argv[1], "multiplicative_decrease")){
      multiplicative_decrease=atoi(argv[2])/100.0;
      if(DEBUG)
	cout<<"Decrease "<<multiplicative_decrease<<endl;
      return(TCL_OK);
    }

    if (!strcasecmp(argv[1], "additive_increase")){
      additive_increase=atoi(argv[2]);
      if(DEBUG)
	cout<<"Increase "<<additive_increase<<endl;
      return(TCL_OK);
    }

    if (!strcasecmp(argv[1], "loss_upper_threshold")){
      loss_upper_threshold = atoi(argv[2])/100.0;
      if(DEBUG)
	cout<<"Upper Threshold "<<loss_upper_threshold<<endl;
      return(TCL_OK);
    }

    if (!strcasecmp(argv[1], "loss_lower_threshold")){
      loss_lower_threshold = atoi(argv[2])/100.0;
      if(DEBUG)
	cout<<"Lower Threshold "<<loss_lower_threshold<<endl;
      return(TCL_OK);
    }

    if (!strcasecmp(argv[1], "congested_recv_fraction")){
      congested_recv_fraction = atoi(argv[2])/100.0;
      if(DEBUG)
	cout<<"Congested Receiver "<<congested_recv_fraction<<endl;
       return(TCL_OK);
    }
    
     if (!strcasecmp(argv[1], "loaded_recv_fraction")){
       loaded_recv_fraction = atoi(argv[2])/100.0;
       if(DEBUG)
	 cout<<"Loaded Receiver "<<loaded_recv_fraction<<endl;
       return(TCL_OK);
    }
    
  }
  return (TclObject::command(argc, argv));
}

void AC::ac_update(int tag,int parameter)
{
  switch ( tag ) {
  case CONTROL: 
    control		= parameter;
    break;
  case ALGORITHM:
    algorithm		= parameter; 
    break;
  case ALPHA:
    alpha			= parameter;
    break;
  case MAX_BANDWIDTH:
    max_bandwidth		= 0.9*parameter;
    break;
  case MIN_BANDWIDTH:
    min_bandwidth		= parameter;
    break;
  case MULTIPLICATIVE_DECREASE:
    multiplicative_decrease	= parameter/100.0;
    break;
  case ADDITATIVE_DECREASE:
    additive_increase	= parameter;
    break;
  case LOSS_LOWER_THRESHOLD:
    loss_lower_threshold	= parameter/100.0;
    break;
  case LOSS_UPPER_THRESHOLD:
    loss_upper_threshold	= parameter/100.0;
    break;
  case CONGESTED_RECV_FRACTION:
    congested_recv_fraction	= parameter/100.0;
    break;
  case LOADED_RECV_FRACTION:
    loaded_recv_fraction	= parameter/100.0;
    break;
  default :
    perror(" ac_update: Unknown parameter ");
  }
}

/*
 * This function analyses the information contained in the RTP database 
 * and returns the actual network state.
 * The network state is set according to the maximum loss rate of all
 * receivers.
 */
network_state_t AC::worst_receiver()
{
  float max_loss;
  mem_t *m;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;

  network_state_t state=UNLOADED;

  max_loss = -1;

  /*
   *  search through the hash table and look for the highest 
   *  filtered loss value
   */
  for (entryPtr = Tcl_FirstHashEntry(&member_table, &search); 
       entryPtr != NULL;
       entryPtr = Tcl_NextHashEntry(&search)) {
    m = (mem_t *) Tcl_GetHashValue(entryPtr);
    if(DEBUG)
      cerr<<"Source: "<<m->ssrc<<" Loss: "<<m->loss<<endl;
    if (m->ssrc == ssrc) continue;  /* neglect entry for sender */
    if (m->loss > max_loss)
      max_loss = m->loss;
  }

  /*
   *  linear regulator with deadzone
   */
  if(DEBUG)
    cerr<<"Max loss: "<<max_loss<<" Threshold: "<<loss_upper_threshold<<endl;
  if (max_loss >= loss_upper_threshold)
    state = CONGESTED;
  else if (max_loss >= loss_lower_threshold) 
    state = LOADED;
  else
    state = UNLOADED;
  return state;
}


/* 
 * Count the number of receivers that are congested, loaded and unloaded
 * and determine network state 
 */
network_state_t AC::congestion_ratio()
{
  network_state_t state;
  int unloaded, loaded, congested, counted;
  mem_t *m=NULL;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  float v;

  unloaded = 0;
  loaded = 0;
  congested = 0;
  counted = 0;

  /*
   *  search through the hash table and determine the number of 
   *  receivers that are congested, loaded or unloaded
   */
  for (entryPtr = Tcl_FirstHashEntry(&member_table, &search); 
       entryPtr != NULL;
       entryPtr = Tcl_NextHashEntry(&search)) {
    m = (mem_t *) Tcl_GetHashValue(entryPtr);
    if(DEBUG)
      printf("myself= %d recv=%d\n", (unsigned long int)ssrc, m->ssrc);
    if (m->ssrc == ssrc) continue;    /* ignore entry for sender */
    v = m->loss;
    if(DEBUG)
      printf("loss of recv %d is %f\n", m->ssrc, v);
    if (v >= loss_upper_threshold)
      congested++;
    else if (v >= loss_lower_threshold) 
      loaded++;
    else 
      unloaded++;
    counted++;
  }

  if ( (double)congested / counted
      >= congested_recv_fraction)
    state = CONGESTED;
  else if ((double)loaded / counted
	   >= loaded_recv_fraction)
    state = LOADED;
  else 
    state = UNLOADED;

  if(DEBUG)
    fprintf(stderr, "receivers=%d #unloaded=%d #loaded=%d #congested=%d"
	    " state=%d", members - 1, unloaded, loaded, congested,state);
  
  return state;
}

/*
 * low-pass filter to compute the short time statistics                 
 * the smaller a, the greater is the influence of the old value 
 */
float AC::ac_filter(float oldval, float newval, float a)
{
  if(DEBUG)
    fprintf(stderr,"old=%f new=%f, a=%f result=%f\n",oldval,newval,a,(1-a)*oldval + a*newval);
  return ( (1-a)*oldval + a*newval );
}


/*
 * This function interpretes the network state and adjusts the bandwidth
 * of the video agent accordingly
 */
int AC::adjust(network_state_t state,int bandwidth) 
{
  /*
   * Compute new bandwidth
   */
  float current_bw=(float) bandwidth;
  if (control) {
    switch(state) {
    case UNLOADED:    current_bw= current_bw+ additive_increase;
      break;
    case LOADED:      break;
    case CONGESTED:   current_bw = (current_bw*multiplicative_decrease);
    }
  }
  if (current_bw <=min_bandwidth)
    current_bw = min_bandwidth;
  else if (current_bw >= max_bandwidth)
    current_bw = max_bandwidth;

  return (int)current_bw;
}


/* the direct_adjustment algorithm reduces the sending bandwidth directly 
 * to the level reported in the receiver reports. the reduction is then
 * maintained untill a new RR arrives from the same receiver indicating that
 * the loss is beneath the lower loss threshold.
 * In the case that after reducing the rate another RR message was received
 * from another receiver with a heigher loss ratio then the rate is only
 * reduced by the difference of the losses reported in this RR and the one
 * that caused the original loss. Afterwards we have to wait for another RR
 * from this receiver to start increasing the rate again.
 */ 
 
int AC::direct_adjustment(int Ssrc, double Loss, int bandwidth)
{
  int bw=bandwidth;
  static int result=UNLOADED;
  if(DEBUG)
    fprintf(stderr," bw=%i ac_p->loss=%f loss=%f ssrc=%i, ac_ssrc=%i\n",bandwidth, loss,Loss,Ssrc,ssrc_loss);

  if(ssrc_loss==Ssrc){ /* RR from the ssrc with the heighest loss */ 
    if(Loss>loss_lower_threshold){
      bw=(int)(bandwidth*(1-Loss+loss_lower_threshold));
      result=CONGESTED;
    }
    else{
      result=UNLOADED;
    }
    loss=Loss;
  } else {
    if(Loss>loss){/* a new loss reported*/
      bw=(int)(bandwidth*(1-Loss+loss+loss_lower_threshold));
      loss=Loss;
      ssrc_loss=Ssrc;
      result=CONGESTED;
    }
  }

  if(result==UNLOADED) {
    bw+=(int)additive_increase;
  }

  /* check if the bandwidth is still in the desired limits */
 /* if (bw <=min_bandwidth)
     bw = min_bandwidth;*/
  else if (bw >= max_bandwidth)
    bw= (int)max_bandwidth;
  return bw;
}

/*
 * update the entries calculation for this member
 */
void AC::ac_update_entry(mem_t* m )
{
  /* determine the minumum seen round trip time and consider it as the 
     propagation delay */
  if(m->min_rtt> m->rtt)
    m->min_rtt=m->rtt;

  if(DEBUG)
    fprintf(stderr," ac_update min_rtt=%f ",m->min_rtt);
  

  /* determine the current queuing delay */
  m->current_Q = m->rtt - m->min_rtt;

  /* set the current update time */
  m->time=ac_time();
  
  if(DEBUG)
    fprintf(stderr," current_Q=%f max_Q=%f time=%u \n", m->current_Q, m->max_Q, m->time);
}

/*
 * calculate for each member the bandwdith it could have been getting
 * under the condition of current bandwdith and the measured delay and loss
 */
double AC::ac_get_LDBA_bw(mem_t* m,double current_bandwidth) 
{
  double bw=current_bandwidth;
  int min_bbw=ac_min_bbw();
  
  if(min_bbw<1)
    min_bbw=(int)max_bandwidth;
  if(DEBUG)
    fprintf(stderr," LDBA ssrc=%i  bw=%f loss=%f MD=%f AIF=%f \n ----min_bbw=%i\n",m->ssrc,bw, m->loss,multiplicative_decrease*100, additive_increase,min_bbw);
  
  if(m->loss > loss_lower_threshold) {
    if(m->loss > loss_upper_threshold) {
      bw*=0.5;
    } else {
      bw-=bw*m->loss*(multiplicative_decrease*100);
    }
    m->max_Q=m->current_Q;
    if(DEBUG)
      fprintf(stderr," loss=%f bw=%f \n",m->loss, bw);
    return (bw);
  }
  /* losses lower than lower threshold */
  {
    double delay;
    double window;
    double aif;
   
    /*delay=m->min_rtt+m->max_Q;*/
    delay=m->rtt;

    /* calculate the TCP equivalent increase during UPDATE_INTERVAL and no 
     * losses and rtt of delay in kb/s*/
    window=1.024*(UPDATE_INTERVAL/1000000.0)*8/(delay*2);

    
    /* are we under the lower loss threshold but still have some losses? */
    if(m->loss>0.00001){
      m->TCP_Rate=(1.22*8*1.024)/(delay*sqrt(m->loss));
      if(DEBUG)
	fprintf(stderr," delay=%f loss=%f TCP_Rate=%f \n",delay,m->loss,m->TCP_Rate);

      /* make sure TCP_Rate has an acceptable value */
      if(m->TCP_Rate>max_bandwidth||m->TCP_Rate<0)
	m->TCP_Rate=max_bandwidth;
      
      if(bw>m->TCP_Rate) {
	bw=bw-(bw*m->loss*multiplicative_decrease*100);
	return(bw);
      }
    }
    { /* no losses */
      /* set the additive factor in accordance with the highest bandwdith */
      double aif_factor=1 - (double)(current_bandwidth/max_bandwidth);
      
      /*make sure the factors have maximally the value 1 and a minimum of 0*/
      aif_factor= ((aif_factor<1) ? aif_factor :1);
      aif_factor= ((aif_factor<0) ? 0: aif_factor );
      
      aif=(aif*aif_factor)+aif;
      if( aif<additive_increase)
	aif=additive_increase;
      if(aif>window)
	aif=window;
      
      bw+=aif;
      if(bw>max_bandwidth)
	bw=max_bandwidth;
      if(DEBUG)
	fprintf(stderr,"window=%f aif=%f  bw=%f aif_factor=%f rtt=%f\n",window,aif,bw,aif_factor,m->rtt);
    }
    return (bw);
  }
}
    

   
/*
 * find the smallest possible bandwdith to use 
 */
double AC::ac_get_min_entry(double current_bandwidth)
{
  mem_t *m;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;

  double bandwidth = max_bandwidth;

  /*
   *  search through the hash table and look for the lowest
   *  bandwidth
   */
  double tmp_bw;
  for (entryPtr = Tcl_FirstHashEntry(&member_table, &search); 
       entryPtr != NULL;
       entryPtr = Tcl_NextHashEntry(&search)) {
    m = (mem_t *) Tcl_GetHashValue(entryPtr);
    if (m->ssrc == ssrc) continue;  /* neglect entry for sender */
    tmp_bw=ac_get_LDBA_bw(m,current_bandwidth);
    bandwidth= (bandwidth > tmp_bw) 
      ?tmp_bw : bandwidth; 
  }
  if (bandwidth <=min_bandwidth)
    bandwidth = min_bandwidth;
  else if (bandwidth >= max_bandwidth)
    bandwidth= max_bandwidth;

  if(bandwidth > current_bandwidth) {
    aif=bandwidth-current_bandwidth;
  } else {
    aif=additive_increase;
  }
  if(DEBUG)
    fprintf(stderr," current_bw=%f, bandwdith=%f  aif=%f \n",current_bandwidth,bandwidth,aif);

  return bandwidth;
}

/* 
 * update the value of the bottleneck bandwidth for member ssrc
 */
void AC::ac_update_bbw(int Ssrc, double bbw)
{
  mem_t *m;
  m = ac_find_entry(Ssrc);
  if (m == NULL) {
    m = ac_add_entry(Ssrc);
  }
  m->bbw=bbw/1000; /* convert into kb/s */
}

int AC::ac_min_bbw()
{
  mem_t *m;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  static double min_bbw=0;
  int first=1;
  

  /*
   *  search through the hash table and look for the smallest
   *  bbw value
   */
  for (entryPtr = Tcl_FirstHashEntry(&member_table, &search); 
       entryPtr != NULL;
       entryPtr = Tcl_NextHashEntry(&search)) {
    m = (mem_t *) Tcl_GetHashValue(entryPtr);
    if (m->ssrc == ssrc) continue;  /* neglect entry for sender */
    if(m->bbw>0) {
      if(first) {
	first=0;
	min_bbw=m->bbw;
      } else {
	if(m->bbw<min_bbw)
	  min_bbw=m->bbw;
      }
    }
    /* reset bbw to -1 */
    m->bbw=-1;
  }
  if(DEBUG)
     fprintf(stderr," ac_min min_bbw=%f \n",min_bbw);
  return((int)min_bbw);
}


/* 
 * call the selected algorithm
 */
int AC::ac_qos_control(int Ssrc, double loss, double rtt,int bandwidth)
{
  network_state_t result=UNLOADED;
  mem_t *m=NULL;

  if(algorithm==LOSS_BASED_ADJUSTMENT){
    return direct_adjustment(Ssrc,loss,bandwidth);
  }
  m = ac_find_entry(Ssrc);
  if (m == NULL) {
    m = ac_add_entry(Ssrc);
  }
  if(DEBUG)
    cerr<<"Source: "<<m->ssrc <<" Loss: "<<m->loss<<endl; 
  /*m->loss = loss;*/
  m->loss=ac_filter(m->loss, loss,alpha);
  m->rtt=rtt;
  
  if(algorithm==LOSS_DELAY_BASED_ADJUSTMENT){
    ac_update_entry(m);
    
    /* check if it is time to do an update */
    if((int)ac_time()> time) 
      time=(int)(ac_time()+UPDATE_INTERVAL); /* set the time for the next 
					       update interval */
    return((int)ac_get_min_entry(bandwidth)); 
  }

  switch(algorithm) {
    case WORST_RECEIVER: 
      result = worst_receiver();
      break;
    case CONGESTION_RATIO:
      result = congestion_ratio();
      break;
  }
  return adjust(result,bandwidth);
}
