 #
# pmm.tcl - this file implements the pmm protcol for vic
#
# Initial Version written by K. Samp 
# Modified and enhanced version written by F. Emanuel
#
# $Id: pmm.tcl,v 1.16 1998/10/09 17:24:22 dor Exp $
# 
# ----------------------------------------------------------------------

# READ THIS
# if a remote source leaves without sending an explicit bye and then restarts with another ssrc 
# vic gives this steream another window entry (another src)
# to account for this the actual part used is indexed by pmm($cname)
# as the other parts might be used in the case of future crashes we need to make sure taht they
# all have a valid window id before reusing them



 
#
# Initialisation of pmm - invoke pmm_module and open_menu
# after vic_main has started 

proc pmm_module {} {    
    if {[string match "vic_main" [info procs vic_main]]} {
	after 1 pmm_init
    } 
}


if {[catch {pmm_module}] > 0} {
    # try again
    catch {pmm_module}
}

# create and hide the control menu
proc open_menu {} {
    after 30 unpack_buttons_vic
    after 30 build.menu
    #	after 1 pack forget .menu
}

if {[catch {open_menu}] > 0} {
    # try again
    catch {open_menu}
}

#
# check if isc has sent a session message, if it did not then exit
#
proc check_isc { } {
    global pmm
    if {$pmm(session) != 1 } {
	adios
    }
}
#
# Called from the pmm_module function.
#
proc pmm_init {} {
    global pmm srcstate srclist active V src_name src_nickname prefix colorbutton nametag myself env
    

    rename adios pmm_adios
    proc adios {} {
	global pmm myself
	xflush
	#set src [srctab local]
	set src $pmm($myself)
	set cname [$src sdes cname]
	if [info exists pmm(sn)] {
	    pmm_send_ " closed"
	}
	pmm_adios
    }

    set pmm(debug) 1
    set pmm(controller) "225.2.0.1 5010"
    set pmm(tool) "vic-pmm"
    set pmm(send) 0
    pmm_init_udp $pmm(controller)

    # Derive a cname for the member generating the session
    # and tell vic about it

    if [info exists env(USER)] {
       set user $env(USER)
    } else {
       set user "unknown"
    }
    if [info exists env(HOSTNAME)] {
       set host $env(HOSTNAME)
    } else {
       set host "unknown"
    }

    set host [getdottedaddr $env(HOSTNAME)]
    set myself "$user@$host"
    [srctab local] sdes cname $myself
    set src [srctab local]
    # pmm($myself) contains the own src handler 
    set pmm($myself) $src 
    # pmm($src) contains the idstring of source src
    set pmm($src) $myself

    # Issue "attached" message
    pmm_send "agent $pmm(tool) attached [pid] video"

    # Set variable traces for events
    trace variable active w trace_proc
    trace variable active u trace_proc
    trace variable srcstate w trace_proc
    trace variable nametag w trace_proc
    trace variable nametag u trace_proc
    trace variable src_name w trace_proc
    trace variable colorbutton w trace_proc
    trace variable V u trace_proc
    #Set bindings
    bind Vic <Map> {bind_proc %W}

    proc bind_proc {w} {
	# Check whether window name starts with .top.grid
	if {[string match ".top.grid*" $w]} {  
	    
	    bind $w <1> {
		# Extract src from widget name
		set w %W
		regexp {\.[^.]+\.[^.]+\.([^.]+)\.} $w dummy src
		if {[info exist pmm($src)]} {
		    pmm_send_ "/$pmm($src) vic_button1"
		}
	    }

	    bind $w <2> {
		# Extract src from widget name
		set w %W
		regexp {\.[^.]+\.[^.]+\.([^.]+)\.} $w dummy src
		if {[info exist pmm($src)]} {
		    pmm_send_ "/$pmm($src) vic_button2"
		}
	    }
	    
	    bind $w <3> {
		# Extract src from widget name
		set w %W
		regexp {\.[^.]+\.[^.]+\.([^.]+)\.} $w dummy src
		if {[info exist pmm($src)]} {
		    pmm_send_ "/$pmm($src) vic_button3"
		}
	    }
	}
	
	# DELETE dismiss button in video window after it has 
	# exposed

	# Check whether window name starts with .vw
	#if {[regexp {^(\.vw)+} $w]} {
	if {[string match ".vw*" $w]} {
	    pack forget $w.bar.dismiss
	}
    }
    # make sure that we have a connection to isc, if we do not see a session message in 1
    # 0 seconds then exit again
    set pmm(session) 0
    after 10000 check_isc
}


#
# Open UDP connection 
#
proc pmm_init_udp {controller} {
   global pmm c
    
    set host [lindex $controller 0]
    set port [lindex $controller 1]
    set c [udp $host $port]
    Tcl_SetSockBuffer $c
    fconfigure $c -ttl 0  
    fconfigure $c -buffering none
    fconfigure $c -member 1
    fileevent $c readable {pmm_read_udp $c}
}    

#
# Receive UDP packet
#
proc pmm_read_udp {sock} {
    pmm_eval [gets $sock]
}

#
# Source and destination port on all multicast packets is 5010
# (sending port uses bind(), so that we have no way of knowing whether
# we see our own transmissions or those from another process.  
#
proc pmm_eval {cmd} {
    global pmm myself

    if {[string length $cmd] > 0} {
	if {[lsearch {proc exec rename} [lindex $cmd 0]] >= 0} {
	    puts stderr "[pid] $pmm(tool): forbidden command: $cmd"
	} elseif {[string match "session *" $cmd]} {
	    # remember that we have seen a session message and thus have a connection to isc
	    set pmm(session) 1
	    set conf [wm title .]
	    regsub -all " " $conf "%20" conf
	    # Set to generic pattern on first invocation, but then only
	    # allow repeat of first 'session' command.  Must not answer
	    # other video session requests.
	    if {![info exists pmm(sn)]} {
		set pmm(sn) ${conf}/video/*
	    }
	    if {[string match $pmm(sn) [lindex $cmd 1]]} {
		set pmm(sn) [lindex $cmd 1]
		eval "proc pmm_$pmm(sn) {c args} {pmm_session $pmm(sn) \$c \$args}"
		eval "proc pmm_$pmm(sn)/$myself {c args} {pmm_member $pmm(sn) $myself \$c \$args}"
		pmm_send_ " created"
		pmm_send_ " hello $myself"
	    }
	} else {
	    if {[info exists pmm(sn)]&&[string match $pmm(sn)* $cmd] && [catch {eval pmm_$cmd} msg]} {
		#puts stderr "$pmm(tool): pmm error: $cmd: $msg"
	    }
	}
    }
}


#
# Send PMM message
#
proc pmm_send {msg} {
    global pmm c debug
    puts -nonewline $c $msg 
}


#
# Like pmm_send, but prefix our session name.
#
proc pmm_send_ {msg} {
    global pmm
    
    if [info exists pmm(sn)] {
	pmm_send "$pmm(sn)$msg"
    }
}

#
# ignore pmm "+" message
#
proc pmm_+ {args} {
}
 
#
# ignore pmm "agent" message
#
proc pmm_agent {args} {
}

proc send_update {} {
    global pmm
    if { [info exists pmm(update_interval)] } {
	pmm_send_ " active $pmm(update_interval)"
	after $pmm(update_interval) send_update
    }
}

proc pmm_ac {} {
    global srcstate pmm myself
    foreach {key value} [array get srcstate] {
	if { $key != $pmm($myself) } {
	    set pmm(bw) [ac ac_qos [$key ssrc] [$key rr_percent_lost] [$key rtt] $pmm(bw)]
	    grabber bps $pmm(bw)
	}
    }
    after 5000 pmm_ac
}
#
# Invoked when a pmm session command was received.
# Vic-specific since it invokes buttons.
#
proc pmm_session {sn c args} {
    global pmm portButton inputPort active myself 
    
    set w .menu

    switch $c {
# Application Control related Parameter

	control {
	    if {$args>0} {
		set pmm(ac) 1
		ac control $args [$pmm($myself) ssrc]
		pmm_ac
	    } else {
		if [info exists pmm(ac)] {
		    unset pmm(ac)
		    after cancel pmm_ac
		}
		ac control -1
	    }
	}

	update_interval {
	    set pmm(update_interval) $args
	    send_update
	}

	alpha {
	    ac alpha $args
	}
	
	algorithm {
	    ac algorithm $args
	}

	loss_lower_threshold {
	    ac loss_lower_threshold $args
	}

	loss_upper_threshold {
	    ac loss_upper_threshold $args
	}

	congested_recv_fraction {
	    ac congested_recv_fraction $args
	}

	loaded_recv_fraction {
	    ac loaded_recv_fraction $args
	}
	 
	multiplicative_decrease {
	    ac multiplicative_decrease $args
	}

	additive_increase  {
	    ac additive_increase $args
	}
#end application control

	open {
	    pmm_send "$sn opened"
	}

	hello {
	    set cname $args
	    if {$cname != $myself} {
		set pmm($pmm($cname)) $cname
		eval "proc pmm_$sn/$cname {c args} {pmm_member $sn $cname \$c \$args}"
	    }
	}

	cname {
	}

	media {
	    if {[winfo exists $w]} {
		xflush
		#set src $pmm($myself)
		#unset pmm($src,stolen) 
		switch [string tolower $args] {
		    nv {
			#select_format  nv
			#$w.encoder.f.h0.format.p0.b0 invoke 
		    }
		    jpeg {
			#select_format jpeg
			$w.encoder.f.h0.format.p1.b0 invoke
		    }
		    cellb {
			#select_format cellb
			$w.encoder.f.h0.format.p0.b2 invoke
		    }
		    h261 {
			#select_format h261
			$w.encoder.f.h0.format.p1.b1 invoke
		    }
		}
	    }
	}

	dither {
	    if {[winfo exists $w.decoder.f.h2.dither.h0]} {
		switch [string tolower $args] {
		    global V
		    ordered {
			$w.decoder.f.h2.dither.h0.b0 invoke
		    }
		    error_diff {
			$w.decoder.f.h2.dither.h0.b1 invoke
		    }
		    quantize {
			$w.decoder.f.h2.dither.h1.b2 invoke
		    }
		    gray {
			$w.decoder.f.h2.dither.h1.b3 invoke
		    }
		} 
	    } 
	}

	gamma {
	    if {[winfo exists $w.decoder.f.h2.bot.gamma]} {
		update_gamma $w.decoder.f.h2.bot.gamma.entry  $args
	    }
	}

	sending_slides{
	    global SendingSlides
	    set SendingSlides=$args
	    setFillRate
	}
	
	use_jpeg_for_h261{
	    global useJPEGforH261
	    set useJPEGforH261=$args
	    restart
	}
	max_bandwidth {
	    if {[winfo exists $w]} {
		ac max_bandwidth $args
		grabber bps $args
		set pmm(bw) $args
		#set_bps $w.cb.frame.right.bps.value $args
	    }
	}

	max_frame_rate {
	    if {[winfo exists $w]} {
		grabber fps $args
		#set_fps $w.cb.frame.right.fps.value $args
	    }
	}

	quality {
	    if {[winfo exists $w]} {
		global qscale_val qscale lastFmt
		if { [lindex [$qscale configure -state] 4] == "normal" } {
		    set cmd [lindex [$qscale configure -command] 4]
		    $cmd $args
		}
		$qscale set $args
	    }
	}

	send {
	    if {[winfo exists $w]} {
		#xflush
		set value  [string index [ lindex $args 0 ] 0 ]
		if {$value == 0 && $pmm(send) == 1} {
		    set pmm(send) 0
		    $w.cb.frame.buttons.send deselect
		    transmit
		} elseif { $value == 1 && $pmm(send) == 0} {
		    set pmm(send) 1
		    $w.cb.frame.buttons.send select
		    transmit
		}
	    }
	    pmm_send "$sn sending $value"
	}

	receive {
	    if [winfo exists $w] {
		xflush
		#set src [srctab local]
		set src $pmm($myself)
		#set cname [$src sdes cname]
		set mutebutton($src) [expr !$args]
		$src mute $mutebutton($src)
	    }
	}

	video_input {
	    if {[winfo exists $w]} {
		if {$args == 0} {
		    set inputPort "S-Video"
		    grabber port "S-Video"
		}
	    }
	    if {[winfo exists $w]} {
		if {$args == 1} {
		    set inputPort "Composite-1"
		    grabber port "Composite-1"
		}
	    }
	    if {[winfo exists $w]} {
		if {$args == 2} {
		    set inputPort "Composite-2"
		    grabber port "Composite-2"
		}
	    }
	    if {[winfo exists $w]} {
		if {$args == 4} {
		    set inputPort "Composite1"
		    grabber port "Composite1"
		}
	    }
	    if {[winfo exists $w]} {
		if {$args == 3} {
		    set inputPort "Television"
		    grabber port "Television"
		}
	    }
	    if {[winfo exists $w]} {
		if {$args == 5} {
		    set inputPort "SVHS"
		    grabber port "SVHS"
		}
	    }
	}

	video_device {
	    global inputDeviceList
	    set flag 0
	    foreach d $inputDeviceList {
		if {$args == [$d nickname] || "xil" == [$d nickname]} {
		    select_device $d
		    set flag 1
		    break
		}
	    }
            if {$args == "video4linux"} {
	        foreach d $inputDeviceList {
		    if {[string match "v4l*" [$d nickname]]} {
			select_device $d
			set flag 1
			break
		    }	
		}
            }
	    if {$flag == 0} {
		puts stderr "$pmm(tool): Warning: No such device: $args"
	    }
	}

	close {
	    adios
	}

	closed {
	    # remove procedures
	    # we have to do this after receiving our own "sn closed"
	    # message to avoid an error message
	    rename $sn {}
	    rename $sn/$myself {}
	}

	enlarge {
	    global pre
	    set src $pmm($myself)
	    set pre($src) $args
	    if {[info exists active($src)]} {
		xflush
		set w .top.grid.$src
		select_thumbnail $w $src
	    }
	}
    }
}


#
# Process member specific pmm messages: issue statistics
# Receive: mute button.
#
proc pmm_member {sn cname c args} {
    global pmm mutebutton fpshat bpshat active srcstate myself pre
    
    set w .menu
    switch $c {
	statistics? {
	    if {$cname != $myself} {
		if { [info exists pmm($cname)] } {
		    set src $pmm($cname)
		} else {
		    return
		}
	    } else {
		set src $pmm($myself)
	    }
	    set key $src
	    set stats [list "ssrc [$src srcid]"]
	    set received [$src np]
	    set expected [$src ns]
	    catch {lappend stats "received $received"}
	    catch {lappend stats "expected $expected"}
	    set lost [expr $expected - $received]
	    if {$lost >= 0} {
		catch {lappend stats "lost $lost"}
	    } else {
		catch {lappend stats "lost 0"}
	    }
	    if {$cname != $myself} {
		set rr_lost [$src rr_total_lost]
		catch {lappend stats "rr_lost $rr_lost"}
		set rr_fraction [$src rr_percent_lost]
		set rr_fraction [format "%1.5f" $rr_fraction]
		catch {lappend stats "rr_fraction $rr_fraction"}
	    }
	    if [info exists active($src)] {
		catch {lappend stats "actual_bandwidth [expr $bpshat($key)/1000]"}
		catch {lappend stats "actual_framerate $fpshat($key)"}
	    }
	    catch {lappend stats "media video"}
	    catch {lappend stats "last_packet {[$src lastdata]}"}
	    catch {lappend stats "last_ctrl {[$src lastctrl]}"}
	    if { $cname != "" } { 
		pmm_send "$sn/$cname statistics $stats"
	    }
	}

	receive {
	    set value  [string index [ lindex $args 0 ] 0 ]
	    if {[info exists pmm($cname)]} {
		set pmm($pmm($cname),receive) $value
		if { $value > 0} {
		    if { [info exists pmm($pmm($cname),stolen)] } { 
			if {$pmm($pmm($cname),stolen) > 0} {
			    xflush
			    set src $pmm($cname)
			    set mutebutton($src) [expr !$value]
			    $src mute $mutebutton($src)
			    if {  $cname != "" } { 
				pmm_send "$sn/$cname receiving $value"
			    }
			}
		    }
		} else {
		    xflush
		    set src $pmm($cname)
		    set mutebutton($src) [expr !$value]
		    $src mute $mutebutton($src)
		    if { $cname != "" } { 
			pmm_send "$sn/$cname receiving $value"
		    }
		}		    
	    }
	}
	enlarge {
	    global pre
	    if {[info exists pmm($cname)]} {
		set src $pmm($cname)
		set pre($src) $args
		xflush
		if {[info exists active($src)]} {
		    set w .top.grid.$src
		    select_thumbnail $w $src
		}
	    }
	}
	vic_win_id {
	    set win_id [lindex [lindex $args 0] 2]
	    if { $win_id == 0 } {
		global member
		#if { [info exists pmm($pmm($cname),active)] && $pmm($pmm($cname),active) > 0} {
		set pmm($pmm($cname),stolen) 1
		if { [info exists pmm($pmm($cname),receive)] && \
			 $pmm($pmm($cname),receive) == 1 } {
		    xflush
		    set src $pmm($cname)
		    set mutebutton($src)  &pmm($pmm($cname),receive)
		    $src mute $mutebutton($src)
		}
	    }
	    # could not steal window
	    if { $win_id < 0 } {
		    set pmm($pmm($cname),stolen) -1
	    }
	}
	    
    }	
}


# set main window geometry, hide certain buttons in main window
# we do not have to evaluate the geometry because 90x0 is sufficent
proc unpack_buttons_vic {} {


#    set minwidth [winfo reqwidth .]
#    set minheight [winfo reqheight .]
#
#    if { $minwidth < 200 } {
#	set minwidth 200
#    }
#    if { $minheight < 100 } {
#	set minheight 100
#    }
#    wm minsize . $minwidth $minheight

#    pack forget .top.bar.menu
#    pack forget .top.bar.help
    if { [winfo exists .top.label ] } {
	.top.label configure -text " "
	pack forget .top.label
    }
    wm withdraw .
    if { [winfo exists .top.bar ] } {
	.top.bar.title configure -text " " 
	.top.bar.menu configure -text " "
	.top.bar.quit configure -text " "
	.top.bar.help configure -text " "
	pack forget .top.bar
    }
    wm minsize . 100 1
    #wm resizable . 1 1
    wm geometry . 100x1+0+0
    wm deiconify .
}	


#
# if a remote source crashes and then starts again vic gives him another src id. To check if
# this prodeurde checks if for the current stream there exists another window
# evaluates the gived command for these sources
#
proc pmm_check_other { src command} {
    global pmm
    set cname [$src sdes cname]

    foreach i [array names pmm *stolen] {
	set name [lindex [split $i ,] 0 ]
	if { $name != $src } {
	    set other_cname [$name sdes cname]
	    if { $other_cname != "" && $other_cname==$cname} {
		$command $name
	    }
	}
    }
}
#
# for existant but unused entries (entries for sources that did not dend a bye message before 
# leaving) steal the windo back 
#
proc pmm_resteal_entry { src } {
    global pmm pre
    if {[info exist pmm($src,stolen)] && $pmm($src,stolen) > 0 } {
	set w .top.grid.${src}.stamp
	set win_id [format "%i" [winfo id $w.video]]
	# try to steal the window back 10 times at the most
	set counter 0
	while {[catch {xaccess eventreparent -parentwidget $w  -windowid $win_id} msg] \
		   && $counter < 10} {
	    set counter [expr $counter + 1]
	    after 500
	    update idletasks
	}
	if { $counter >=10 } {
	    adios
	}
	unset pmm($src,stolen)
    }
}    

#
# check if the window has been stolen
#
proc pmm_check_stealing {cname src w win_id} {
    global pmm
    if { [info exists pmm($cname)] && $pmm($cname) == $src } {
	if { [info exists pmm($src,stolen)] && $pmm($src,stolen) != 0 } {
	    # try it another time 
	    if { $cname != "" } { 
		pmm_send_ "/$cname vic_win_id $cname $w $win_id"
	    }
	 }
    }
}
#
# inform isc aboue the window to steal
#
proc pmm_steal_win { src } {
    global pmm
    set cname [$src sdes cname]

    foreach type {name email tool} { 
	if {![info exists pmm($src,$type)] || $pmm($src,$type) == ""} {
	    set pmm($src,$type) [$src sdes $type]
	    if { $cname != "" && $pmm($src,$type) != "" } { 
		pmm_send_ "/$cname $type $pmm($src,$type)"
	    }
	}
    }

    # make first sure that all the other entries that might be attached to this steam
    # have a valid window by stealing the original window back from isc
    pmm_check_other $src pmm_resteal_entry

    set cname [$src sdes cname]
    #Send vic window name/id binding to session controller
    set w .top.grid.${src}.stamp.video
    if {$cname != "" && [winfo exists $w] } {
	if { ![info exists pmm($src,stolen)] } {
	    set win_id [format "%i" [winfo id $w]]
	    
	    # while we are at make sure that isc knows this member is active
	    if { $cname != "" } { 
		pmm_send_ "/$cname active"
	    }

	    if { $cname != "" } { 
		pmm_send_ "/$cname vic_win_id $cname $w $win_id"
	    }

	    # check if the window was actually stolen and if not repeat that
	    after 5000 "pmm_check_stealing $cname $src $w $win_id"

	    set pmm($cname,src) 1
	    unpack_buttons_src $src
	} else {
	    if { [info exists pmm($src,stolen)] && $pmm($src,stolen)<0 } {
		set win_id [format "%i" [winfo id $w]]
		if { $cname != "" } { 
		    pmm_send_ "/$cname vic_win_id $cname $w $win_id"
		}

		# check if the window was actually stolen and if not repeat that
		after 5000 "pmm_check_stealing $cname $src $w $win_id"

		# while we are at make sure that isc knows this member is active
		if { $cname != "" } { 
		    pmm_send_ "/$cname active"
		}
		set pmm($cname,src) 1
		unpack_buttons_src $src
	    }
	}
    }
}	    

#
# unset all variables attached to a specific src
proc pmm_unset_src { src }  {
    global pmm
    foreach e [array names pmm] {
	if {[string match $src,* $e]} {
	    unset pmm($e)
	}
    }
}


proc trace_proc {var_name var_ind op} {
    global pmm srcstate active src_name nametag pre member

    set src $var_ind
    set cname [$src sdes cname]
    #set pmm($src,active) 1
  
    switch $var_name {
	active {
	    if {$op == "w"}  {
		if  {  [info exists pmm($src)]  }   {
		    if  { $active($src) == 0 }  {
			set pmm($src,active) 0
			pmm_resteal_entry $src
			# close any big window attached to this member
			if [info exist pre($src)] {
			    if { $pre($src) > 0 } {
				xflush
				set w .top.grid.$src
				select_thumbnail $w $src
			    }
			    unset pre($src)
			}
			if { [info exists pmm($cname)] && $pmm($cname) == $src } {
			    if { $cname != "" } { 
				pmm_send_ "/$cname !active"	
			    }
			}
		    } else {
			if { $cname != "" } { 
			    pmm_send_ "/$cname active"
			}
			# set this entry to become the actual one for this stream
			set pmm($cname) $src
			set pmm($src,active) 1


			# inform isc to steal the window
			pmm_steal_win $src
		    }
		}
		pmm_activate $src 
		# If preview is already set to 1, open preview window 
		if  { [info exists pre($src)] && $pre($src) == 1 } {
		    set w .top.grid.$src
		    select_thumbnail $w $src
		}
	
	    } elseif  {$op == "u"}  {
		pmm_deactivate
		#pmm_unregister $src
	    }
	}

	srcstate {
	    if {$op == "w"} {
		if {$cname != ""  && ![info exists pmm($src)]} {
		    set pmm($cname) $src
		    if { $cname != "" } {
			pmm_send_ " hello $cname"
		    }
		    # inform isc to steal the window
		    pmm_steal_win $src
		} elseif {$cname != ""} {
		    if {$srcstate($src) == 1} {
			if { $cname != "" } {
			    pmm_send_ "/$cname alive"
			}
			# set this entry to become the actual one for this stream
			set pmm($cname) $src

			# inform isc to steal the window
			pmm_steal_win $src
		    }
		    if {$srcstate($src) == 2 } {
			# steal the window back
			pmm_resteal_entry $src

			if { [info exists pmm($cname)] && $pmm($cname) == $src } {
			    if { $cname != "" } {
				pmm_send_ "/$cname !alive"
			    }
			}
		    }
		}
	    }
	    
	}

	src_name {
	    if {$op == "w"} {
		if {$cname != ""  && ![info exists pmm($src)]} {
		    set pmm($cname) $src
		    if { $cname != "" } {
			pmm_send_ " hello $cname"
		    }
		    # inform isc to steal the window
		    pmm_steal_win $src
		}
	    }
	}

	nametag {
	    set w .top.grid.${src}.stamp.video
	    if {$op == "u" } {
		xflush
		if { $cname != "" } {
		    pmm_send_ "/$cname bye"
		}
		#ac delete_entry [$src ssrc]
		if {[info exists pmm($src)]} {
		    #unset srcstate($src)
		    unset pmm($src)
		}
		if  { [info exists  pmm($cname) ] } { 
		    unset pmm($cname)
		}
		if { [info exists $pmm(sn) ] && [info exists cname]} {
		    rename pmm_$pmm(sn)/$cname {}
		}
		# unset all variables attached to this src
		pmm_unset_src $src

		if [info exist pre($src)] {
		    unset pre($src)
		}
	    }	
	}
	
	colorbutton {
	    if {$op == "w"} {
		global colorbut
		if {![info exists colorbut($src)]} {
		    set colorbut($src) 1
		    unpack_buttons_src $src
		}
	    }
	}
    }


}

# change the size of vic window when the new user joins the session
proc pmm_activate {src} {
    set geom [winfo geometry .]
    regsub {([0-9]+\x)([0-9]+)(\+.*)} $geom {\2} temp
    set temp [expr $temp + 65]
    regsub {\x([0-9]+)\+} $geom \x$temp\+ x
#    .top.grid.$src config  -geometry 80x0
    wm deiconify .
}

# change the size of vic window when the user leaves the session
proc pmm_deactivate {} {
    set geom [winfo geometry .]
    regsub {([0-9]+\x)([0-9]+)(\+.*)} $geom {\2} temp
    set temp [expr $temp - 65]
    regsub {\x([0-9]+)\+} $geom \x$temp\+ x
    #wm geometry . $x
    set geom $x
}

# hide certain buttons in control menu
proc unpack_buttons_src {src} {
    pack forget .top.grid.$src.r.ctrl.mute
    pack forget .top.grid.$src.r.ctrl.info
    pack forget .top.grid.$src.r.ctrl.color
    
}

proc send_sdes_resources {src} {
    global src_info src_nickname src_name
    set cname [$src sdes cname]
    foreach sdes_item [resource sdesList] {
	if {[$src sdes $sdes_item] != ""} {
	    if { $cname != "" } {
		pmm_send_ "/$cname $sdes_item [$src sdes $sdes_item]"
	    }
	}
    }
    if {([info exists src_info($src)]) && ($src_info($src) != "")} {
	if { $cname != "" } {
	    pmm_send_ "/$cname cname $src_name($src)"
	}
    }
    if {([info exists src_nickname($src)]) && ($src_nickname($src) != "")} {
	if { $cname != "" } {
	    pmm_send_ "/$cname name $src_nickname($src)"
	}
    }
}

proc redraw_windows {} {
    foreach src [session active] {
	set d [$src handler]
	if { $d != "" } {
	    #dputs "$d redraw"
	    $d redraw
		}
	}
}

proc dputs {arg} {
    global pmm
    if {$pmm(debug)} {
	puts stdout $arg
    }
}
    
